/*
 * Direct3D8 textures
 *  Copyright (c) 2002, TransGaming Technologies Inc.
*/

#include "config.h"

#include "winbase.h"
#include "d3d8.h"
#include "debugtools.h"

#include "d3d8_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

static D3D_Light default_light = {
  {D3DLIGHT_DIRECTIONAL,
   {1,1,1,0},
   {0,0,0,0},
   {0,0,0,0},
   {0,0,0},
   {0,0,1},
   0,0,0,0,0,0,0},
  0, 0, {0,0,0,0}
};

/* table to convert from D3D8 world transform to D3D7 world transform */
static UINT wxfrm[4] = {1,4,5,6};

HRESULT WINAPI Direct3DDevice8_SetTransform(LPDIRECT3DDEVICE8 iface,
					    D3DTRANSFORMSTATETYPE dtstType,
					    CONST D3DMATRIX* lpD3DMatrix)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETTRANSFORM, 0, {1} };
  D3DHAL_DP2SETTRANSFORM xfrm;

  TRACE("(%p)->(%d,%p)\n", This, dtstType, lpD3DMatrix);
  if (!dtstType) return D3DERR_INVALIDCALL;

  if (dtstType >= 256) {
    unsigned i = dtstType - 256;
    if (i >= 4) return D3DERR_INVALIDCALL;
    dtstType = wxfrm[dtstType-256];
  }

  if (dtstType >= D3DTS_TEXTURE0) {
    unsigned i = dtstType - D3DTS_TEXTURE0;
    if (i >= 8) return D3DERR_INVALIDCALL;
    This->d3dp.txfrm[i] = *lpD3DMatrix;
  }
  else if (dtstType < 8) {
    This->d3dp.xfrm[dtstType] = *lpD3DMatrix;
  }
  xfrm.xfrmType = dtstType;
  xfrm.matrix = *lpD3DMatrix;
  return This->emit_dp2(This->ip_ctx, &cmd, &xfrm, sizeof(xfrm), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_GetTransform(LPDIRECT3DDEVICE8 iface,
					    D3DTRANSFORMSTATETYPE dtstType,
					    D3DMATRIX* lpD3DMatrix)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%d,%p)\n", This, dtstType, lpD3DMatrix);
  if (!dtstType) return D3DERR_INVALIDCALL;

  if (dtstType >= 256) {
    unsigned i = dtstType - 256;
    if (i >= 4) return D3DERR_INVALIDCALL;
    dtstType = wxfrm[dtstType-256];
  }

  if (dtstType >= D3DTS_TEXTURE0) {
    unsigned i = dtstType - D3DTS_TEXTURE0;
    if (i >= 8) return D3DERR_INVALIDCALL;
    *lpD3DMatrix = This->d3dp.txfrm[i];
  }
  else if (dtstType < 8) {
    *lpD3DMatrix = This->d3dp.xfrm[dtstType];
  }
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetViewport(LPDIRECT3DDEVICE8 iface,
					   CONST D3DVIEWPORT8* lpData)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { 0, 0, {1} };
  D3DHAL_DP2VIEWPORTINFO view;
  D3DHAL_DP2ZRANGE zrange;
  HRESULT hr;

  TRACE("(%p)->(%p)\n", This, lpData);
  memcpy(&This->d3dp.viewport, lpData, sizeof(D3DVIEWPORT7));
  view.dwX = lpData->X;
  view.dwY = lpData->Y;
  view.dwWidth  = lpData->Width;
  view.dwHeight = lpData->Height;
  zrange.dvMinZ = lpData->MinZ;
  zrange.dvMaxZ = lpData->MaxZ;
  cmd.bCommand = D3DDP2OP_VIEWPORTINFO;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &view, sizeof(view), NULL, 0);
  /* FIXME: do we also need to do anything about WINFO? */
  if (SUCCEEDED(hr)) {
      cmd.bCommand = D3DDP2OP_ZRANGE;
      hr = This->emit_dp2(This->ip_ctx, &cmd, &zrange, sizeof(zrange), NULL, 0);
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice8_GetViewport(LPDIRECT3DDEVICE8 iface,
					   D3DVIEWPORT8* lpData)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%p)\n", This, lpData);
  memcpy(lpData, &This->d3dp.viewport, sizeof(D3DVIEWPORT7));
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetMaterial(LPDIRECT3DDEVICE8 iface,
					   CONST D3DMATERIAL8* lpMat)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETMATERIAL, 0, {1} };

  TRACE("(%p)->(%p)\n", This, lpMat);
#if 0
  _dump_colorvalue("Diffuse", lpMat->u.diffuse);
  _dump_colorvalue("Ambient", lpMat->u1.ambient);
  _dump_colorvalue("Specular", lpMat->u2.specular);
  _dump_colorvalue("Emissive", lpMat->u3.emissive);
  TRACE(" Power : %f\n", lpMat->u4.power);
#endif
  return This->emit_dp2(This->ip_ctx, &cmd,
			(D3DMATERIAL7*)lpMat, sizeof(This->d3dp.material), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_GetMaterial(LPDIRECT3DDEVICE8 iface,
					   D3DMATERIAL8* lpMat)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%p)\n", This, lpMat);
  *(D3DMATERIAL7*)lpMat = This->d3dp.material;
  return D3D_OK;
}

static HRESULT CheckExpandLights(IDirect3DDevice8Impl *This, DWORD dwIndex)
{
  if (dwIndex >= This->d3dp.numlights) {
    DWORD newlights = dwIndex+1, cnt;
    D3DHAL_DP2COMMAND cmd = { D3DDP2OP_CREATELIGHT, 0, {1} };
    D3DHAL_DP2CREATELIGHT data;
    D3D_Light *newlight;
    HRESULT hr;

    newlight = HeapReAlloc(GetProcessHeap(),0,This->d3dp.light,sizeof(D3D_Light)*newlights);
    if (!newlight) {
      ERR("out of memory\n");
      return DDERR_OUTOFMEMORY;
    }
    for (cnt=This->d3dp.numlights; cnt<newlights; cnt++)
      memcpy(&newlight[cnt], &default_light, sizeof(D3D_Light));
    data.dwIndex = dwIndex;
    hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
    if (FAILED(hr)) {
      ERR("CREATELIGHT failed (%ld)\n", hr);
      return hr;
    }
    This->d3dp.light = newlight;
    This->d3dp.numlights = newlights;
  }
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetLight(LPDIRECT3DDEVICE8 iface,
					DWORD dwIndex,
					CONST D3DLIGHT8* lpLight)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETLIGHT, 0, {1} };
  D3DHAL_DP2SETLIGHT data;
  HRESULT hr;

  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, lpLight);
  hr = CheckExpandLights(This, dwIndex);
  if (FAILED(hr)) return hr;
  This->d3dp.light[dwIndex].l = *(D3DLIGHT7*)lpLight;
  data.dwIndex = dwIndex;
  data.dwDataType = D3DHAL_SETLIGHT_DATA;
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data),
			&This->d3dp.light[dwIndex].l, sizeof(D3DLIGHT7));
}

HRESULT WINAPI Direct3DDevice8_GetLight(LPDIRECT3DDEVICE8 iface,
					DWORD dwIndex,
					D3DLIGHT8* lpLight)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, lpLight);
  if (dwIndex < This->d3dp.numlights)
    *(D3DLIGHT7*)lpLight = This->d3dp.light[dwIndex].l;
  else
    *(D3DLIGHT7*)lpLight = default_light.l;
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_LightEnable(LPDIRECT3DDEVICE8 iface,
					   DWORD dwIndex,
					   BOOL bEnable)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_SETLIGHT, 0, {1} };
  D3DHAL_DP2SETLIGHT data;
  DWORD oldflags;
  HRESULT hr;

  TRACE("(%p)->(%ld,%d)\n", This, dwIndex, bEnable);
  hr = CheckExpandLights(This, dwIndex);
  if (FAILED(hr)) return hr;
  oldflags = This->d3dp.light[dwIndex].flags;
  if (bEnable) {
    This->d3dp.light[dwIndex].flags |= D3DLIGHT_ACTIVE;
    data.dwDataType = D3DHAL_SETLIGHT_ENABLE;
  } else {
    This->d3dp.light[dwIndex].flags &= ~D3DLIGHT_ACTIVE;
    data.dwDataType = D3DHAL_SETLIGHT_DISABLE;
  }
  data.dwIndex = dwIndex;
  if (This->d3dp.light[dwIndex].flags == oldflags) return D3D_OK;
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_GetLightEnable(LPDIRECT3DDEVICE8 iface,
					      DWORD dwIndex,
					      LPBOOL lpbEnable)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, lpbEnable);
  if (dwIndex < This->d3dp.numlights)
    *lpbEnable = This->d3dp.light[dwIndex].flags & D3DLIGHT_ACTIVE;
  else
    *lpbEnable = FALSE;
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetClipPlane(LPDIRECT3DDEVICE8 iface,
					    DWORD dwIndex,
					    CONST float* pPlane)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_RENDERSTATE, 0, {1} };
  D3DHAL_DP2SETCLIPPLANE data;

  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, pPlane);
  memcpy(&This->d3dp.clipplane[dwIndex], pPlane, sizeof(float)*4);
  data.dwIndex = dwIndex;
  memcpy(&data.plane, pPlane, sizeof(float)*4);
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_GetClipPlane(LPDIRECT3DDEVICE8 iface,
					    DWORD dwIndex,
					    float* pPlane)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%ld,%p)\n", This, dwIndex, pPlane);
  memcpy(pPlane, &This->d3dp.clipplane[dwIndex], sizeof(float)*4);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetRenderState(LPDIRECT3DDEVICE8 iface,
					      D3DRENDERSTATETYPE d3dType,
					      DWORD dwState)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_RENDERSTATE, 0, {1} };
  D3DHAL_DP2RENDERSTATE stat;

  TRACE("(%p)->(%d,%08lx)\n", This, d3dType, dwState);
  if (d3dType >= 256) {
    WARN("out-of-range render state (%d)\n", d3dType);
    return D3DERR_INVALIDCALL;
  }
  /* according to the HAL docs, it seems the driver is
   * responsible for updating the render state array
   * to what it's capable of handling */
  stat.RenderState = d3dType;
  stat.u.dwState = dwState;
  return This->emit_dp2(This->ip_ctx, &cmd, &stat, sizeof(stat), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_GetRenderState(LPDIRECT3DDEVICE8 iface,
					      D3DRENDERSTATETYPE d3dType,
					      LPDWORD lpdwState)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  if (d3dType >= 256) {
    WARN("out-of-range render state (%d)\n", d3dType);
    return D3DERR_INVALIDCALL;
  }
  *lpdwState = This->d3dp.renderstate[d3dType];
  TRACE("(%p)->(%d,%p) => %08lx\n", This, d3dType, lpdwState, *lpdwState);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_BeginStateBlock(LPDIRECT3DDEVICE8 iface)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3D8StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)\n", This);
  if (This->current->state_block) return D3DERR_INVALIDCALL;

  block = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(D3D8StateBlock));

  hr = This->alloc_state_block(This->ip_ctx, &block->state_block);
  if (FAILED(hr)) {
    HeapFree(GetProcessHeap(), 0, block);
    return hr;
  }

  data.dwOperation = D3DHAL_STATESETBEGIN;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (SUCCEEDED(hr)) This->current = block;
  return hr;
}

HRESULT WINAPI Direct3DDevice8_EndStateBlock(LPDIRECT3DDEVICE8 iface,
					     LPDWORD lpdwBlock)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3D8StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;

  TRACE("(%p)->(%p)\n", This, lpdwBlock);
  if (!This->current->state_block) return D3DERR_INVALIDCALL;

  block = This->current;
  *lpdwBlock = (DWORD)block;
  This->current = &This->state;
  TRACE(" => %ld\n", *lpdwBlock);

  data.dwOperation = D3DHAL_STATESETEND;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  return This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_ApplyStateBlock(LPDIRECT3DDEVICE8 iface,
					       DWORD dwBlock)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3D8StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)->(%ld)\n", This, dwBlock);
  if (This->current->state_block) return D3DERR_INVALIDCALL;

  block = (D3D8StateBlock*)dwBlock;

  data.dwOperation = D3DHAL_STATESETEXECUTE;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (SUCCEEDED(hr)) {
    /* update current state to match new device state */
    unsigned u;
    for (u=0; u<8; u++) {
      if (This->state.tex[u] != block->tex[u]) {
        if (block->tex[u]) IDirect3DBaseTexture8_AddRef(block->tex[u]);
        if (This->state.tex[u]) IDirect3DBaseTexture8_Release(This->state.tex[u]);
        This->state.tex[u] = block->tex[u];
      }
    }
    This->dirty_textures = TRUE; /* just in case */
    /* we could also do states that the driver doesn't know about here */
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice8_CaptureStateBlock(LPDIRECT3DDEVICE8 iface,
						 DWORD dwBlock)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3D8StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)->(%ld)\n", This, dwBlock);
  if (This->current->state_block) return D3DERR_INVALIDCALL;

  block = (D3D8StateBlock*)dwBlock;

  data.dwOperation = D3DHAL_STATESETCAPTURE;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (SUCCEEDED(hr)) {
    /* update state block to match current state */
    unsigned u;
    for (u=0; u<8; u++) {
      if (This->state.tex[u] != block->tex[u]) {
        if (This->state.tex[u]) IDirect3DBaseTexture8_AddRef(This->state.tex[u]);
        if (block->tex[u]) IDirect3DBaseTexture8_Release(block->tex[u]);
        block->tex[u] = This->state.tex[u];
      }
    }
    /* we could also do states that the driver doesn't know about here */
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice8_DeleteStateBlock(LPDIRECT3DDEVICE8 iface,
						DWORD dwBlock)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3D8StateBlock* block;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_STATESET, 0, {1} };
  D3DHAL_DP2STATESET data;
  HRESULT hr;

  TRACE("(%p)->(%ld)\n", This, dwBlock);

  block = (D3D8StateBlock*)dwBlock;

  data.dwOperation = D3DHAL_STATESETDELETE;
  data.dwParam = block->state_block;
  data.sbType = D3DSBT_ALL;
  hr = This->emit_dp2(This->ip_ctx, &cmd, &data, sizeof(data), NULL, 0);
  if (FAILED(hr)) return hr;
  hr = This->free_state_block(This->ip_ctx, dwBlock);
  if (SUCCEEDED(hr)) {
    /* release referenced objects */
    unsigned u;
    for (u=0; u<8; u++)
      if (block->tex[u]) IDirect3DBaseTexture8_Release(block->tex[u]);
    HeapFree(GetProcessHeap(), 0, block);
  }
  return hr;
}

HRESULT WINAPI Direct3DDevice8_GetTexture(LPDIRECT3DDEVICE8 iface,
					  DWORD dwStage,
					  LPDIRECT3DBASETEXTURE8* lplpTexture)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  *lplpTexture = This->state.tex[dwStage];
  TRACE("(%p)->(%ld,%p) => %p\n", This, dwStage, lplpTexture, *lplpTexture);
  if (*lplpTexture) IDirect3DBaseTexture8_AddRef(*lplpTexture);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetTexture(LPDIRECT3DDEVICE8 iface,
					  DWORD dwStage,
					  LPDIRECT3DBASETEXTURE8 lpTexture)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  LPDIRECT3DBASETEXTURE8 oldtex;
  IDirect3DBaseTexture8Impl* tex = (IDirect3DBaseTexture8Impl*)lpTexture;
  /* in reality, Direct3D allocates "surface handles" that it
   * uses for the state here, but since we don't have the memory
   * restrictions that real drivers do (and we're very lazy too),
   * we'll just cast real pointers here, at least for now */
  DWORD dwState = 0;
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_TEXTURESTAGESTATE, 0, {1} };
  D3DHAL_DP2TEXTURESTAGESTATE stat;

  TRACE("(%p)->(%ld,%p)\n", This, dwStage, lpTexture);

  oldtex = This->current->tex[dwStage];
  if (lpTexture == oldtex) return D3D_OK;

  /* maintain reference count */
  if (lpTexture) IDirect3DBaseTexture8_AddRef(lpTexture);
  This->current->tex[dwStage] = lpTexture;
  if (oldtex) IDirect3DBaseTexture8_Release(oldtex);

  /* find texture object's first surface */
  if (lpTexture)
    dwState = TEXTURE_HANDLE(tex);

  if (!This->current->state_block) {
    This->d3dp.texstagestate[dwStage][D3DTSS_TEXTUREMAP] = dwState;
    if (lpTexture && tex->IsDirty) This->dirty_textures = TRUE;
    TRACE("saving\n");
  }
  TRACE("val=%08lx\n", dwState);
  stat.wStage = dwStage;
  stat.TSState = D3DTSS_TEXTUREMAP;
  stat.dwValue = dwState;
  return This->emit_dp2(This->ip_ctx, &cmd, &stat, sizeof(stat), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_GetTextureStageState(LPDIRECT3DDEVICE8 iface,
						    DWORD dwStage,
						    D3DTEXTURESTAGESTATETYPE d3dType,
						    LPDWORD lpdwState)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  DWORD dwState;

  /* tables to convert from D3D7 texture filters to D3D8 texture filters */
  static const DWORD magfilter[6] = {0, 1, 2, 4, 5, 3};
  static const DWORD minfilter[4] = {0, 1, 2, 3};
  static const DWORD mipfilter[4] = {0, 0, 1, 2};

  if (d3dType >= 32) {
    WARN("out-of-range texture stage state %d\n", d3dType);
    return D3DERR_INVALIDCALL;
  }
  dwState = This->d3dp.texstagestate[dwStage][d3dType];

  /* convert from D3D7 (HAL) texture filters to D3D8 texture filters */
  switch (d3dType) {
  case D3DTSS_MAGFILTER:
    dwState = magfilter[dwState];
    break;
  case D3DTSS_MINFILTER:
    dwState = minfilter[dwState];
    break;
  case D3DTSS_MIPFILTER:
    dwState = mipfilter[dwState];
    break;
  default:
    break;
  }

  TRACE("(%p)->(%ld,%d,%p) => %ld\n", This, dwStage, d3dType, lpdwState, dwState);
  *lpdwState = dwState;

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetTextureStageState(LPDIRECT3DDEVICE8 iface,
						    DWORD dwStage,
						    D3DTEXTURESTAGESTATETYPE d3dType,
						    DWORD dwState)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  D3DHAL_DP2COMMAND cmd = { D3DDP2OP_TEXTURESTAGESTATE, 0, {1} };
  D3DHAL_DP2TEXTURESTAGESTATE stat;

  /* tables to convert from D3D8 texture filters to D3D7 texture filters */
  static const DWORD magfilter[6] = {0, 1, 2, 5, 3, 4};
  static const DWORD minfilter[6] = {0, 1, 2, 3, 3, 3};
  static const DWORD mipfilter[6] = {1, 2, 3, 3, 3, 3};

  TRACE("(%p)->(%ld,%d,%ld)\n", This, dwStage, d3dType, dwState);

  if (d3dType >= 32) {
    WARN("out-of-range texture stage state %d\n", d3dType);
    return D3DERR_INVALIDCALL;
  }

  /* convert from D3D8 texture filters to D3D7 (HAL) texture filters */
  switch (d3dType) {
  case D3DTSS_MAGFILTER:
    dwState = magfilter[dwState];
    break;
  case D3DTSS_MINFILTER:
    dwState = minfilter[dwState];
    break;
  case D3DTSS_MIPFILTER:
    dwState = mipfilter[dwState];
    break;
  default:
    break;
  }

  /* the doc implies that the app is allowed to set any state it wants,
   * _then_ check (with ValidateDevice) whether the hardware groks the
   * resulting state vector, so I guess no checks should be performed here
   */
  if (!This->current->state_block)
    This->d3dp.texstagestate[dwStage][d3dType] = dwState;
  stat.wStage = dwStage;
  stat.TSState = d3dType;
  stat.dwValue = dwState;
  return This->emit_dp2(This->ip_ctx, &cmd, &stat, sizeof(stat), NULL, 0);
}

HRESULT WINAPI Direct3DDevice8_ValidateDevice(LPDIRECT3DDEVICE8 iface,
					      DWORD* pNumPasses)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%p)\n", This, pNumPasses);
  return This->validate_device(This->ip_ctx, pNumPasses);
}

HRESULT WINAPI Direct3DDevice8_SetStreamSource(LPDIRECT3DDEVICE8 iface,
					       UINT StreamNumber,
					       LPDIRECT3DVERTEXBUFFER8 pStreamData,
					       UINT Stride)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  LPDIRECT3DVERTEXBUFFER8 olddata;
  IDirect3DVertexBuffer8Impl* buffer = (IDirect3DVertexBuffer8Impl *)pStreamData;

  TRACE("(%p)->(%d,%p,%d)\n", This, StreamNumber, pStreamData, Stride);

  olddata = This->streamsource[StreamNumber];

  if (pStreamData != olddata) {
    if (pStreamData) IDirect3DVertexBuffer8_AddRef(pStreamData);
    This->streamsource[StreamNumber] = pStreamData;
    if (olddata) IDirect3DVertexBuffer8_Release(olddata);
  }

  This->d3dp.stream[StreamNumber].data_t = &(buffer->t);
  This->d3dp.stream[StreamNumber].stride = Stride;

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_GetStreamSource(LPDIRECT3DDEVICE8 iface,
					       UINT StreamNumber,
					       LPDIRECT3DVERTEXBUFFER8* ppStreamData,
					       UINT* pStride)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  *ppStreamData = This->streamsource[StreamNumber];
  *pStride = This->d3dp.stream[StreamNumber].stride;
  TRACE("(%p)->(%d,%p,%p)\n", This, StreamNumber, ppStreamData, pStride);
  if (*ppStreamData) IDirect3DVertexBuffer8_AddRef(*ppStreamData);
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetIndices(LPDIRECT3DDEVICE8 iface,
					  LPDIRECT3DINDEXBUFFER8 pIndexData,
					  UINT BaseIndex)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  LPDIRECT3DINDEXBUFFER8 olddata;

  TRACE("(%p)->(%p,%d)\n", This, pIndexData, BaseIndex);

  olddata = This->index_data;

  if (pIndexData != olddata) {
    if (pIndexData) IDirect3DIndexBuffer8_AddRef(pIndexData);
    This->index_data = pIndexData;
    if (olddata) IDirect3DIndexBuffer8_Release(olddata);
  }

  This->base_index = BaseIndex;

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_GetIndices(LPDIRECT3DDEVICE8 iface,
					  LPDIRECT3DINDEXBUFFER8* ppIndexData,
					  UINT* pBaseIndex)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  *ppIndexData = This->index_data;
  *pBaseIndex = This->base_index;
  TRACE("(%p)->(%p,%p)\n", This, ppIndexData, pBaseIndex);
  if (*ppIndexData) IDirect3DIndexBuffer8_AddRef(*ppIndexData);
  return D3D_OK;
}
