/*	Direct3D driver for OpenGL
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */

#include "config.h"

#include "winbase.h"
#include "wine/winbase16.h"
#include "winreg.h"
#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "d3dgl_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define IS_OPTION_TRUE(ch) \
    ((ch) == 'y' || (ch) == 'Y' || (ch) == 't' || (ch) == 'T' || (ch) == '1')

/* config options */
int use_anisotropic_texture_filtering=0;  /* default to no */
int use_vertex_shaders=1;                 /* default to yes */
int use_clipspace_fix=1;                  /* default to yes */

static void fill_prim_caps(DD_GL_data *gld, LPD3DPRIMCAPS pc);
static void check_exts(DD_GL_data *gld);
static void add_texture_fmts(DD_GL_data *gld, LPD3DHAL_GLOBALDRIVERDATA lpData);
static void add_zbuffer_fmts(DD_GL_data *gld, LPDDRAWI_DIRECTDRAW_GBL lpDD);


/***********************************************************************
 * get_config_key
 *
 * Get a config key from either the app-specific or the default config
 */
inline static DWORD get_config_key( HKEY defkey, HKEY appkey, const char *name,
                                    char *buffer, DWORD size )
{
    if (appkey && !RegQueryValueExA( appkey, name, 0, NULL, buffer, &size )) return 0;
    return RegQueryValueExA( defkey, name, 0, NULL, buffer, &size );
}

/***********************************************************************
 * setup_options
 *
 * setup d3dgl specific options
 */
static void setup_options(void)
{
  char buffer[MAX_PATH+16];
  HKEY hkey, appkey = 0;

  if (RegCreateKeyExA( HKEY_LOCAL_MACHINE, "Software\\Wine\\Wine\\Config\\d3dgl", 0, NULL,
                       REG_OPTION_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL )) {
    ERR("Cannot create config registry key\n");
    return;
  }
  
  /* open the app-specific key */
  if (GetModuleFileName16( GetCurrentTask(), buffer, MAX_PATH ) ||
      GetModuleFileNameA( 0, buffer, MAX_PATH )) {
    HKEY tmpkey;
    char *p, *appname = buffer;
    if ((p = strrchr( appname, '/' ))) appname = p + 1;
    if ((p = strrchr( appname, '\\' ))) appname = p + 1;
    strcat( appname, "\\d3dgl" );
    if (!RegOpenKeyA( HKEY_LOCAL_MACHINE, "Software\\Wine\\Wine\\Config\\AppDefaults", &tmpkey ))
    {
	if (RegOpenKeyA( tmpkey, appname, &appkey )) appkey = 0;
	RegCloseKey( tmpkey );
    }
  }

  if (!get_config_key( hkey, appkey, "AnisotropicTextureFiltering", buffer, sizeof(buffer) ))
    use_anisotropic_texture_filtering = IS_OPTION_TRUE( buffer[0] );
  TRACE("AnisotropicTextureFiltering=(%d)\n", use_anisotropic_texture_filtering);

  if (!get_config_key( hkey, appkey, "VertexShaders", buffer, sizeof(buffer) ))
    use_vertex_shaders = IS_OPTION_TRUE( buffer[0] );
  TRACE("VertexShaders=(%d)\n", use_vertex_shaders);

  if (!get_config_key( hkey, appkey, "ClipSpaceFix", buffer, sizeof(buffer) ))
    use_clipspace_fix = IS_OPTION_TRUE( buffer[0] );
  TRACE("ClipSpaceFix=(%d)\n", use_clipspace_fix);
  
  if (appkey) RegCloseKey( appkey );
  RegCloseKey( hkey );

}

void WINAPI D3DGL_Init(LPDDRAWI_DIRECTDRAW_GBL dd)
{
    DD_GL_data *gld;
    /* the "cdecl" tool is useful for something... */
    void (*(*gl_GetProcAddress)(const GLubyte *))(void);
    LPD3DHAL_GLOBALDRIVERDATA lpData;
    LPD3DHAL_D3DEXTENDEDCAPS lpExt;
    LPD3DHAL_CALLBACKS lpCB;
    ULONG_PTR dwhContext;
    GLint tmpi;

    if (dd->dwReserved3) /* already initialized - but since we're being called
                            again, the previous data has been wiped, so we have
                            to fill it in again */
    {
        gld = (DD_GL_data *)dd->dwReserved3;
    }
    else
    {
        gld = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(DD_GL_data));
        dd->dwReserved3 = (ULONG_PTR)gld;
    }

    lpData = (LPD3DHAL_GLOBALDRIVERDATA)dd->lpD3DGlobalDriverData;
    lpExt = (LPD3DHAL_D3DEXTENDEDCAPS)dd->lpD3DExtendedCaps;
    lpCB = (LPD3DHAL_CALLBACKS)dd->lpD3DHALCallbacks;

    /* x11drv puts glXGetProcAddress in here */
    gl_GetProcAddress = lpCB->lpReserved10;
    /* retrieve available extension procs */
    if (gl_GetProcAddress) {
	void (*extproc)(void);
#define DD_GL_PROC(x,y) extproc = gl_GetProcAddress(#y); gld->y = (x)extproc;
	DD_GL_PROC_LIST
#undef DD_GL_PROC
    } else {
#define DD_GL_PROC(x,y) gld->y = (x)NULL;
	DD_GL_PROC_LIST
#undef DD_GL_PROC
    }

    /* In order to retrieve actual device capabilities for an OpenGL device,
     * a context must be created for it first... eek */
    {
	D3DHAL_CONTEXTCREATEDATA data;
	data.u1.lpDDLcl = NULL;
	data.u2.lpDDSLcl = NULL;
	data.u3.lpDDSZLcl = NULL;
	data.u4.dwPID = 0;
	lpCB->ContextCreate(&data);
	if (data.ddrval != D3D_OK) {
	    ERR("cannot acquire OpenGL device capabilities\n");
	    return;
	}
	dwhContext = data.dwhContext;
    }
    TRACE("created temp context %08x\n", dwhContext);
    check_exts(gld);

    /* load options from config file */
    setup_options();

    /* Disable texture compression for now... Black&White requires the
     * Blt method to support texture decompression if we enable this... */
    gld->gl_ext[ARB_texture_compression] = 0;

    /* disable anistropic texture filters -- they cause a major slowdowns when used! 
     * (unless user explicitly turned them on in config file) */
    if( !use_anisotropic_texture_filtering ) {
      TRACE("disabling anisotropic filtering\n");
      gld->gl_ext[EXT_texture_filter_anisotropic] = 0; 
    }
      
    /* disable vertex programs/shaders if explicitly turned off in config file */
    if( !use_vertex_shaders ) {
      TRACE("disabling vertex shaders\n");
      gld->gl_ext[NV_vertex_program] = 0;
      gld->gl_ext[NV_vertex_program1_1] = 0;
      gld->gl_ext[NV_vertex_program2] = 0;
      gld->gl_ext[ARB_vertex_program] = 0;
    }

    dd->ddCaps.dwCaps2 |= DDCAPS2_CANMANAGETEXTURE | DDCAPS2_CANRENDERWINDOWED;
    dd->ddCaps.dwZBufferBitDepths |= DDBD_16;
/*
    dd->ddCaps.ddsCaps.dwCaps |= 0;
*/
    if (gld->gl_ext[ARB_texture_cube_map])
	dd->ddsCapsMore.dwCaps2 |= DDSCAPS2_CUBEMAP | DDSCAPS2_CUBEMAP_ALLFACES;

    lpData->hwCaps.dwFlags |= D3DDD_COLORMODEL | D3DDD_DEVCAPS | D3DDD_TRANSFORMCAPS | D3DDD_LIGHTINGCAPS |
			      D3DDD_BCLIPPING | D3DDD_LINECAPS | D3DDD_TRICAPS | D3DDD_DEVICERENDERBITDEPTH |
			      D3DDD_DEVICEZBUFFERBITDEPTH | D3DDD_MAXBUFFERSIZE | D3DDD_MAXVERTEXCOUNT;
    lpData->hwCaps.dcmColorModel = D3DCOLOR_RGB;
    lpData->hwCaps.dwDevCaps = D3DDEVCAPS_FLOATTLVERTEX | D3DDEVCAPS_EXECUTESYSTEMMEMORY |
			       D3DDEVCAPS_TLVERTEXSYSTEMMEMORY | D3DDEVCAPS_TLVERTEXVIDEOMEMORY |
			       D3DDEVCAPS_TEXTURESYSTEMMEMORY | D3DDEVCAPS_TEXTUREVIDEOMEMORY |
			       D3DDEVCAPS_DRAWPRIMTLVERTEX | D3DDEVCAPS_CANRENDERAFTERFLIP |
			       D3DDEVCAPS_TEXTURENONLOCALVIDMEM | D3DDEVCAPS_HWTRANSFORMANDLIGHT |
			       D3DDEVCAPS_CANBLTSYSTONONLOCAL | D3DDEVCAPS_HWRASTERIZATION;
    lpData->hwCaps.dtcTransformCaps.dwSize = sizeof(D3DTRANSFORMCAPS);
    lpData->hwCaps.dtcTransformCaps.dwCaps = D3DTRANSFORMCAPS_CLIP;
    lpData->hwCaps.bClipping = TRUE;
    lpData->hwCaps.dlcLightingCaps.dwSize = sizeof(D3DLIGHTINGCAPS);
    /* FIXME: we have GLSPOT, but we can't really do plain SPOT... should it be removed? */
    lpData->hwCaps.dlcLightingCaps.dwCaps = D3DLIGHTCAPS_POINT | D3DLIGHTCAPS_SPOT | D3DLIGHTCAPS_DIRECTIONAL |
					    D3DLIGHTCAPS_PARALLELPOINT | D3DLIGHTCAPS_GLSPOT;
    lpData->hwCaps.dlcLightingCaps.dwLightingModel = D3DLIGHTINGMODEL_RGB;
    lpData->hwCaps.dlcLightingCaps.dwNumLights = 1024; /* how much memory can we waste storing inactive lights? */
    fill_prim_caps(gld, &lpData->hwCaps.dpcLineCaps);
    fill_prim_caps(gld, &lpData->hwCaps.dpcTriCaps);
    /* FIXME: get real bit depth data */
    lpData->hwCaps.dwDeviceRenderBitDepth = DDBD_16;
    lpData->hwCaps.dwDeviceZBufferBitDepth = DDBD_16;
    /* FIXME: fix these numbers */
    lpData->hwCaps.dwMaxBufferSize = 0;
    lpData->hwCaps.dwMaxVertexCount = 65536;

    lpExt->dwSize = sizeof(D3DHAL_D3DEXTENDEDCAPS);

    tmpi = 1;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &tmpi);
    lpExt->dwMinTextureWidth  = 1;
    lpExt->dwMinTextureHeight = 1;
    lpExt->dwMaxTextureWidth  = tmpi;
    lpExt->dwMaxTextureHeight = tmpi;

    /* FIXME: fix these numbers */
    lpExt->dwMinStippleWidth  = 1;
    lpExt->dwMinStippleHeight = 1;
    lpExt->dwMaxStippleWidth  = 32;
    lpExt->dwMaxStippleHeight = 32;
    lpExt->dwMaxTextureRepeat = 256;
    lpExt->dwMaxTextureAspectRatio = 0;

    tmpi = 1;
    if (gld->gl_ext[EXT_texture_filter_anisotropic])
	glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &tmpi);
    lpExt->dwMaxAnisotropy = tmpi;

    /* FIXME: fix these numbers */
    lpExt->dvGuardBandLeft = -4096.0;
    lpExt->dvGuardBandTop = -4096.0;
    lpExt->dvGuardBandRight = 4095.0;
    lpExt->dvGuardBandBottom = 4095.0;
    lpExt->dvExtentsAdjust = 0;

    lpExt->dwStencilCaps = D3DSTENCILCAPS_KEEP | D3DSTENCILCAPS_ZERO | D3DSTENCILCAPS_REPLACE |
			   D3DSTENCILCAPS_INCRSAT | D3DSTENCILCAPS_DECRSAT | D3DSTENCILCAPS_INVERT;
    if (gld->gl_ext[EXT_stencil_wrap])
	lpExt->dwStencilCaps = D3DSTENCILCAPS_INCR | D3DSTENCILCAPS_DECR;

    lpExt->dwFVFCaps = 8;

    lpExt->dwTextureOpCaps = D3DTEXOPCAPS_DISABLE | D3DTEXOPCAPS_SELECTARG1 | D3DTEXOPCAPS_SELECTARG2 |
			     D3DTEXOPCAPS_MODULATE | D3DTEXOPCAPS_BLENDTEXTUREALPHA;
    if (gld->gl_ext[ARB_texture_env_add] || gld->gl_ext[EXT_texture_env_add])
	lpExt->dwTextureOpCaps |= D3DTEXOPCAPS_ADD;
    if (gld->gl_ext[ARB_texture_env_combine])
	lpExt->dwTextureOpCaps |= D3DTEXOPCAPS_MODULATE | D3DTEXOPCAPS_MODULATE2X | D3DTEXOPCAPS_MODULATE4X |
				  D3DTEXOPCAPS_ADD | D3DTEXOPCAPS_ADDSIGNED | D3DTEXOPCAPS_ADDSIGNED2X |
				  D3DTEXOPCAPS_SUBTRACT | D3DTEXOPCAPS_BLENDDIFFUSEALPHA |
				  D3DTEXOPCAPS_BLENDTEXTUREALPHA | D3DTEXOPCAPS_BLENDFACTORALPHA |
				  D3DTEXOPCAPS_BLENDCURRENTALPHA | D3DTEXOPCAPS_LERP;
    else if (gld->gl_ext[EXT_texture_env_combine])
	lpExt->dwTextureOpCaps |= D3DTEXOPCAPS_MODULATE | D3DTEXOPCAPS_MODULATE2X | D3DTEXOPCAPS_MODULATE4X |
				  D3DTEXOPCAPS_ADD | D3DTEXOPCAPS_ADDSIGNED | D3DTEXOPCAPS_ADDSIGNED2X |
				  D3DTEXOPCAPS_BLENDDIFFUSEALPHA | D3DTEXOPCAPS_BLENDTEXTUREALPHA |
				  D3DTEXOPCAPS_BLENDFACTORALPHA | D3DTEXOPCAPS_BLENDCURRENTALPHA |
				  D3DTEXOPCAPS_LERP;
#if 0 /* not implemented yet */
    if (gld->gl_ext[NV_texture_shader])
	lpExt->dwTextureOpCaps |= D3DTEXOPCAPS_BUMPENVMAP | D3DTEXOPCAPS_BUMPENVMAPLUMINANCE;
#endif
    if (gld->gl_ext[ARB_texture_env_dot3] || gld->gl_ext[EXT_texture_env_dot3])
	lpExt->dwTextureOpCaps |= D3DTEXOPCAPS_DOTPRODUCT3;

    tmpi = 1;
    if (gld->gl_ext[ARB_multitexture])
	glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &tmpi);
    if (tmpi > 8) tmpi = 8; /* D3D can't handle more */
    lpExt->wMaxTextureBlendStages = tmpi;
    lpExt->wMaxSimultaneousTextures = tmpi;
    TRACE("texture blend stages: %d/%d\n",
          lpExt->wMaxTextureBlendStages,
          lpExt->wMaxSimultaneousTextures);

    tmpi = 0;
    glGetIntegerv(GL_MAX_LIGHTS, &tmpi);
    if (tmpi > 32) tmpi = 32; /* gllight.c can't handle more */
    lpExt->dwMaxActiveLights = tmpi;

    lpExt->dvMaxVertexW = 1.0; /* XXX */

    tmpi = 0;
    glGetIntegerv(GL_MAX_CLIP_PLANES, &tmpi);
    if (tmpi > 32) tmpi = 32; /* D3DMAXUSERCLIPPLANES */
    lpExt->wMaxUserClipPlanes = tmpi;

    tmpi = 0;
    if (gld->gl_ext[ARB_vertex_blend])
	glGetIntegerv(GL_MAX_VERTEX_UNITS_ARB, &tmpi);
    else if (gld->gl_ext[EXT_vertex_weighting])
	tmpi = 2;
    if (tmpi > 4) tmpi = 4; /* D3D can't handle more */
    lpExt->wMaxVertexBlendMatrices = tmpi;

    lpExt->dwVertexProcessingCaps = D3DVTXPCAPS_TEXGEN | D3DVTXPCAPS_VERTEXFOG |
				    D3DVTXPCAPS_DIRECTIONALLIGHTS | D3DVTXPCAPS_POSITIONALLIGHTS |
				    D3DVTXPCAPS_LOCALVIEWER | D3DVTXPCAPS_MATERIALSOURCE7;

    lpExt->dwReserved1 = 0;
    lpExt->dwReserved2 = 0;
    lpExt->dwReserved3 = 0;
    lpExt->dwReserved4 = 0;

    add_texture_fmts(gld, lpData);
    add_zbuffer_fmts(gld, dd);

    /* DirectX 8 caps */
    gld->caps8.DeviceType = D3DDEVTYPE_HAL;
    gld->caps8.AdapterOrdinal = 0;
    gld->caps8.Caps = dd->ddCaps.dwCaps;
    gld->caps8.Caps2 = dd->ddCaps.dwCaps2;
    gld->caps8.Caps3 = 0;
    gld->caps8.DevCaps = lpData->hwCaps.dwDevCaps;
    /* FIXME: add caps for patches */
    gld->caps8.PrimitiveMiscCaps  = lpData->hwCaps.dpcTriCaps.dwMiscCaps;
    gld->caps8.RasterCaps         = lpData->hwCaps.dpcTriCaps.dwRasterCaps;
    gld->caps8.ZCmpCaps           = lpData->hwCaps.dpcTriCaps.dwZCmpCaps;
    gld->caps8.SrcBlendCaps       = lpData->hwCaps.dpcTriCaps.dwSrcBlendCaps;
    gld->caps8.DestBlendCaps      = lpData->hwCaps.dpcTriCaps.dwDestBlendCaps;
    gld->caps8.AlphaCmpCaps       = lpData->hwCaps.dpcTriCaps.dwAlphaCmpCaps;
    gld->caps8.ShadeCaps          = lpData->hwCaps.dpcTriCaps.dwShadeCaps;
    gld->caps8.TextureCaps        = lpData->hwCaps.dpcTriCaps.dwTextureCaps;
    gld->caps8.TextureFilterCaps  = lpData->hwCaps.dpcTriCaps.dwTextureFilterCaps;
    gld->caps8.TextureAddressCaps = lpData->hwCaps.dpcTriCaps.dwTextureAddressCaps;
    if (gld->gl_ext[ARB_texture_cube_map])
	gld->caps8.CubeTextureFilterCaps = gld->caps8.TextureFilterCaps;
    /* FIXME: VolumeTextureFilterCaps */
    gld->caps8.LineCaps = D3DLINECAPS_ALPHACMP | D3DLINECAPS_BLEND | D3DLINECAPS_FOG |
			  D3DLINECAPS_TEXTURE | D3DLINECAPS_ZTEST;
    gld->caps8.MaxTextureWidth       = lpExt->dwMaxTextureWidth;
    gld->caps8.MaxTextureHeight      = lpExt->dwMaxTextureHeight;
    /* FIXME: MaxVolumeExtent */
    gld->caps8.MaxTextureRepeat      = lpExt->dwMaxTextureRepeat;
    gld->caps8.MaxTextureAspectRatio = lpExt->dwMaxTextureAspectRatio;
    gld->caps8.MaxAnisotropy         = lpExt->dwMaxAnisotropy;
    gld->caps8.MaxVertexW            = lpExt->dvMaxVertexW;
    gld->caps8.GuardBandLeft   = lpExt->dvGuardBandLeft;
    gld->caps8.GuardBandTop    = lpExt->dvGuardBandTop;
    gld->caps8.GuardBandRight  = lpExt->dvGuardBandRight;
    gld->caps8.GuardBandBottom = lpExt->dvGuardBandBottom;
    gld->caps8.ExtentsAdjust = lpExt->dvExtentsAdjust;
    gld->caps8.StencilCaps   = lpExt->dwStencilCaps;
    gld->caps8.FVFCaps       = lpExt->dwFVFCaps; /* FIXME: PSIZE */
    gld->caps8.TextureOpCaps = lpExt->dwTextureOpCaps;
    gld->caps8.MaxTextureBlendStages   = lpExt->wMaxTextureBlendStages;
    gld->caps8.MaxSimultaneousTextures = lpExt->wMaxSimultaneousTextures;
    gld->caps8.VertexProcessingCaps = lpExt->dwVertexProcessingCaps;
    gld->caps8.MaxActiveLights   = lpExt->dwMaxActiveLights;
    gld->caps8.MaxUserClipPlanes = lpExt->wMaxUserClipPlanes;
    gld->caps8.MaxVertexBlendMatrices = lpExt->wMaxVertexBlendMatrices;
    gld->caps8.MaxVertexBlendMatrixIndex = 0; /* FIXME */
    gld->caps8.MaxPointSize = 1.0f;
    gld->caps8.MaxPrimitiveCount = 0xFFFF;
    gld->caps8.MaxVertexIndex = 0xFFFF;
    gld->caps8.MaxStreams = 16; 
    gld->caps8.MaxStreamStride = 0xFFFF;
    if( gld->gl_ext[NV_vertex_program] || gld->gl_ext[NV_vertex_program1_1] ) {
      /* add ARB_vertex_program when implemented */
      gld->caps8.VertexShaderVersion = D3DVS_VERSION(1,1);
      gld->caps8.MaxVertexShaderConst = 96;
    } else {
      gld->caps8.VertexShaderVersion = 0;
      gld->caps8.MaxVertexShaderConst = 0;
    }
    gld->caps8.PixelShaderVersion = 0;
    gld->caps8.MaxPixelShaderValue = 0;

    {
	D3DHAL_CONTEXTDESTROYDATA data;
	data.dwhContext = dwhContext;
	lpCB->ContextDestroy(&data);
    }
    TRACE("device capabilities completed\n");
}

static void fill_prim_caps(DD_GL_data *gld, LPD3DPRIMCAPS pc)
{
    pc->dwSize = sizeof(D3DPRIMCAPS);
    pc->dwMiscCaps = D3DPMISCCAPS_MASKZ | D3DPMISCCAPS_LINEPATTERNREP | D3DPMISCCAPS_CONFORMANT |
		     D3DPMISCCAPS_CULLNONE | D3DPMISCCAPS_CULLCW | D3DPMISCCAPS_CULLCCW |
		     D3DPMISCCAPS_COLORWRITEENABLE;
    if (gld->gl_ext[ARB_imaging])
	pc->dwMiscCaps |= D3DPMISCCAPS_BLENDOP;
    pc->dwRasterCaps = D3DPRASTERCAPS_DITHER | D3DPRASTERCAPS_ZTEST | D3DPRASTERCAPS_FOGVERTEX |
		       D3DPRASTERCAPS_FOGTABLE | D3DPRASTERCAPS_STIPPLE | D3DPRASTERCAPS_MIPMAPLODBIAS |
		       D3DPRASTERCAPS_ZBIAS | D3DPRASTERCAPS_FOGRANGE | D3DPRASTERCAPS_WFOG;
    if (gld->gl_ext[EXT_texture_filter_anisotropic])
	pc->dwRasterCaps |= D3DPRASTERCAPS_ANISOTROPY;
    pc->dwZCmpCaps = 0xFF; /* All Z tests can be done */
    pc->dwSrcBlendCaps = D3DPBLENDCAPS_ZERO | D3DPBLENDCAPS_ONE | D3DPBLENDCAPS_SRCALPHA |
			 D3DPBLENDCAPS_INVSRCALPHA | D3DPBLENDCAPS_DESTALPHA | D3DPBLENDCAPS_INVDESTALPHA |
			 D3DPBLENDCAPS_DESTCOLOR | D3DPBLENDCAPS_INVDESTCOLOR | D3DPBLENDCAPS_SRCALPHASAT;
    pc->dwDestBlendCaps = D3DPBLENDCAPS_ZERO | D3DPBLENDCAPS_ONE | D3DPBLENDCAPS_SRCCOLOR |
			  D3DPBLENDCAPS_INVSRCCOLOR | D3DPBLENDCAPS_SRCALPHA | D3DPBLENDCAPS_INVSRCALPHA |
			  D3DPBLENDCAPS_DESTALPHA | D3DPBLENDCAPS_INVDESTALPHA;
    if (gld->gl_ext[NV_blend_square]) {
	pc->dwSrcBlendCaps |= D3DPBLENDCAPS_SRCCOLOR | D3DPBLENDCAPS_INVSRCCOLOR;
	pc->dwDestBlendCaps |= D3DPBLENDCAPS_DESTCOLOR | D3DPBLENDCAPS_INVDESTCOLOR;
    }
    pc->dwAlphaCmpCaps  = 0xFF; /* All alpha tests can be done */
    pc->dwShadeCaps = D3DPSHADECAPS_COLORFLATRGB | D3DPSHADECAPS_COLORGOURAUDRGB |
		      D3DPSHADECAPS_SPECULARFLATRGB | D3DPSHADECAPS_SPECULARGOURAUDRGB |
		      D3DPSHADECAPS_ALPHAFLATBLEND | D3DPSHADECAPS_ALPHAGOURAUDBLEND |
		      D3DPSHADECAPS_FOGFLAT | D3DPSHADECAPS_FOGGOURAUD;
    pc->dwTextureCaps = D3DPTEXTURECAPS_PERSPECTIVE | D3DPTEXTURECAPS_POW2 | D3DPTEXTURECAPS_ALPHA |
			D3DPTEXTURECAPS_TRANSPARENCY | D3DPTEXTURECAPS_BORDER | D3DPTEXTURECAPS_MIPMAP;
    if (gld->gl_ext[ARB_texture_cube_map])
	pc->dwTextureCaps |= D3DPTEXTURECAPS_CUBEMAP | D3DPTEXTURECAPS_MIPCUBEMAP | D3DPTEXTURECAPS_CUBEMAP_POW2;
    pc->dwTextureFilterCaps = D3DPTFILTERCAPS_NEAREST | D3DPTFILTERCAPS_LINEAR |
			      D3DPTFILTERCAPS_MIPNEAREST | D3DPTFILTERCAPS_MIPLINEAR |
			      D3DPTFILTERCAPS_LINEARMIPNEAREST | D3DPTFILTERCAPS_LINEARMIPLINEAR |
			      D3DPTFILTERCAPS_MINFPOINT | D3DPTFILTERCAPS_MINFLINEAR |
			      D3DPTFILTERCAPS_MIPFPOINT | D3DPTFILTERCAPS_MIPFLINEAR |
			      D3DPTFILTERCAPS_MAGFPOINT | D3DPTFILTERCAPS_MAGFLINEAR;
    if (gld->gl_ext[EXT_texture_filter_anisotropic])
	pc->dwTextureFilterCaps |= D3DPTFILTERCAPS_MINFANISOTROPIC | D3DPTFILTERCAPS_MAGFANISOTROPIC;
    pc->dwTextureBlendCaps = D3DPTBLENDCAPS_DECAL | D3DPTBLENDCAPS_MODULATE |
			     D3DPTBLENDCAPS_DECALALPHA | D3DPTBLENDCAPS_MODULATEALPHA |
			     D3DPTBLENDCAPS_COPY;
    if (gld->gl_ext[ARB_texture_env_add] || gld->gl_ext[EXT_texture_env_add] ||
	gld->gl_ext[ARB_texture_env_combine] || gld->gl_ext[EXT_texture_env_combine])
	pc->dwTextureBlendCaps |= D3DPTBLENDCAPS_ADD;
    pc->dwTextureAddressCaps = D3DPTADDRESSCAPS_WRAP | D3DPTADDRESSCAPS_CLAMP | D3DPTADDRESSCAPS_INDEPENDENTUV;
    if (gld->gl_ext[ARB_texture_border_clamp])
	pc->dwTextureAddressCaps |= D3DPTADDRESSCAPS_BORDER;
    if (gld->gl_ext[ARB_texture_mirrored_repeat] || gld->gl_ext[IBM_texture_mirrored_repeat])
	pc->dwTextureAddressCaps |= D3DPTADDRESSCAPS_MIRROR;
    if (gld->gl_ext[ATI_texture_mirror_once])
	pc->dwTextureAddressCaps |= D3DPTADDRESSCAPS_MIRRORONCE;
    pc->dwStippleWidth = 32;
    pc->dwStippleHeight = 32;
}

#define DD_GL_EXT(x) "GL_"#x
static const char*const GL_ext_names[DD_GL_EXTENSION_COUNT] = {
    DD_GL_EXTENSION_LIST
};
#undef DD_GL_EXT

static void check_exts(DD_GL_data *gld)
{
    const char *exts = glGetString(GL_EXTENSIONS);
    int i;

    TRACE("checking available GL extensions\n");
    if (!exts) WARN("failed to acquire extension list\n");
    for (i=0; i<DD_GL_EXTENSION_COUNT; i++) {
	gld->gl_ext[i] = exts ? (strstr(exts, GL_ext_names[i]) != NULL) : FALSE;
	TRACE(" %s: %s\n", GL_ext_names[i], gld->gl_ext[i] ? "YES" : "NO");
    }
}

static void add_texture_fmts(DD_GL_data *gld, LPD3DHAL_GLOBALDRIVERDATA lpData)
{
    DDSURFACEDESC temp;
    LPDDPIXELFORMAT pfmt;
    DWORD count = 8 + 4;

    memset(&temp, 0, sizeof(DDSURFACEDESC));
    lpData->lpTextureFormats = HeapAlloc(GetProcessHeap(),0,sizeof(DDSURFACEDESC)*count);
    temp.dwSize = sizeof(DDSURFACEDESC);
    temp.dwFlags = DDSD_PIXELFORMAT | DDSD_CAPS;
    temp.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
    pfmt = &temp.ddpfPixelFormat;
    pfmt->dwSize = sizeof(DDPIXELFORMAT);
    pfmt->dwFourCC = 0;
    count = 0;

    /* X8R8G8B8 */
    pfmt->dwFlags = DDPF_RGB;
    pfmt->u1.dwRGBBitCount     = 32;
    pfmt->u2.dwRBitMask        = 0x00FF0000;
    pfmt->u3.dwGBitMask        = 0x0000FF00;
    pfmt->u4.dwBBitMask        = 0x000000FF;
    pfmt->u5.dwRGBAlphaBitMask = 0x00000000;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* A8R8G8B8 */
    pfmt->dwFlags = DDPF_RGB | DDPF_ALPHAPIXELS;
    pfmt->u1.dwRGBBitCount     = 32;
    pfmt->u2.dwRBitMask        = 0x00FF0000;
    pfmt->u3.dwGBitMask        = 0x0000FF00;
    pfmt->u4.dwBBitMask        = 0x000000FF;
    pfmt->u5.dwRGBAlphaBitMask = 0xFF000000;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* R5G6B5 */
    pfmt->dwFlags = DDPF_RGB;
    pfmt->u1.dwRGBBitCount     = 16;
    pfmt->u2.dwRBitMask        = 0xF800;
    pfmt->u3.dwGBitMask        = 0x07E0;
    pfmt->u4.dwBBitMask        = 0x001F;
    pfmt->u5.dwRGBAlphaBitMask = 0x0000;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* X1R5G5B5 */
    pfmt->dwFlags = DDPF_RGB;
    pfmt->u1.dwRGBBitCount     = 16;
    pfmt->u2.dwRBitMask        = 0x7C00;
    pfmt->u3.dwGBitMask        = 0x03E0;
    pfmt->u4.dwBBitMask        = 0x001F;
    pfmt->u5.dwRGBAlphaBitMask = 0x0000;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* A1R5G5B5 */
    pfmt->dwFlags = DDPF_RGB | DDPF_ALPHAPIXELS;
    pfmt->u1.dwRGBBitCount     = 16;
    pfmt->u2.dwRBitMask        = 0x7C00;
    pfmt->u3.dwGBitMask        = 0x03E0;
    pfmt->u4.dwBBitMask        = 0x001F;
    pfmt->u5.dwRGBAlphaBitMask = 0x8000;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* A4R4G4B4 */
    pfmt->dwFlags = DDPF_RGB | DDPF_ALPHAPIXELS;
    pfmt->u1.dwRGBBitCount     = 16;
    pfmt->u2.dwRBitMask        = 0x0F00;
    pfmt->u3.dwGBitMask        = 0x00F0;
    pfmt->u4.dwBBitMask        = 0x000F;
    pfmt->u5.dwRGBAlphaBitMask = 0xF000;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* L8 */
    pfmt->dwFlags = DDPF_LUMINANCE;
    pfmt->u1.dwLuminanceBitCount       = 8;
    pfmt->u2.dwLuminanceBitMask        = 0xFF;
    pfmt->u3.dwGBitMask                = 0x00;
    pfmt->u4.dwBBitMask                = 0x00;
    pfmt->u5.dwRGBAlphaBitMask         = 0x00;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* A8L8 */
    pfmt->dwFlags = DDPF_LUMINANCE | DDPF_ALPHAPIXELS;
    pfmt->u1.dwLuminanceBitCount       = 16;
    pfmt->u2.dwLuminanceBitMask        = 0x00FF;
    pfmt->u3.dwGBitMask                = 0x0000;
    pfmt->u4.dwBBitMask                = 0x0000;
    pfmt->u5.dwLuminanceAlphaBitMask   = 0xFF00;
    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

    /* FIXME: bumpmap format */

    if (gld->gl_ext[ARB_texture_compression]) {
	/* compressed texture formats */
	if (gld->gl_ext[EXT_texture_compression_s3tc]) {
	    pfmt->dwFlags = DDPF_RGB | DDPF_FOURCC | DDPF_COMPRESSED;
	    pfmt->dwFourCC = 0x31545844; /* DXT1 */
	    pfmt->u2.dwRBitMask        = 0x00F800;
	    pfmt->u3.dwGBitMask        = 0x0007E0;
	    pfmt->u4.dwBBitMask        = 0x00001F;
	    pfmt->u5.dwRGBAlphaBitMask = 0x000000;
	    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

	    pfmt->dwFlags = DDPF_RGB | DDPF_ALPHAPIXELS | DDPF_FOURCC | DDPF_COMPRESSED;
	    pfmt->dwFourCC = 0x31545844; /* DXT1 */
	    pfmt->u5.dwRGBAlphaBitMask = 0x010000;
	    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

	    pfmt->dwFourCC = 0x33545844; /* DXT3 */
	    pfmt->u5.dwRGBAlphaBitMask = 0x0F0000;
	    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));

	    pfmt->dwFourCC = 0x35545844; /* DXT5 */
	    pfmt->u5.dwRGBAlphaBitMask = 0xFF0000;
	    memcpy(&lpData->lpTextureFormats[count++], &temp, sizeof(DDSURFACEDESC));
	}
    }

    lpData->dwNumTextureFormats = count;
}

static void add_zbuffer_fmts(DD_GL_data *gld, LPDDRAWI_DIRECTDRAW_GBL lpDD)
{
    DDPIXELFORMAT temp;
    DWORD count = 2;
    memset(&temp, 0, sizeof(DDPIXELFORMAT));
    lpDD->lpZPixelFormats = HeapAlloc(GetProcessHeap(),0,sizeof(DDPIXELFORMAT)*count);
    temp.dwSize = sizeof(DDPIXELFORMAT);
    temp.dwFlags = DDPF_ZBUFFER;
    temp.dwFourCC = 0;
    count = 0;

    temp.u1.dwZBufferBitDepth = 16;
    temp.u2.dwStencilBitDepth = 0;
    temp.u3.dwZBitMask = 0xffff;
    temp.u4.dwStencilBitMask = 0;
    temp.u5.dwRGBZBitMask = 0; /* What does this field mean in this case? */
    memcpy(&lpDD->lpZPixelFormats[count++], &temp, sizeof(DDPIXELFORMAT));

    /* This is pretty silly. Can we safely enumerate a 24/8 format here? */
    temp.dwFlags |= DDPF_STENCILBUFFER;
    temp.u2.dwStencilBitDepth = 8;
    temp.u3.dwZBitMask = 0x00ff;
    temp.u4.dwStencilBitMask = 0xff00;
    memcpy(&lpDD->lpZPixelFormats[count++], &temp, sizeof(DDPIXELFORMAT));

    lpDD->dwNumZPixelFormats = count;
}

LPVOID WINAPI D3DGL_GetCaps(LPDDRAWI_DIRECTDRAW_GBL dd, DWORD d3d_version)
{
    DD_GL_data *gld = (DD_GL_data *)dd->dwReserved3;
    if (d3d_version == 8)
	return &gld->caps8;
    return NULL;
}
