/*	Direct3D OpenGL driver
 *
 * Copyright 2001 TransGaming Technologies Inc.
 */

/* This is not a self-contained driver - in the same way that OpenGL requires
 * an OS-specific interface like GLX or WGL to operate, this GL driver is
 * supposed to be layered onto a platform-specific Direct3D driver
 */

#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "winbase.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"
#include "debugtools.h"

#undef STATEBLOCK_DISPLAYLIST /* this may make lighting unreliable, don't define */

DEFAULT_DEBUG_CHANNEL(ddraw);

static HRESULT GL_replay_cmds(GL_D3D_priv* priv, struct GL_D3D_saved_cmd **cmd, DWORD dwFlags);
#define REPLAY_DEL	1
#define REPLAY_EXEC	2
#define REPLAY_APPLY	4
#define REPLAY_CAPTURE	8

static BOOL DDHAL_GetDriverInfo(LPDDHAL_GETDRIVERINFO dd_get_driver_info,
				const GUID* guidInfo,
				LPVOID lpData,
				DWORD dwExpectedSize)
{
    DDHAL_GETDRIVERINFODATA data;

    if (!dd_get_driver_info) return FALSE;

    data.dwSize = sizeof(DDHAL_GETDRIVERINFODATA);
    data.dwFlags = 0;
    data.guidInfo = *guidInfo;
    data.dwExpectedSize = dwExpectedSize;
    data.lpvData = lpData;
    data.dwActualSize = 0;
    data.ddRVal = DD_OK;
    data.dwContext = 0; /* XXX */

    dd_get_driver_info(&data);
    if (FAILED(data.ddRVal)) return FALSE;

    *(LPDWORD)lpData = data.dwActualSize;
    return TRUE;
}

GL_D3D_priv* WINAPI D3DGL_Construct(LPD3D_P d3dp, DWORD d3d_version)
{
    LPDDRAWI_DIRECTDRAW_LCL dd_lcl = d3dp->lcl;
    LPDDRAWI_DIRECTDRAW_GBL dd_gbl = dd_lcl->lpGbl;
    GL_D3D_priv* priv;

    TRACE("(%p)\n", d3dp);

    priv = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(GL_D3D_priv));
    priv->d3dp = d3dp;
    priv->dd_lcl = dd_lcl;
    priv->dd_gbl = dd_gbl;
    priv->gl_data = (DD_GL_data*)dd_gbl->dwReserved3;
    priv->lpCaps = (LPD3DHAL_D3DEXTENDEDCAPS)dd_gbl->lpD3DExtendedCaps;
    priv->glFlags = GLF_NEW_TRANSFORM | GLF_NEW_LIGHTING;
    priv->material_flags = GLMF_ALL_MATERIAL_DIRTY;
    priv->texture_flags = GLTF_DEFAULT_TEXCOORD_TRANSFORM;
    priv->d3d_version = d3d_version;

    priv->color_array_size = 8192;
    priv->color_array = VirtualAlloc(NULL, priv->color_array_size,
				     MEM_COMMIT | MEM_RESERVE,
				     PAGE_READWRITE);
    priv->vertex_array_size = 8192;
    priv->vertex_array = VirtualAlloc(NULL, priv->vertex_array_size,
				      MEM_COMMIT | MEM_RESERVE,
				      PAGE_READWRITE);

    DDHAL_GetDriverInfo(d3dp->gdi, &GUID_WineGLD3DCallbacks,
			&priv->callbacks, sizeof(priv->callbacks));

    /* record "init" command */
    priv->cmds = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(struct GL_D3D_saved_cmd));
    priv->cmds->cmd.bCommand = 0;

    return priv;
}

void WINAPI D3DGL_Destruct(GL_D3D_priv* priv)
{
    VirtualFree(priv->color_array, 0, MEM_RELEASE);
    VirtualFree(priv->vertex_array, 0, MEM_RELEASE);
    while (priv->cmds) {
	struct GL_D3D_saved_cmd *ncmd = priv->cmds;
	priv->cmds = ncmd->next;
	if (ncmd->data1) HeapFree(GetProcessHeap(),0,ncmd->data1);
	if (ncmd->data2) HeapFree(GetProcessHeap(),0,ncmd->data2);
	HeapFree(GetProcessHeap(),0,ncmd);
    }
}

static const float zero4f[4] = { 0, 0, 0, 0 };
static const float one4f[4] = { 1, 1, 1, 1 };
static const DWORD dummy_texture_pixels;

/* This routine should only initialize anything non-render-state-specific
 * (the default render states are set separately) */
HRESULT D3D_GL_init(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    static const GLfloat sgenparams[] = {1.0, 0.0, 0.0, 0.0};
    static const GLfloat tgenparams[] = {0.0, 1.0, 0.0, 0.0};
    static const GLfloat rgenparams[] = {0.0, 0.0, 1.0, 0.0};
    unsigned u;

    TRACE("initializing GL context\n");

#ifndef HAVE_OSMESA
    glFrontFace(GL_CW);
#else
    /* OSMESA y-swap */
    glFrontFace(GL_CCW);
#endif

    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, one4f);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, zero4f);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, zero4f);
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, zero4f);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, zero4f);
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
    priv->current_color_material = GL_AMBIENT_AND_DIFFUSE;

    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

    glDisable(GL_DITHER);
    glEnable(GL_SCISSOR_TEST);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    for (u=0; u < priv->lpCaps->wMaxTextureBlendStages; u++) {
	GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	glTexGenfv(GL_S, GL_EYE_PLANE, sgenparams);
	glTexGenfv(GL_T, GL_EYE_PLANE, tgenparams);
	glTexGenfv(GL_R, GL_EYE_PLANE, rgenparams);
    }

    glPopMatrix();

    /* if texture_env_combine is available, enable it */
    if (HAVE_GL_EXT(priv, ARB_texture_env_combine) || HAVE_GL_EXT(priv, EXT_texture_env_combine)) {
	for (u=0; u<priv->lpCaps->wMaxTextureBlendStages; u++) {
	    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
	}
    }

    /* create a dummy texture for stages without a real texture */
    glGenTextures(1, &priv->dummy_texture);
    glBindTexture(GL_TEXTURE_2D, priv->dummy_texture);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
		 1, 1,
		 0, GL_RGB, GL_UNSIGNED_BYTE,
		 &dummy_texture_pixels);
    priv->texture_target[0] = GL_TEXTURE_2D;
    for (u=1; u<priv->lpCaps->wMaxTextureBlendStages; u++) {
	GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	glBindTexture(GL_TEXTURE_2D, priv->dummy_texture);
	priv->texture_target[u] = GL_TEXTURE_2D;
    }

    if (HAVE_GL_EXT(priv, ARB_vertex_blend))
	glEnable(GL_WEIGHT_SUM_UNITY_ARB);

    glDisable(GL_LIGHTING);
    priv->isLightingEnabled = FALSE;
    priv->isVertexProgramEnabled = FALSE;

    priv->pVertexArray = NULL;
    priv->pNormalArray = NULL;
    priv->pColorArray = NULL;
    priv->pSecondaryColorArray = NULL;
    for (u=0; u < 8; u++) 
	priv->pTexCoordArray[u] = NULL;

    return D3D_OK;
}

HRESULT WINAPI D3DGL_AllocStateBlock(GL_D3D_priv* priv, LPDWORD lpdwBlock)
{
    struct GL_D3D_state_block *block = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
						 sizeof(struct GL_D3D_state_block));
    TRACE("creating state block %p\n", block);
    /* won't bother to allocate a handle yet, just cast the pointer... */
    *lpdwBlock = (DWORD)block;
    return D3D_OK;
}

HRESULT WINAPI D3DGL_FreeStateBlock(GL_D3D_priv* priv, DWORD dwBlock)
{
    /* don't do anything here, let D3DHAL_STATESETDELETE take care of it */
    return D3D_OK;
}

static BOOL GL_update_cmd(struct GL_D3D_saved_cmd *ocmd,
			  LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    if (ocmd->cmd.bCommand != cmd->bCommand) return FALSE;
    switch (ocmd->cmd.bCommand) {
    case D3DDP2OP_RENDERSTATE:
	{
	    LPD3DHAL_DP2RENDERSTATE stat1 = ocmd->data1;
	    LPD3DHAL_DP2RENDERSTATE stat2 = data1;
	    if (stat1->RenderState == stat2->RenderState) {
		stat1->u.dwState = stat2->u.dwState;
		return TRUE;
	    }
	}
	break;
    case D3DDP2OP_TEXTURESTAGESTATE:
	{
	    LPD3DHAL_DP2TEXTURESTAGESTATE stat1 = ocmd->data1;
	    LPD3DHAL_DP2TEXTURESTAGESTATE stat2 = data1;
	    if (stat1->wStage != stat2->wStage) break;
	    if (stat1->TSState == stat2->TSState) {
		stat1->dwValue = stat2->dwValue;
		return TRUE;
	    }
	}
	break;
    case D3DDP2OP_VIEWPORTINFO:
	{
	    LPD3DHAL_DP2VIEWPORTINFO stat1 = ocmd->data1;
	    LPD3DHAL_DP2VIEWPORTINFO stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    case D3DDP2OP_WINFO:
	{
	    LPD3DHAL_DP2WINFO stat1 = ocmd->data1;
	    LPD3DHAL_DP2WINFO stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    /* FIXME: SETPALETTE, UPDATEPALETTE */
    case D3DDP2OP_ZRANGE:
	{
	    LPD3DHAL_DP2ZRANGE stat1 = ocmd->data1;
	    LPD3DHAL_DP2ZRANGE stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    case D3DDP2OP_SETMATERIAL:
	{
	    LPD3DHAL_DP2SETMATERIAL stat1 = ocmd->data1;
	    LPD3DHAL_DP2SETMATERIAL stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    case D3DDP2OP_SETLIGHT:
	{
	    LPD3DHAL_DP2SETLIGHT stat1 = ocmd->data1;
	    LPD3DHAL_DP2SETLIGHT stat2 = data1;
	    if (stat1->dwIndex != stat2->dwIndex) return FALSE;
	    if (stat1->dwDataType < D3DHAL_SETLIGHT_DATA &&
		stat2->dwDataType < D3DHAL_SETLIGHT_DATA) {
		stat1->dwDataType = stat2->dwDataType;
		return TRUE;
	    }
	    if (stat1->dwDataType == D3DHAL_SETLIGHT_DATA &&
		stat2->dwDataType == D3DHAL_SETLIGHT_DATA) {
		LPD3DLIGHT7 l1 = ocmd->data2;
		LPD3DLIGHT7 l2 = data2;
		*l1 = *l2;
		return TRUE;
	    }
	}
	break;
    case D3DDP2OP_CREATELIGHT:
	{
	    LPD3DHAL_DP2CREATELIGHT stat1 = ocmd->data1;
	    LPD3DHAL_DP2CREATELIGHT stat2 = data1;
	    if (stat1->dwIndex < stat2->dwIndex)
		stat1->dwIndex = stat2->dwIndex;
	    return TRUE;
	}
	break;
    case D3DDP2OP_SETTRANSFORM:
	{
	    LPD3DHAL_DP2SETTRANSFORM stat1 = ocmd->data1;
	    LPD3DHAL_DP2SETTRANSFORM stat2 = data1;
	    if (stat1->xfrmType == stat2->xfrmType) {
		stat1->matrix = stat2->matrix;
		return TRUE;
	    }
	}
	break;
    /* FIXME: SETPRIORITY, SETTEXLOD, SETCLIPPLANE */
    default:
	break;
    }
    return FALSE;
}

static void GL_capture_cmd(GL_D3D_priv* priv,
			   struct GL_D3D_saved_cmd *ocmd)
{
    LPD3D_P d3dp = priv->d3dp;
    switch (ocmd->cmd.bCommand) {
    case D3DDP2OP_RENDERSTATE:
	{
	    LPD3DHAL_DP2RENDERSTATE stat = ocmd->data1;
	    stat->u.dwState = d3dp->renderstate[stat->RenderState];
	}
	break;
    case D3DDP2OP_TEXTURESTAGESTATE:
	{
	    LPD3DHAL_DP2TEXTURESTAGESTATE stat = ocmd->data1;
	    stat->dwValue = d3dp->texstagestate[stat->wStage][stat->TSState];
	}
	break;
    case D3DDP2OP_VIEWPORTINFO:
	{
	    LPD3DHAL_DP2VIEWPORTINFO stat = ocmd->data1;
	    stat->dwX = d3dp->viewport.dwX;
	    stat->dwY = d3dp->viewport.dwY;
	    stat->dwWidth = d3dp->viewport.dwWidth;
	    stat->dwHeight = d3dp->viewport.dwHeight;
	}
	break;
    case D3DDP2OP_WINFO:
	/* ? */
	break;
    case D3DDP2OP_ZRANGE:
	{
	    LPD3DHAL_DP2ZRANGE stat = ocmd->data1;
	    stat->dvMinZ = d3dp->viewport.dvMinZ;
	    stat->dvMaxZ = d3dp->viewport.dvMaxZ;
	}
	break;
    case D3DDP2OP_SETMATERIAL:
	{
	    LPD3DHAL_DP2SETMATERIAL stat = ocmd->data1;
	    *stat = d3dp->material;
	}
	break;
    case D3DDP2OP_SETLIGHT:
	{
	    LPD3DHAL_DP2SETLIGHT stat = ocmd->data1;
	    switch (stat->dwDataType) {
	    case D3DHAL_SETLIGHT_ENABLE:
	    case D3DHAL_SETLIGHT_DISABLE:
		if (d3dp->light[stat->dwIndex].flags & D3DLIGHT_ACTIVE)
		    stat->dwDataType = D3DHAL_SETLIGHT_ENABLE;
		else
		    stat->dwDataType = D3DHAL_SETLIGHT_DISABLE;
		break;
	    case D3DHAL_SETLIGHT_DATA:
		*(LPD3DLIGHT7)ocmd->data2 = d3dp->light[stat->dwIndex].l;
		break;
	    }
	}
	break;
    case D3DDP2OP_SETTRANSFORM:
	{
	    LPD3DHAL_DP2SETTRANSFORM stat = ocmd->data1;
	    stat->matrix = d3dp->xfrm[stat->xfrmType];
	}
	break;
    /* FIXME: SETCLIPPLANE */
    }
}

static void GL_apply_cmd(GL_D3D_priv* priv,
			 struct GL_D3D_saved_cmd *ocmd)
{
    LPD3D_P d3dp = priv->d3dp;
    switch (ocmd->cmd.bCommand) {
    case D3DDP2OP_RENDERSTATE:
	{
	    LPD3DHAL_DP2RENDERSTATE stat = ocmd->data1;
	    d3dp->renderstate[stat->RenderState] = stat->u.dwState;
	}
	break;
    case D3DDP2OP_TEXTURESTAGESTATE:
	{
	    LPD3DHAL_DP2TEXTURESTAGESTATE stat = ocmd->data1;
	    d3dp->texstagestate[stat->wStage][stat->TSState] = stat->dwValue;
	}
	break;
    case D3DDP2OP_VIEWPORTINFO:
	{
	    LPD3DHAL_DP2VIEWPORTINFO stat = ocmd->data1;
	    d3dp->viewport.dwX = stat->dwX;
	    d3dp->viewport.dwY = stat->dwY;
	    d3dp->viewport.dwWidth = stat->dwWidth;
	    d3dp->viewport.dwHeight = stat->dwHeight;
	}
	break;
    case D3DDP2OP_WINFO:
	/* ? */
	break;
    case D3DDP2OP_ZRANGE:
	{
	    LPD3DHAL_DP2ZRANGE stat = ocmd->data1;
	    d3dp->viewport.dvMinZ = stat->dvMinZ;
	    d3dp->viewport.dvMaxZ = stat->dvMaxZ;
	}
	break;
    case D3DDP2OP_SETMATERIAL:
	{
	    LPD3DHAL_DP2SETMATERIAL stat = ocmd->data1;
	    d3dp->material = *stat;
	}
	break;
    case D3DDP2OP_SETLIGHT:
	{
	    LPD3DHAL_DP2SETLIGHT stat = ocmd->data1;
	    switch (stat->dwDataType) {
	    case D3DHAL_SETLIGHT_ENABLE:
		d3dp->light[stat->dwIndex].flags |= D3DLIGHT_ACTIVE;
		break;
	    case D3DHAL_SETLIGHT_DISABLE:
		d3dp->light[stat->dwIndex].flags &= ~D3DLIGHT_ACTIVE;
		break;
	    case D3DHAL_SETLIGHT_DATA:
		d3dp->light[stat->dwIndex].l = *(LPD3DLIGHT7)ocmd->data2;
		break;
	    }
	}
	break;
    case D3DDP2OP_SETTRANSFORM:
	{
	    LPD3DHAL_DP2SETTRANSFORM stat = ocmd->data1;
	    d3dp->xfrm[stat->xfrmType] = stat->matrix;
	}
	break;
    /* FIXME: SETCLIPPLANE */
    }
}

static struct GL_D3D_saved_cmd*
GL_duplicate_cmd(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd,
		 LPVOID data1, DWORD dwSize1, LPVOID data2, DWORD dwSize2)
{
    struct GL_D3D_saved_cmd *ncmd;

    ncmd = HeapAlloc(GetProcessHeap(),0,sizeof(struct GL_D3D_saved_cmd));
    if (!ncmd) return ncmd;
    ncmd->next = NULL;
    ncmd->cmd = *cmd;
    ncmd->data1 = data1 ? HeapAlloc(GetProcessHeap(),0,dwSize1) : NULL;
    if (ncmd->data1) memcpy(ncmd->data1, data1, dwSize1);
    ncmd->data2 = data2 ? HeapAlloc(GetProcessHeap(),0,dwSize2) : NULL;
    if (ncmd->data2) memcpy(ncmd->data2, data2, dwSize2);
    return ncmd;
}

static void
GL_record_cmd(GL_D3D_priv* priv, struct GL_D3D_saved_cmd **cmds,
	      LPD3DHAL_DP2COMMAND cmd, LPVOID data1, DWORD dwSize1, LPVOID data2, DWORD dwSize2)
{
    struct GL_D3D_saved_cmd *ncmd = *cmds, *lcmd = NULL;

    while (ncmd) {
	if (GL_update_cmd(ncmd, cmd, data1, data2))
	    return;
	lcmd = ncmd;
	ncmd = ncmd->next;
    }
    ncmd = GL_duplicate_cmd(priv, cmd, data1, dwSize1, data2, dwSize2);
    if (lcmd) lcmd->next = ncmd;
    else *cmds = ncmd;
}

static void GL_delete_cmd(GL_D3D_priv* priv, struct GL_D3D_saved_cmd **cmd)
{
    struct GL_D3D_saved_cmd *ncmd = *cmd;
    *cmd = ncmd->next;
    if (ncmd->data1) HeapFree(GetProcessHeap(),0,ncmd->data1);
    if (ncmd->data2) HeapFree(GetProcessHeap(),0,ncmd->data2);
    HeapFree(GetProcessHeap(),0,ncmd);
}

void WINAPI D3DGL_DemolishTexture(GL_D3D_priv* priv, DWORD dwTexture)
{
    struct GL_D3D_saved_cmd **cmd = &priv->cmds;

    while (*cmd) {
	BOOL kill_it = FALSE;
	switch ((*cmd)->cmd.bCommand) {
	case D3DDP2OP_TEXBLT:
	    {
		LPD3DHAL_DP2TEXBLT dat = (*cmd)->data1;
		if (dat->dwDDDestSurface == dwTexture ||
		    dat->dwDDSrcSurface == dwTexture) kill_it = TRUE;
	    }
	    break;
	default:
	    break;
	}
	if (kill_it)
	    GL_delete_cmd(priv, cmd);
	else
	    cmd = &(*cmd)->next;
    }
}

static HRESULT D3D_GL_async_state_set(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2STATESET dat = (LPD3DHAL_DP2STATESET)data1;
    struct GL_D3D_state_block *block = (struct GL_D3D_state_block *)dat->dwParam;

    switch (dat->dwOperation) {
    case D3DHAL_STATESETBEGIN:
	TRACE("recording to state block %p\n", block);
	priv->block = block;
	break;
    case D3DHAL_STATESETEND:
	TRACE("state block %p complete\n", block);
	priv->block = NULL;
	break;
    default:
	break;
    }
    return D3D_OK;
}

static HRESULT D3D_GL_sync_state_set(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2STATESET dat = (LPD3DHAL_DP2STATESET)data1;
    struct GL_D3D_state_block *block = (struct GL_D3D_state_block *)dat->dwParam;

    switch (dat->dwOperation) {
    case D3DHAL_STATESETBEGIN:
    case D3DHAL_STATESETEND:
	/* handled in D3D_GL_async_state_set */
	break;
    case D3DHAL_STATESETDELETE:
	TRACE("deleting state block %p\n", block);
#ifdef STATEBLOCK_DISPLAYLIST
	if (block->list) {
	    glDeleteLists(block->list, 1);
	    block->list = 0;
	}
#endif
	while (block->cmds) GL_delete_cmd(priv, &block->cmds);
	HeapFree(GetProcessHeap(), 0, block);
	CHECK_GL_ERROR("DeleteStateBlock");
	break;
    case D3DHAL_STATESETEXECUTE:
	TRACE("executing state block %p\n", block);
#ifdef STATEBLOCK_DISPLAYLIST
	if (!block->list) {
	    block->list = glGenLists(1);
	    glNewList(block->list, GL_COMPILE_AND_EXECUTE);
	    /* FIXME: lights might not work reliably this way? */
#endif
	    GL_replay_cmds(priv, &block->cmds, REPLAY_EXEC|REPLAY_APPLY);
#ifdef STATEBLOCK_DISPLAYLIST
	    glEndList();
	} else {
	    glCallList(block->list);
	    GL_replay_cmds(priv, &block->cmds, REPLAY_APPLY);
	}
#endif
	CHECK_GL_ERROR("ApplyStateBlock");
	break;
    case D3DHAL_STATESETCAPTURE:
	TRACE("capturing state block %p\n", block);
#ifdef STATEBLOCK_DISPLAYLIST
	if (block->list) {
	    glDeleteLists(block->list, 1);
	    block->list = 0;
	}
#endif
	GL_replay_cmds(priv, &block->cmds, REPLAY_CAPTURE);
	break;
    }
    return D3D_OK;
}

typedef HRESULT (*GL_dp2_handler)(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2);

#define PRIM 1
#define STAT 2
#define OTHR 3

static const struct {
    GL_dp2_handler s_func, a_func;
    int type;
    const char*name;
} GL_dp2_func[] = {
    {D3D_GL_init,		NULL,			  OTHR, "INIT"},
    {D3D_GL_draw,		NULL,			  PRIM, "POINTS"},
    {NULL,			NULL,			  PRIM, "INDEXEDLINELIST"},
    {NULL,			NULL,			  PRIM, "INDEXEDTRIANGLELIST"},
    {NULL}, {NULL}, {NULL}, {NULL},
    {D3D_GL_sync_render_state, 	D3D_GL_async_render_state,STAT, "RENDERSTATE"},
    {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL},
    {D3D_GL_draw,		NULL,			  PRIM, "LINELIST"},
    {D3D_GL_draw,		NULL,			  PRIM, "LINESTRIP"},
    {D3D_GL_draw,		NULL,			  PRIM, "INDEXEDLINESTRIP"},
    {D3D_GL_draw,		NULL,			  PRIM, "TRIANGLELIST"},
    {D3D_GL_draw,		NULL,			  PRIM, "TRIANGLESTRIP"},
    {D3D_GL_draw,		NULL,			  PRIM, "INDEXEDTRIANGLESTRIP"},
    {D3D_GL_draw,		NULL,			  PRIM, "TRIANGLEFAN"},
    {D3D_GL_draw,		NULL,			  PRIM, "INDEXEDTRIANGLEFAN"},
    {NULL,			NULL,			  PRIM, "TRIANGLEFAN_IMM"},
    {NULL,			NULL,			  PRIM, "LINELIST_IMM"},
    {D3D_GL_texture_stage_state,NULL,			  STAT, "TEXTURESTAGESTATE"},
    {D3D_GL_draw,		NULL,			  PRIM, "INDEXEDTRIANGLELIST2"},
    {D3D_GL_draw,		NULL,			  PRIM, "INDEXEDLINELIST2"},
    {D3D_GL_viewport_info,	NULL,			  STAT, "VIEWPORTINFO"},
    {NULL,			NULL,			  STAT, "WINFO"},
    {D3D_GL_set_palette,	NULL,			  OTHR, "SETPALETTE"},
    {D3D_GL_update_palette,	NULL,			  OTHR, "UPDATEPALETTE"},
    {D3D_GL_z_range,		NULL,			  STAT, "ZRANGE"},
    {D3D_GL_set_material,	NULL,			  STAT, "SETMATERIAL"},
    {D3D_GL_set_light,		NULL,			  STAT, "SETLIGHT"},
    {D3D_GL_create_light,	NULL,			  OTHR, "CREATELIGHT"},
    {D3D_GL_set_transform,	NULL,			  STAT, "SETTRANSFORM"},
    {NULL,			NULL,			  OTHR, "EXT"},
    {D3D_GL_texblt,		NULL,			  OTHR, "TEXBLT"},
    {D3D_GL_sync_state_set,	D3D_GL_async_state_set,	  OTHR, "STATESET"},
    {NULL,			NULL,			  OTHR, "SETPRIORITY"},
    {NULL,			NULL,			  OTHR, "SETRENDERTARGET"},
    {D3D_GL_clear,		NULL,			  OTHR, "CLEAR"},
    {NULL,			NULL,			  OTHR, "SETTEXLOD"},
    {D3D_GL_set_clip_plane,	NULL,			  STAT, "SETCLIPPLANE"},
    {D3D_GL_create_vertex_shader,NULL,			  OTHR, "CREATEVERTEXSHADER"},
    {D3D_GL_delete_vertex_shader,NULL,			  OTHR, "DELETEVERTEXSHADER"},
    {D3D_GL_set_vertex_shader,	NULL,			  STAT, "SETVERTEXSHADER"},
    {D3D_GL_set_vertex_shader_const,NULL,		  STAT, "SETVERTEXSHADERCONST"},
    {NULL,			NULL,			  OTHR, "SETSTREAMSOURCE"},
    {NULL,			NULL,			  OTHR, "SETSTREAMSOURCEUM"},
    {NULL,			NULL,			  OTHR, "SETINDICES"},
    {NULL,			NULL,			  PRIM, "DRAWPRIMITIVE"},
    {NULL,			NULL,			  PRIM, "DRAWINDEXEDPRIMITIVE"},
    {NULL,			NULL,			  OTHR, "CREATEPIXELSHADER"},
    {NULL,			NULL,			  OTHR, "DELETEPIXELSHADER"},
    {NULL,			NULL,			  STAT, "SETPIXELSHADER"},
    {NULL,			NULL,			  STAT, "SETPIXELSHADERCONST"},
    {NULL,			NULL,			  PRIM, "CLIPPEDTRIANGLEFAN"},
    {NULL,			NULL,			  PRIM, "DRAWPRIMITIVE2"},
    {NULL,			NULL,			  PRIM, "DRAWINDEXEDPRIMITIVE2"},
    {NULL,			NULL,			  PRIM, "DRAWRECTPATCH"},
    {NULL,			NULL,			  PRIM, "DRAWTRIPATCH"},
    {NULL,			NULL,			  OTHR, "VOLUMEBLT"},
    {NULL,			NULL,			  OTHR, "BUFFERBLT"},
    {NULL,			NULL,			  OTHR, "MULTIPLYTRANSFORM"},
    {NULL,			NULL,			  OTHR, "ADDDIRTYRECT"},
    {NULL,			NULL,			  OTHR, "ADDDIRTYBOX"}
};

HRESULT WINAPI D3DGL_StartDP2SD(GL_D3D_priv* priv, DWORD dwFlags,
				DWORD dwVertexType, LPVOID lpVertices,
				DWORD dwVertexOffset, DWORD dwVertexLength,
				DWORD dwReqVertexBufSize, DWORD dwReqCommandBufSize,
				LPD3DDRAWPRIMITIVESTRIDEDDATA lpStridedData)
{
    priv->glFlags &= ~(GLF_STRIDED_DATA|GLF_VERTEX_BUFFER);
    priv->dp2_VertexBuffer = NULL;
    priv->dp2_lpVertices = NULL;

    if (lpStridedData)
    {
	priv->sd = *lpStridedData;
	priv->glFlags |= GLF_STRIDED_DATA;
    }
    else if (!(dwFlags & D3DHALDP2_USERMEMVERTICES))
    {
	priv->dp2_VertexBuffer = lpVertices;
	priv->glFlags |= GLF_VERTEX_BUFFER;
    }
    else
    {
	priv->dp2_lpVertices = lpVertices;
    }

    if (GL_CHECK_NEW_TNL(dwVertexType, priv->dp2_dwVertexType))
	priv->glFlags |= GLF_NEW_TNL_MODE;

    priv->dp2_dwVertexType = dwVertexType;
    priv->dp2_dwVertexOffset = dwVertexOffset;
    priv->dp2_dwVertexLength = dwVertexLength;
    priv->dp2_dwFlags = dwFlags;

    return D3D_OK;
}

HRESULT WINAPI D3DGL_StartDP2(GL_D3D_priv* priv, DWORD dwFlags,
			      DWORD dwVertexType, LPVOID lpVertices,
			      DWORD dwVertexOffset, DWORD dwVertexLength,
			      DWORD dwReqVertexBufSize, DWORD dwReqCommandBufSize)
{
    return D3DGL_StartDP2SD(priv, dwFlags, dwVertexType, lpVertices,
			    dwVertexOffset, dwVertexLength,
			    dwReqVertexBufSize, dwReqCommandBufSize, NULL);
}

HRESULT WINAPI D3DGL_EmitDP2(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd,
			     LPVOID data1, DWORD dwSize1, LPVOID data2, DWORD dwSize2)
{
    HRESULT hr = D3D_OK;
    GL_dp2_handler s_func, a_func;

    s_func = GL_dp2_func[cmd->bCommand].s_func;
    a_func = GL_dp2_func[cmd->bCommand].a_func;
    if (!s_func) {
	FIXME("handler for %s not implemented\n", GL_dp2_func[cmd->bCommand].name);
	/* return D3DERR_EXECUTE_FAILED; */
	return D3D_OK; /* pretend success */
    }
    if (priv->block && GL_dp2_func[cmd->bCommand].type == STAT) {
	/* record to state block */
	TRACE("saving state %s\n", GL_dp2_func[cmd->bCommand].name);
	GL_record_cmd(priv, &priv->block->cmds, cmd, data1, dwSize1, data2, dwSize2);
	return D3D_OK;
    }
    if (a_func) hr = a_func(priv, cmd, data1, data2);
    if (!priv->d3dp->in_scene) {
	if (s_func == D3D_GL_draw) return D3DERR_INVALIDCALL;
	/* record command */
	TRACE("recording command %s\n", GL_dp2_func[cmd->bCommand].name);
	GL_record_cmd(priv, &priv->cmds, cmd, data1, dwSize1, data2, dwSize2);
	return hr;
    }
    return s_func(priv, cmd, data1, data2);
}

HRESULT WINAPI D3DGL_FlushDP2(GL_D3D_priv* priv)
{
    if (!priv->d3dp->in_scene) return D3D_OK;

    CHECK_GL_ERROR("FlushDP2 (enter)");

    GL_replay_cmds(priv, &priv->cmds, REPLAY_EXEC|REPLAY_DEL);

    glFlush();
    CHECK_GL_ERROR("FlushDP2 (exit)");
    return D3D_OK;
}

static HRESULT GL_replay_cmds(GL_D3D_priv* priv, struct GL_D3D_saved_cmd **cmd, DWORD dwFlags)
{
    struct GL_D3D_saved_cmd *ncmd = *cmd;

    while (ncmd) {
	TRACE("replaying command %s\n", GL_dp2_func[ncmd->cmd.bCommand].name);
	if (dwFlags & REPLAY_APPLY)
	    GL_apply_cmd(priv, ncmd);
	if (dwFlags & REPLAY_CAPTURE)
	    GL_capture_cmd(priv, ncmd);
	if (dwFlags & REPLAY_EXEC) {
	    GL_dp2_func[ncmd->cmd.bCommand].s_func(priv, &ncmd->cmd, ncmd->data1, ncmd->data2);
	    /* FIXME: what to do in case of an error? */
	}

	ncmd = ncmd->next;
	if (dwFlags & REPLAY_DEL) GL_delete_cmd(priv, cmd);
    }
    return D3D_OK;
}
