/*	Direct3D OpenGL driver
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "winbase.h"
#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"

/* #define DEBUG_VERTEX_DATA */

DEFAULT_DEBUG_CHANNEL(ddraw);

static const float one4f[4] = { 1, 1, 1, 1 };
static const float zero4f[4] = { 0, 0, 0, 0 };
extern int use_clipspace_fix;

/* stuff we use that may be missing from glext.h */
#ifndef GL_ARB_texture_cube_map
#define GL_NORMAL_MAP_ARB	0x8511
#define GL_REFLECTION_MAP_ARB	0x8512
#endif

#define D3DFVF_TEXCOUNT(x) (((x) & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT)


static inline void GL_convert_position(const float* in, float* out)
{
    float w;

    /* Reportedly Microsoft uses 2^20 in their ProcessVertices impl. */
    if (in[3] == 0.0) w = 1048576.0;
    else w = 1.0 / in[3];

    out[0] = in[0] * w;
    out[1] = in[1] * w;
    out[2] = in[2] * w;
    out[3] = w;
}


static HRESULT GL_build_position_array(GL_D3D_priv* priv, DWORD fvf,
				       DWORD vstart, DWORD vcount,
				       LPGLDRAWPRIMITIVESTRIDEDDATA data,
				       const WORD* indices, DWORD nindices,
				       DWORD ibase)
{
    LPBYTE in;
    LPD3DVALUE out;
    DWORD stride = data->position.dwStride;
    DWORD size;

    size = vcount * 4 * sizeof(D3DVALUE);
    if (size > priv->vertex_array_size)
    {
	LPVOID new_array;

	size = max(size, 2 * priv->vertex_array_size);

	new_array = VirtualAlloc(NULL, size, MEM_COMMIT | MEM_RESERVE,
				 PAGE_READWRITE);

	if (new_array == NULL) return E_OUTOFMEMORY;

	VirtualFree(priv->vertex_array, 0, MEM_RELEASE);

	priv->vertex_array = new_array;
	priv->vertex_array_size = size;
    }

    in = data->position.lpvData;
    out = priv->vertex_array;

    if (indices != NULL && nindices < vcount)
    {
	in += ibase * stride;
	out -= (vstart - ibase) * 4;
	while (nindices--)
	{
	    DWORD index = *indices++;

	    GL_convert_position((LPFLOAT)(in + index * stride),
				(LPFLOAT)(out + index * 4));
	}
    }
    else
    {
	in += vstart * stride;
	while (vcount--)
	{
	    GL_convert_position((LPFLOAT)in, (LPFLOAT)out);

	    in += stride;
	    out += 4;
	}
    }

    return DD_OK;
}

static void GL_fixup_position_array(GL_D3D_priv* priv, DWORD fvf,
				    DWORD vstart, DWORD vcount,
				    LPGLDRAWPRIMITIVESTRIDEDDATA data,
				    const WORD* indices, DWORD nindices,
				    DWORD ibase)
{
    if (((fvf & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW)
	&& !(fvf & D3DFVF_GLW))
    {
	DWORD stride = 4 * sizeof(D3DVALUE);

	GL_build_position_array(priv, fvf, vstart, vcount, data, indices, nindices, ibase);

	data->position.lpvData = (LPBYTE)priv->vertex_array - vstart * stride;
	data->position.dwStride = stride;
    }
}


static inline DWORD rot16(DWORD x)
{
    return (x << 16) | (x >> 16);
}

static inline D3DCOLOR GL_reorder_color(D3DCOLOR x)
{
    return (x & 0xFF00FF00) | (rot16(x) & 0x00FF00FF);
}

/* This considers the D3DFVF_DIFFUSE and D3DFVF_SPECULAR bits in vt and
 * converts those arrays into GL colour format. It doesn't pass them to GL.
 *
 * The return value is a pointer to the colour arrays, which each colour
 * stored as 4 unsigned bytes. If both colours are requested then the
 * arrays are interleaved with diffuse first.
 */
static HRESULT GL_build_color_arrays(GL_D3D_priv* priv, DWORD fvf,
				     DWORD vstart, DWORD vcount,
				     LPGLDRAWPRIMITIVESTRIDEDDATA psd,
				     const WORD* indices, DWORD nindices,
				     DWORD ibase)
{
    LPDWORD out;
    DWORD which;
    DWORD size;

    which = fvf & (D3DFVF_DIFFUSE|D3DFVF_SPECULAR);

    /* do the calculation assuming DIFFUSE and SPECULAR are set. */
    size = vcount * 2 * sizeof(D3DCOLOR);
    if (size > priv->color_array_size)
    {
	LPVOID new_array;

	size = max(size, 2 * priv->color_array_size);

	new_array = VirtualAlloc(NULL, size, MEM_COMMIT | MEM_RESERVE,
				 PAGE_READWRITE);

	if (new_array == NULL) return E_OUTOFMEMORY;

	VirtualFree(priv->color_array, 0, MEM_RELEASE);

	priv->color_array = new_array;
	priv->color_array_size = size;
    }

    out = priv->color_array;

    if (which == (D3DFVF_DIFFUSE | D3DFVF_SPECULAR)) {
	/* This loop is written under the assumption that the diffuse and
	 * specular data are interleaved. If not then the cache striding
	 * will probably be very bad. */

	DWORD diffuse_stride = psd->color[0].dwStride;
	LPBYTE diffuse_in = (LPBYTE)psd->color[0].lpvData;
	DWORD specular_stride = psd->color[1].dwStride;
	LPBYTE specular_in = (LPBYTE)psd->color[1].lpvData;

	if (indices != NULL && nindices < vcount)
	{
	    diffuse_in += ibase * diffuse_stride;
	    specular_in += ibase * specular_stride;
	    out -= (vstart - ibase) * 2;
	    while (nindices--)
	    {
		DWORD index = *indices++;

		out[2*index+0]
		    = GL_reorder_color(*(LPDWORD)(diffuse_in
						  + index * diffuse_stride));
		out[2*index+1]
		    = GL_reorder_color(*(LPDWORD)(specular_in
						  + index * specular_stride));
	    }
	}
	else
	{
	    diffuse_in += vstart * diffuse_stride;
	    specular_in += vstart * specular_stride;
	    while (vcount--)
	    {
		out[0] = GL_reorder_color(*(LPDWORD)diffuse_in);
		out[1] = GL_reorder_color(*(LPDWORD)specular_in);

		diffuse_in += diffuse_stride;
		specular_in += specular_stride;
		out += 2;
	    }
	}
    }
    else
    {
	DWORD stride;
	LPBYTE in;

	if (which == D3DFVF_DIFFUSE)
	{
	    stride = psd->color[0].dwStride;
	    in = (LPBYTE)psd->color[0].lpvData;
	}
	else
	{
	    stride = psd->color[1].dwStride;
	    in = (LPBYTE)psd->color[1].lpvData;
	}

	if (indices != NULL && nindices < vcount)
	{
	    in += ibase * stride;
	    out -= vstart - ibase;
	    while (nindices--)
	    {
		DWORD index = *indices++;

		out[index] = GL_reorder_color(*(LPDWORD)(in + index * stride));
	    }
	}
	else
	{
	    in += vstart * stride;
	    while (vcount--)
	    {
		*out++ = GL_reorder_color(*(LPDWORD)in);
		in += stride;
	    }
	}
    }

    return DD_OK;
}

static void GL_fixup_color_arrays(GL_D3D_priv* priv, DWORD fvf,
				  DWORD vstart, DWORD vcount,
				  LPGLDRAWPRIMITIVESTRIDEDDATA data,
				  const WORD* indices, DWORD nindices,
				  DWORD ibase)
{
    if ((fvf & (D3DFVF_DIFFUSE|D3DFVF_SPECULAR))
	&& !(fvf & D3DFVF_GLCOLORORDER))
    {
	GL_build_color_arrays(priv, fvf, vstart, vcount, data, indices, nindices, ibase);

	switch (fvf & (D3DFVF_DIFFUSE|D3DFVF_SPECULAR))
	{
	case D3DFVF_DIFFUSE|D3DFVF_SPECULAR:
	    data->color[0].lpvData = (LPBYTE)priv->color_array - vstart * 8;
	    data->color[0].dwStride = 8;
	    data->color[1].lpvData = (LPBYTE)priv->color_array+sizeof(D3DCOLOR) - vstart * 8;
	    data->color[1].dwStride = 8;
	    break;

	case D3DFVF_DIFFUSE:
	    data->color[0].lpvData = (LPBYTE)priv->color_array - vstart * 4;
	    data->color[0].dwStride = 4;
	    break;

	case D3DFVF_SPECULAR:
	    data->color[1].lpvData = (LPBYTE)priv->color_array - vstart * 4;
	    data->color[1].dwStride = 4;
	    break;
	}
    }
}

/* Set up for unlit rendering in which the materials not listed in
 * DIRTY will be using the global (IDirect3DDevice7::SetMaterial)
 * values.
 *
 * All dirty materials are cleaned, except those listed in DIRTY.
 * Then DIRTY is added to the set of dirty materials.
 */
static void GL_reset_materials(GL_D3D_priv* priv, DWORD dirty)
{
    LPD3D_P d3dp = priv->d3dp;
    DWORD flags = priv->material_flags & ~dirty;

    if (flags & GLMF_DIFFUSE_DIRTY)
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,
		     (float*)&d3dp->material.u.diffuse);

    if (flags & GLMF_AMBIENT_DIRTY)
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,
		     (float*)&d3dp->material.u1.ambient);

    if (d3dp->renderstate[D3DRENDERSTATE_SPECULARENABLE])
    {
	if (flags & GLMF_SPECULAR_DIRTY)
	    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,
			 (float*)&d3dp->material.u2.specular);

	if (flags & GLMF_SHININESS_DIRTY)
	    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS,
			d3dp->material.u4.power);

	flags &= ~(GLMF_SPECULAR_DIRTY|GLMF_SHININESS_DIRTY);
    }
    else
    {
	if (flags & GLMF_SPECULAR_DIRTY)
	    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, zero4f);

	flags &= ~GLMF_SPECULAR_DIRTY;
    }

    if (flags & GLMF_EMISSIVE_DIRTY)
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION,
		     (float*)&d3dp->material.u3.emissive);

    flags &= ~(GLMF_DIFFUSE_DIRTY|GLMF_AMBIENT_DIRTY|GLMF_EMISSIVE_DIRTY);

    priv->material_flags = flags | dirty;
}

/* Set up the mapping from vertex colour to material.
 *
 * This is used to draw unlit vertices when D3DRSS_COLORVERTEX is on.
 * It walks through the D3DRSS_*MATERIALSOURCE renderstates to decide
 * which will be mapped to the current color using glColorMaterial.
 *
 * NOTE: When this returns, GL_COLOR_MATERIAL may be enabled, so turn it off
 * once you are done rendering.
 *
 * Each material is either
 * - loaded from the current colour (i.e. by glColor(Pointer))
 * - marked with the GLMF_DO_material flag
 * - or clean.
 *
 * Materials that will be altered by the current colour or are marked
 * with a GLMF_DO_material flag are also marked dirty.
 */
static DWORD GL_fixup_vertex_materials(GL_D3D_priv* priv,
				       DWORD fvf,
				       LPGLDRAWPRIMITIVESTRIDEDDATA data)
{
    LPD3D_P d3dp = priv->d3dp;

    /* These will be set to the effective colour source for each material.
     * ("Effective" means after removing sources not present in the FVF
     * and specular if disabled.) */
    DWORD diffuse, specular, ambient, emissive;
    /* For the Diffuse&Ambient combo hack. (Sacrifice) */
    DWORD ambient2;

    /* Keep track of which colour source will be loaded into the array.
     * 0 means that it hasn't been assigned yet, otherwise it is
     * D3DMCS_COLOR1 or D3DMCS_COLOR2. */
    DWORD color_array = 0;
    GLenum color_material = 0;

    /* Keep flags for which materials will be changed while drawing. We won't
     * bother to clean these materials, and we'll mark them dirty right
     * here, knowing that the'll be changed shortly. */
    DWORD inuse = 0;

    DWORD material_flags = priv->material_flags;

    /* Pick out D3DFVF_DIFFUSE and D3DFVF_SPECULAR, mapping them to
     * D3DMCS_COLOR1 and D3DMCS_COLOR2 respectively. */
    DWORD vertex_colors = (fvf >> 6) & 3;

    /* Don't do anything the vertices are prelit, or vertex materials are
     * not in use at all. */
    if (((fvf & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW)
        || (!(fvf & D3DFVF_NORMAL) && priv->d3d_version < 8)
	|| !d3dp->renderstate[D3DRENDERSTATE_COLORVERTEX]
	|| !d3dp->renderstate[D3DRENDERSTATE_LIGHTING])
    {
	/* We don't have to disable GL_COLOR_MATERIAL when the vertices are
	 * prelit. Is it better to disable it or not? */
	if (material_flags & GLMF_COLOR_MATERIAL)
	{
	    glDisable(GL_COLOR_MATERIAL);
	    priv->material_flags &= ~GLMF_COLOR_MATERIAL;
	}

	if (((fvf & D3DFVF_POSITION_MASK) != D3DFVF_XYZRHW)
	    && ((fvf & D3DFVF_NORMAL) || priv->d3d_version >= 8)
	    && d3dp->renderstate[D3DRENDERSTATE_LIGHTING])
	{
	    GL_reset_materials(priv, 0);

	    /* Lighting is on, but color vertex is off, so we won't be
	     * needing these. */
	    fvf &= ~(D3DFVF_DIFFUSE|D3DFVF_SPECULAR);
	}

	return fvf;
    }

    material_flags &= ~(GLMF_DO_DIFFUSE|GLMF_DO_SPECULAR|GLMF_DO_AMBIENT
			|GLMF_DO_EMISSIVE);

    diffuse = d3dp->renderstate[D3DRENDERSTATE_DIFFUSEMATERIALSOURCE]
	& vertex_colors;
    specular = d3dp->renderstate[D3DRENDERSTATE_SPECULARMATERIALSOURCE]
	& vertex_colors;
    ambient = d3dp->renderstate[D3DRENDERSTATE_AMBIENTMATERIALSOURCE]
	& vertex_colors;
    emissive = d3dp->renderstate[D3DRENDERSTATE_EMISSIVEMATERIALSOURCE]
	& vertex_colors;

    /* ignore specular if it is disabled. */
    if (!d3dp->renderstate[D3DRENDERSTATE_SPECULARENABLE])
	specular = D3DMCS_MATERIAL;

    /* for the purposes of deciding whether or not to fallback,
     * ignore ambient if it is the same as diffuse */
    ambient2 = ambient;
    if (ambient2 == diffuse)
	ambient2 = D3DMCS_MATERIAL;

    /* If two materials are using vertex colours (except for the diffuse
     * & ambient combo), then we will need to use immediate-mode calls. */
    if (((diffuse != D3DMCS_MATERIAL) + (ambient2 != D3DMCS_MATERIAL)
	 + (specular != D3DMCS_MATERIAL)
	 + (emissive != D3DMCS_MATERIAL)) > 1) {
	priv->glFlags &= ~GLF_USE_VERTEX_ARRAYS;
    }

    /* Choose a material that will be mapped to the current colour.
     * Diffuse goes first so that we can prefer the diffuse & ambient combo.
     * Also mark which materials will have to be loaded by glMaterial. */
    if (diffuse != D3DMCS_MATERIAL)
    {
	if (diffuse == ambient)
	{
	    assert(ambient2 == D3DMCS_MATERIAL);
	    color_material = GL_AMBIENT_AND_DIFFUSE;
	    inuse |= GLMF_AMBIENT_DIRTY;
	    TRACE("choosing GL_DIFFUSE_AND_AMBIENT\n");
	}
	else
	{
	    color_material = GL_DIFFUSE;
	    TRACE("choosing GL_DIFFUSE\n");
	}

	color_array = diffuse;
	inuse |= GLMF_DIFFUSE_DIRTY;
    }

    if (ambient2 != D3DMCS_MATERIAL)
    {
	assert(ambient2 == ambient);

	if (color_array == 0)
	{
	    color_material = GL_AMBIENT;
	    color_array = ambient;
	    TRACE("choosing GL_AMBIENT\n");
	}
	else
	{
	    material_flags |= GLMF_DO_AMBIENT;
	}

	inuse |= GLMF_AMBIENT_DIRTY;
    }

    if (specular != D3DMCS_MATERIAL)
    {
	if (color_array == 0)
	{
	    color_material = GL_SPECULAR;
	    color_array = specular;
	    TRACE("choosing GL_SPECULAR\n");
	}
	else
	{
	    material_flags |= GLMF_DO_SPECULAR;
	}

	inuse |= GLMF_SPECULAR_DIRTY;
    }

    if (emissive != D3DMCS_MATERIAL)
    {
	if (color_array == 0)
	{
	    color_material = GL_EMISSION;
	    color_array = emissive;
	    TRACE("choosing GL_EMISSIVE\n");
	}
	else
	{
	    material_flags |= GLMF_DO_EMISSIVE;
	}

	inuse |= GLMF_EMISSIVE_DIRTY;
    }

    priv->material_flags = material_flags;

    GL_reset_materials(priv, inuse);

    if (color_array != 0)
    {
	if (color_material != priv->current_color_material)
	{
	    glColorMaterial(GL_FRONT_AND_BACK, color_material);
	    priv->current_color_material = color_material;
	}

	if (!(material_flags & GLMF_COLOR_MATERIAL))
	{
	    glEnable(GL_COLOR_MATERIAL);
	    priv->material_flags |= GLMF_COLOR_MATERIAL;
	}

	/* Now put the in-use colour into slot 0, and tell the colour
	 * fixup and loader code to use that. */

	if (color_array == D3DMCS_COLOR2)
	    data->color[0] = data->color[1];

	return (fvf &~ D3DFVF_SPECULAR) | D3DFVF_DIFFUSE;
    }
    else
    {
	/* We should detect when vertex colours are specified in the fvf,
	 * but not actually used (because only global materials are in use)
	 * and mask them off here. */
	return fvf;
    }
}

static void GL_load_arrays(GL_D3D_priv* priv, DWORD fvf, DWORD start,
			   LPGLDRAWPRIMITIVESTRIDEDDATA stride)
{
    DWORD position_size, num_weights;
    DWORD inuse, tstages;
    DWORD i;

    if ((fvf & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW)
	position_size = 4;
    else
	position_size = 3;

    num_weights = max((fvf & D3DFVF_POSITION_MASK) / 2, 2) - 2;
    if (num_weights && (fvf & D3DFVF_LASTBETA_UBYTE4)) num_weights--;

#define STRIDE_DATA_STRIDE(x) \
  stride->x.dwStride
#define STRIDE_DATA_PTR(x) \
  ((LPBYTE)stride->x.lpvData + start * stride->x.dwStride)
#define STRIDE_DATA(x) \
  STRIDE_DATA_STRIDE(x), STRIDE_DATA_PTR(x)

    if (!priv->pVertexArray)
	glEnableClientState(GL_VERTEX_ARRAY);
    if (priv->pVertexArray != STRIDE_DATA_PTR(position))
    {
	glVertexPointer(position_size, GL_FLOAT, STRIDE_DATA(position));
	priv->pVertexArray = STRIDE_DATA_PTR(position);
    }

    if (fvf & D3DFVF_NORMAL)
    {
	if (!priv->pNormalArray)
	    glEnableClientState(GL_NORMAL_ARRAY);
	if (priv->pNormalArray != STRIDE_DATA_PTR(normal) ||
	    priv->sNormalArray != STRIDE_DATA_STRIDE(normal))
	{
	    glNormalPointer(GL_FLOAT, STRIDE_DATA(normal));
	    priv->pNormalArray = STRIDE_DATA_PTR(normal);
	    priv->sNormalArray = STRIDE_DATA_STRIDE(normal);
	}
    }
    else
    {
	if (priv->pNormalArray)
	{
	    glDisableClientState(GL_NORMAL_ARRAY);
	    priv->pNormalArray = NULL;
	    priv->sNormalArray = 0;

	    /* YES it's necessary - D3D8 allows (ambient) lighting on vertices
	     * without normals, don't leave it undefined */
	    glNormal3f(0, 0, 0);
	}
    }

    if (fvf & D3DFVF_DIFFUSE)
    {
	if (!priv->pColorArray)
	    glEnableClientState(GL_COLOR_ARRAY);
	if (priv->pColorArray != STRIDE_DATA_PTR(color[0]) ||
	    priv->sColorArray != STRIDE_DATA_STRIDE(color[0]))
	{
	    glColorPointer(4, GL_UNSIGNED_BYTE, STRIDE_DATA(color[0]));
	    priv->pColorArray = STRIDE_DATA_PTR(color[0]);
	    priv->sColorArray = STRIDE_DATA_STRIDE(color[0]);
	}
    }
    else
    {
	if (priv->pColorArray)
	{
	    glDisableClientState(GL_COLOR_ARRAY);
	    priv->pColorArray = NULL;
	    priv->sColorArray = 0;

	    /* YES it's necessary - D3D (all versions) defaults to white
	     * when no colors are specified, don't leave it undefined */
	    glColor4fv(one4f);
	}
    }

    if (HAVE_GL_EXT(priv, EXT_secondary_color))
    {
	if (fvf & D3DFVF_SPECULAR)
	{
	    if (!priv->pSecondaryColorArray) 
		glEnableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
	    if (priv->pSecondaryColorArray != STRIDE_DATA_PTR(color[1]) ||
	        priv->sSecondaryColorArray != STRIDE_DATA_STRIDE(color[1]))
	    {
		GLD(priv)->glSecondaryColorPointerEXT(3, GL_UNSIGNED_BYTE,
						      STRIDE_DATA(color[1]));
		priv->pSecondaryColorArray = STRIDE_DATA_PTR(color[1]);
		priv->sSecondaryColorArray = STRIDE_DATA_STRIDE(color[1]);
	    }
	}
	else
	{
	    if (priv->pSecondaryColorArray)
	    {
		glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
		priv->pSecondaryColorArray = NULL;
		priv->sSecondaryColorArray = 0;

		/* YES it's necessary - we don't want to leave
		 * the secondary color undefined */
		GLD(priv)->glSecondaryColor3ubEXT(0, 0, 0);
	    }
	}
    }

    tstages = priv->lpCaps->wMaxTextureBlendStages;
    inuse = D3DFVF_TEXCOUNT(fvf);

    for (i = 0; i < tstages; i++)
    {
	DWORD coord_idx = priv->d3dp->texstagestate[i][D3DTSS_TEXCOORDINDEX];

	GLD(priv)->glClientActiveTextureARB(GL_TEXTURE0_ARB + i);

	if ((coord_idx & D3DTSS_TCI_MASK) || (coord_idx >= inuse))
	{
	    /* Automatic texture coordinate generation is in use,
	     * or coord_idx is out of range */
	    if (priv->pTexCoordArray[i])
	    {
		GLD(priv)->glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		priv->pTexCoordArray[i] = NULL;
		priv->sTexCoordArray[i] = 0;

		/* this may be less necessary, not sure */
		GLD(priv)->glMultiTexCoord2fARB(GL_TEXTURE0_ARB + i, 0, 0);
	    }
	}
	else
	{
	    if (!priv->pTexCoordArray[i])
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	    if (priv->pTexCoordArray[i] != STRIDE_DATA_PTR(textureCoords[coord_idx]) ||
	        priv->sTexCoordArray[i] != STRIDE_DATA_STRIDE(textureCoords[coord_idx]))
	    {
		glTexCoordPointer(2, GL_FLOAT,
				  STRIDE_DATA(textureCoords[coord_idx]));
		priv->pTexCoordArray[i] = STRIDE_DATA_PTR(textureCoords[coord_idx]);
		priv->sTexCoordArray[i] = STRIDE_DATA_STRIDE(textureCoords[coord_idx]);
	    }
	}
    }

    if (HAVE_GL_EXT(priv, ARB_vertex_blend))
    {
	if (num_weights)
	{
	    glEnableClientState(GL_WEIGHT_ARRAY_ARB);
	    GLD(priv)->glWeightPointerARB(num_weights, GL_FLOAT,
					  STRIDE_DATA(blend_weights[0]));
	}
	else
	{
	    glDisableClientState(GL_WEIGHT_ARRAY_ARB);
	}
	/* TODO: matrix palette indices */
    }
    else if (HAVE_GL_EXT(priv, EXT_vertex_weighting))
    {
	if (num_weights)
	{
	    glEnableClientState(GL_VERTEX_WEIGHT_ARRAY_EXT);
	    GLD(priv)->glVertexWeightPointerEXT(num_weights, GL_FLOAT,
						STRIDE_DATA(blend_weights[0]));
	}
	else
	{
	    glDisableClientState(GL_VERTEX_WEIGHT_ARRAY_EXT);
	}
    }
    /* TODO: fog */

#undef STRIDE_DATA
}

static void GL_clear_texture_transforms(GL_D3D_priv* priv)
{
    DWORD tstages = priv->lpCaps->wMaxTextureBlendStages;
    DWORD i;

    if (priv->texture_flags & GLTF_DEFAULT_TEXCOORD_TRANSFORM) return;

    glMatrixMode(GL_TEXTURE);

    for (i = 0; i < tstages; i++)
    {
	GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + i);

	glLoadIdentity();

	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
    }
}

static void GL_setup_texture_transforms(GL_D3D_priv* priv)
{
    LPD3D_P d3dp = priv->d3dp;
    DWORD tstages = priv->lpCaps->wMaxTextureBlendStages;
    DWORD i;

    if (priv->texture_flags & GLTF_DEFAULT_TEXCOORD_TRANSFORM) return;

    glMatrixMode(GL_TEXTURE);

    for (i=0; i < tstages; i++)
    {
	DWORD coord_idx = d3dp->texstagestate[i][D3DTSS_TEXCOORDINDEX];
	BOOL want_texgen = FALSE;
	GLenum texgen_mode = 0;

	GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + i);

	if (coord_idx & D3DTSS_TCI_MASK)
	{
	    switch (coord_idx >> 16)
	    {
	    case D3DTSS_TCI_CAMERASPACEPOSITION >> 16:
		/* We assume that the eye plane has been set up to map
		 * x,y,z to s,t,r. See D3D_GL_init. */
		texgen_mode = GL_EYE_LINEAR;
		want_texgen = TRUE;
		break;

	    case D3DTSS_TCI_CAMERASPACENORMAL >> 16:
		if (HAVE_GL_EXT(priv, ARB_texture_cube_map)
		    || HAVE_GL_EXT(priv, NV_texgen_reflection))
		{
		    texgen_mode = GL_NORMAL_MAP_ARB;
		    want_texgen = TRUE;
		}
		break;

	    case D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR >> 16:
		if (HAVE_GL_EXT(priv, ARB_texture_cube_map)
		    || HAVE_GL_EXT(priv, NV_texgen_reflection))
		{
		    texgen_mode = GL_REFLECTION_MAP_ARB;
		    want_texgen = TRUE;
		}
		break;
	    }
	}

	if (want_texgen)
	{
	    glEnable(GL_TEXTURE_GEN_S);
	    glEnable(GL_TEXTURE_GEN_T);
	    glEnable(GL_TEXTURE_GEN_R);
	    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, texgen_mode);
	    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, texgen_mode);
	    glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, texgen_mode);
	}
	else
	{
	    glDisable(GL_TEXTURE_GEN_S);
	    glDisable(GL_TEXTURE_GEN_T);
	    glDisable(GL_TEXTURE_GEN_R);
	}

	if (d3dp->texstagestate[i][D3DTSS_TEXTURETRANSFORMFLAGS]
	    == D3DTTFF_DISABLE)
	{
	    glLoadIdentity();
	}
	else
	{
	    /* We interpret D3DTTFF_COUNTX and D3DTTFF_PROJECTED as
	     * follows:
	     *
	     * All extra dimensions (beyond the count setting) are set
	     * to 0, except for q which is set to 1.
	     * If D3DTTFF_PROJECTED is set then the last dimension is
	     * sent to q, and is set to 0.
	     *
	     * So for example with D3DTTFF_COUNT3|D3DTTFF_PROJECTED,
	     * have s' = s, t'=t, r'=0, q'=r.
	     *
	     * Note that D3DTTFF_COUNT4 and
	     * D3DTTFF_COUNT4|D3DTTFF_PROJECTED are the same because
	     * GL always divides by the 4th coordinate.
	     */
	    D3DTEXTURETRANSFORMFLAGS flags
		= d3dp->texstagestate[i][D3DTSS_TEXTURETRANSFORMFLAGS];

	    unsigned coord_count = flags & D3DTTFF_COUNT_MASK;

	    if (coord_count < 4)
	    {
		static const float count_plain[3][4][4] =
		    {
			/* COUNT1: t' = r' = 0 */
			{
			    { 1, 0, 0, 0 },
			    { 0, 0, 0, 0 },
			    { 0, 0, 0, 0 },
			    { 0, 0, 0, 1 }
			},
			/* COUNT2: r' = 0 */
			{
			    { 1, 0, 0, 0 },
			    { 0, 1, 0, 0 },
			    { 0, 0, 0, 0 },
			    { 0, 0, 0, 1 }
			},
			/* COUNT3: identity */
			{
			    { 1, 0, 0, 0 },
			    { 0, 1, 0, 0 },
			    { 0, 0, 1, 0 },
			    { 0, 0, 0, 1 }
			}
		    };

		static const float count_projected[3][4][4] =
		    {
			/* COUNT1: q' = s, s' = t' = r' = 0 */
			{
			    { 0, 0, 0, 1 },
			    { 0, 0, 0, 0 },
			    { 0, 0, 0, 0 },
			    { 0, 0, 0, 0 }
			},
			/* COUNT2: q' = t, t' = r' = 0 */
			{
			    { 1, 0, 0, 0 },
			    { 0, 0, 0, 1 },
			    { 0, 0, 0, 0 },
			    { 0, 0, 0, 0 },
			},
			/* COUNT3: q' = r, r' = 0 */
			{
			    { 1, 0, 0, 0 },
			    { 0, 1, 0, 0 },
			    { 0, 0, 0, 1 },
			    { 0, 0, 0, 0 },
			}
		    };

		if (flags & D3DTTFF_PROJECTED)
		{
		    glLoadMatrixf((const float*)count_projected[coord_count-1]);
		    /* Because we load q from another coordinate,
		     * we don't have to force it to 1. */
		    glMultMatrixf((const float *)&d3dp->txfrm[i]);
		}
		else
		{
		    float tmp[4][4];

		    glLoadMatrixf((const float *)count_plain[coord_count-1]);

		    /* We need to ensure that the output q' = 1.
		     * Since the original q texture coordinate was 1,
		     * we can achieve this by setting the 4th row of
		     * the matrix to 0, 0, 0, 1, so that q' = q = 1.
		     */
		    memcpy(tmp, &d3dp->txfrm[i], 4*4*sizeof(float));

		    tmp[0][3] = 0;
		    tmp[1][3] = 0;
		    tmp[2][3] = 0;
		    tmp[3][3] = 1;

		    glMultMatrixf((const float*)tmp);
		}
	    }
	    else
	    {
		glLoadMatrixf((const float *)&d3dp->txfrm[i]);
	    }
	}
    }
}

#ifndef HAVE_OSMESA

static const GLfloat clipspace_fix[] =
{
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 2, 0,
    0, 0, -1, 1
};

#else

/* OSMESA y-swap */
static const GLfloat clipspace_fix[] =
{
    1, 0, 0, 0,
    0, -1, 0, 0, /* FLIP */
    0, 0, 2, 0,
    0, 0, -1, 1
};

#endif

static void GL_setup_transforms(GL_D3D_priv* priv, DWORD vt)
{
    LPD3D_P d3dp = priv->d3dp;
    if (priv->glFlags & (GLF_NEW_TNL_MODE | GLF_NEW_TRANSFORM)) {
	if ((vt & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW) {
	    GLdouble left, top, height, width, minZ, maxZ;

	    TRACE("Transformed Vertices\n");

	    /* Do not put any transformation matrixes */
	    glMatrixMode(GL_MODELVIEW);
	    glLoadIdentity();
	    /* the Red Book (under "OpenGL Correctness Tips") recommends offseting coordinates
	     * by 0.375 to be sure that integer pixel coordinates can be used in all cases...
	     * I'm not sure what the best value to use is for us, though (0.5? 0.25?),
	     * but in order to get nearest-filtered textures mapped correctly, it must be
	     * *something*, due to the half-pixel issue discussed below...
	     * so I think I'll use the book's recommendation for now
	     * (this fixes ugly EverQuest fonts seen on some systems) */
	    glTranslatef(0.375, 0.375, 0);

	    glMatrixMode(GL_PROJECTION);
	    glLoadIdentity();
	    if (HAVE_GL_EXT(priv, ARB_vertex_blend))
		glDisable(GL_VERTEX_BLEND_ARB);
	    else if (HAVE_GL_EXT(priv, EXT_vertex_weighting))
		glDisable(GL_VERTEX_WEIGHTING_EXT);

	    left   = d3dp->viewport.dwX;
	    top    = d3dp->viewport.dwY;
	    height = d3dp->viewport.dwHeight;
	    width  = d3dp->viewport.dwWidth;
	    minZ   = d3dp->viewport.dvMinZ;
	    maxZ   = d3dp->viewport.dvMaxZ;

	    /* TODO: Use 0 and 1 for Znear, Zfar. */
	    /* TODO: glDepthRange(0, 1); */
	    glOrtho(left, left + width, top + height, top, -minZ, -maxZ);

	    GL_clear_texture_transforms(priv);
	}
	else {
	    TRACE("Untransformed Vertices\n");

	    /* Need to put the correct transformation again */
	    switch (d3dp->renderstate[D3DRENDERSTATE_VERTEXBLEND]) {
	    default:
	    case D3DVBLEND_3WEIGHTS:
		glMatrixMode(GL_MODELVIEW3_ARB);
		glLoadMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_VIEW]);
		glMultMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_WORLD3]);
		/* fall thru */
	    case D3DVBLEND_2WEIGHTS:
		glMatrixMode(GL_MODELVIEW2_ARB);
		glLoadMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_VIEW]);
		glMultMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_WORLD2]);
		/* fall thru */
	    case D3DVBLEND_1WEIGHT:
		glMatrixMode(GL_MODELVIEW1_ARB);
		glLoadMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_VIEW]);
		glMultMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_WORLD1]);
		/* fall thru */
	    case D3DVBLEND_DISABLE:
		glMatrixMode(GL_MODELVIEW0_ARB); /* aka standard GL_MODELVIEW */
		glLoadMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_VIEW]);
		glMultMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_WORLD]);
	    }

	    glMatrixMode(GL_PROJECTION);
	    if( use_clipspace_fix ) {
	      /* in opengl the z clipping is -W < z <= W   but in d3d it is 0 < z <= w
	       * this matrix makes openlg clip like d3d */
	      glLoadMatrixf(clipspace_fix);
	    } else {
	      /* however, with vertex programs, we cannot easily apply such a fix, and
	       * apps the mix fixed-function rendering with programmable, end up with
	       * conflicting depth info.  This lets us disable it on a per-app basis. */
	      glLoadIdentity(); 
	    }
	    /* the OpenGL spec states:
	     *  "A fragment is located by its lower-left corner, which lies on integer grid coordinates"
	     * the Direct3D rasterization rules state:
	     *  "Pixel centers are at integer coordinates"
	     * so it seems we need to translate the pre-rasterization coordinates by
	     * half a pixel in order to get nearest-filtered texture maps (and the polygons
	     * that they are mapped onto) mapped correctly
	     * (this fixes ugly WarCraft III fonts seen on some systems)
	     */
	    glTranslatef(2*0.5/d3dp->viewport.dwWidth, 2*0.5/d3dp->viewport.dwHeight, 0);
	    glMultMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_PROJECTION]);

	    GL_setup_texture_transforms(priv);

	    /* TODO: Reload glDepthRange. */

	    /* Configure vertex blending */
	    if (HAVE_GL_EXT(priv, ARB_vertex_blend)) {
		if (d3dp->renderstate[D3DRENDERSTATE_VERTEXBLEND]) {
		    glEnable(GL_VERTEX_BLEND_ARB);
		    GLD(priv)->glVertexBlendARB(d3dp->renderstate[D3DRENDERSTATE_VERTEXBLEND] + 1);
		}
		else
		    glDisable(GL_VERTEX_BLEND_ARB);
	    } else if (HAVE_GL_EXT(priv, EXT_vertex_weighting)) {
		if (d3dp->renderstate[D3DRENDERSTATE_VERTEXBLEND])
		    glEnable(GL_VERTEX_WEIGHTING_EXT);
		else
		    glDisable(GL_VERTEX_WEIGHTING_EXT);
	    }
	}
	priv->glFlags &= ~(GLF_NEW_TRANSFORM | GLF_NEW_TEXCOORD_TRANSFORM);
    }
    else if (priv->glFlags & GLF_NEW_TEXCOORD_TRANSFORM)
    {
	GL_setup_texture_transforms(priv);
	priv->glFlags &= ~GLF_NEW_TEXCOORD_TRANSFORM;
    }
}

static void GL_setup_lighting(GL_D3D_priv* priv, DWORD vt)
{
    LPD3D_P d3dp = priv->d3dp;

    /* DX8: Only the render state should be used to distinguish prelit,
     * but apparently only for untransformed vertices */
    if (d3dp->renderstate[D3DRENDERSTATE_LIGHTING] &&
	((vt & D3DFVF_POSITION_MASK) != D3DFVF_XYZRHW) &&
	((vt & D3DFVF_NORMAL) || (priv->d3d_version >= 8))) {
	TRACE("Unlit Vertices\n");
	if (!priv->isLightingEnabled)
	{
	    glEnable(GL_LIGHTING);
	    priv->isLightingEnabled = TRUE;
	}
	GL_LoadLightPositions(priv, priv->glFlags & GLF_NEW_VIEW);
	priv->glFlags &= ~GLF_NEW_VIEW;
    } else {
	TRACE("Lit Vertices\n");
	if (priv->isLightingEnabled)
	{
	    glDisable(GL_LIGHTING);
	    priv->isLightingEnabled = FALSE;
	}
    }
}

#define STRIDE(data,idx) \
 ((LPVOID)((LPBYTE)(data).lpvData + (idx) * (data).dwStride))

/* It is assumed that materials being set are marked dirty, and materials
 * not being set are clean. */
void GL_draw_unlit_vertices(GL_D3D_priv* priv, GLenum gltype, DWORD vt,
			    DWORD start, DWORD count, GL_IndexFunc index,
			    LPVOID index_data)
{
    LPD3D_P d3dp = priv->d3dp;
    DWORD stop = start + count;

    const D3DDP_PTRSTRIDE *diffuse = NULL;
    const D3DDP_PTRSTRIDE *specular = NULL;
    const D3DDP_PTRSTRIDE *ambient = NULL;
    const D3DDP_PTRSTRIDE *emissive = NULL;

    BOOL do_diffuse = priv->material_flags & GLMF_DO_DIFFUSE;
    BOOL do_specular = priv->material_flags & GLMF_DO_SPECULAR;
    BOOL do_ambient = priv->material_flags & GLMF_DO_AMBIENT;
    BOOL do_emissive = priv->material_flags & GLMF_DO_EMISSIVE;

    DWORD cnt;

    assert(vt & D3DFVF_NORMAL || priv->d3d_version >= 8);
    assert(vt & D3DFVF_POSITION_MASK && (vt & D3DFVF_POSITION_MASK) != D3DFVF_XYZRHW);
    assert(glIsEnabled(GL_LIGHTING));
    assert(d3dp->renderstate[D3DRENDERSTATE_COLORVERTEX]);

    TRACE("%d: %ld %ld %ld %ld\n",
	  do_diffuse+do_specular+do_ambient+do_emissive,
	  d3dp->renderstate[D3DRENDERSTATE_DIFFUSEMATERIALSOURCE],
	  d3dp->renderstate[D3DRENDERSTATE_SPECULARMATERIALSOURCE],
	  d3dp->renderstate[D3DRENDERSTATE_AMBIENTMATERIALSOURCE],
	  d3dp->renderstate[D3DRENDERSTATE_EMISSIVEMATERIALSOURCE]);

    if (do_diffuse)
	diffuse = &priv->stride.color[d3dp->renderstate[D3DRENDERSTATE_DIFFUSEMATERIALSOURCE] - 1];

    if (do_specular)
	specular = &priv->stride.color[d3dp->renderstate[D3DRENDERSTATE_SPECULARMATERIALSOURCE] - 1];

    if (do_ambient)
	ambient = &priv->stride.color[d3dp->renderstate[D3DRENDERSTATE_AMBIENTMATERIALSOURCE] - 1];

    if (do_emissive)
	emissive = &priv->stride.color[d3dp->renderstate[D3DRENDERSTATE_EMISSIVEMATERIALSOURCE] - 1];

    glBegin(gltype);
    for (cnt = start; cnt < stop; cnt++)
    {
	DWORD cur = index(cnt, index_data);

	if (do_diffuse)
	{
	    DWORD col = *(LPDWORD)STRIDE(*diffuse, cur);

	    GLfloat color[4] =
		{ RGBA_GETRED(col) / 255.0,
		  RGBA_GETGREEN(col) / 255.0,
		  RGBA_GETBLUE(col) / 255.0,
		  RGBA_GETALPHA(col) / 255.0
		};

	    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color);
	}

	if (do_specular)
	{
	    DWORD col = *(LPDWORD)STRIDE(*specular, cur);

	    GLfloat color[4] =
		{ RGBA_GETRED(col) / 255.0,
		  RGBA_GETGREEN(col) / 255.0,
		  RGBA_GETBLUE(col) / 255.0,
		  RGBA_GETALPHA(col) / 255.0
		};

	    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);
	}

	if (do_ambient)
	{
	    DWORD col = *(LPDWORD)STRIDE(*ambient, cur);

	    GLfloat color[4] =
		{ RGBA_GETRED(col) / 255.0,
		  RGBA_GETGREEN(col) / 255.0,
		  RGBA_GETBLUE(col) / 255.0,
		  RGBA_GETALPHA(col) / 255.0
		};

	    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color);
	}

	if (do_emissive)
	{
	    DWORD col = *(LPDWORD)STRIDE(*emissive, cur);

	    GLfloat color[4] =
		{ RGBA_GETRED(col) / 255.0,
		  RGBA_GETGREEN(col) / 255.0,
		  RGBA_GETBLUE(col) / 255.0,
		  RGBA_GETALPHA(col) / 255.0
		};

	    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);
	}

	glArrayElement(cur);
    }

    glEnd();
}

/* Determine if vt is directly supported by OpenGL given the current
 * renderstates and settings. */
static BOOL GL_IsDirectlyRenderable(GL_D3D_priv* priv, DWORD vt)
{
    LPD3D_P d3dp = priv->d3dp;

    /* RHW is directly supported if it has been converted to OpenGL format. */
    if ((vt & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW && !(vt & D3DFVF_GLW))
	return FALSE;

    if (vt & (D3DFVF_DIFFUSE|D3DFVF_SPECULAR))
    {
	DWORD vertex_colors = (vt >> 6) & 3;
	DWORD diffuse, specular, ambient, emissive;

	/* If the vertex is prelit, then any specified vertex colours are
	 * in use. */
	if ((vt & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW
	    || (!(vt & D3DFVF_NORMAL) && priv->d3d_version < 8)
	    || !d3dp->renderstate[D3DRENDERSTATE_COLORVERTEX]
	    || !d3dp->renderstate[D3DRENDERSTATE_LIGHTING])
	    return (vt & D3DFVF_GLCOLORORDER);

	diffuse = d3dp->renderstate[D3DRENDERSTATE_DIFFUSEMATERIALSOURCE]
	    & vertex_colors;
	specular = d3dp->renderstate[D3DRENDERSTATE_SPECULARMATERIALSOURCE]
	    & vertex_colors;
	ambient = d3dp->renderstate[D3DRENDERSTATE_AMBIENTMATERIALSOURCE]
	    & vertex_colors;
	emissive = d3dp->renderstate[D3DRENDERSTATE_EMISSIVEMATERIALSOURCE]
	    & vertex_colors;

	/* ignore specular if it is disabled. */
	if (!d3dp->renderstate[D3DRENDERSTATE_SPECULARENABLE])
	    specular = D3DMCS_MATERIAL;

	/* for the purposes of deciding whether or not to fallback,
	 * ignore ambient if it is the same as diffuse */
	if (ambient == diffuse)
	    ambient = D3DMCS_MATERIAL;

	/* If two materials are using vertex colours (except for the diffuse
	 * & ambient combo), then we will need to use immediate-mode calls. */
	if (((diffuse != D3DMCS_MATERIAL) + (ambient != D3DMCS_MATERIAL)
	     + (specular != D3DMCS_MATERIAL)
	     + (emissive != D3DMCS_MATERIAL)) > 1)
	    return FALSE;

	/* If any vertex colours are in use, then we need them to be in
	 * OpenGL order.
	 * NOTE: The bitwise | is a silly optimisation, not a typo. */
	if (diffuse | ambient | specular | emissive)
	    return (vt & D3DFVF_GLCOLORORDER);
    }

    return TRUE;
}

/* This is common to setting up from either a D3D stride, or from
 * a pointer and FVF. */
static void
GL_SetupStrideBlendWeights(DWORD fvf, LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    int num_weights = min((fvf & D3DFVF_POSITION_MASK) / 2, 2) - 2;
    int i;

    /* Blend weights immediately follow the position data.
     * There are a maximum of 4: XYZB5 does not actually specify 5 blend
     * weights since it must be used with D3DFVF_LASTBETA_UBYTE4 flag,
     * meaning that the last float is actually 4 matrix palette indices. */
    for (i=0; i < 4; i++)
    {
	gl->blend_weights[i].lpvData = (LPFLOAT)gl->position.lpvData + 3 + i;
	gl->blend_weights[i].dwStride = gl->position.dwStride;
    }

    /* Setup the last blend weight as the matrix index pointer.
     * (In case that's what it actually is.) */
    gl->matrix_indices.lpvData
	= (LPFLOAT)gl->position.lpvData + 3 + num_weights - 1;
}

/* Apparently there is a super-secret fog value hidden in specular alpha. */
static void
GL_SetupStrideFogCoordinate(DWORD fvf, LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    gl->fog_coordinate.lpvData = (LPBYTE)gl->color[1].lpvData + 3;
    gl->fog_coordinate.dwStride = gl->color[1].dwStride;
}

static void
GL_ConvertD3DStrideToGLStride(DWORD fvf,
			      const D3DDRAWPRIMITIVESTRIDEDDATA *d3d,
			      LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    int i;

    gl->position = d3d->position;
    gl->normal   = d3d->normal;
    gl->color[0] = d3d->diffuse;
    gl->color[1] = d3d->specular;

    for (i = 0; i < D3DDP_MAXTEXCOORD; i++)
	gl->textureCoords[i] = d3d->textureCoords[i];

    GL_SetupStrideBlendWeights(fvf, gl);
    GL_SetupStrideFogCoordinate(fvf, gl);
}

static DWORD
GL_VertexDeclarationToGLStride(GL_D3D_priv* priv, LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    VertexShader *vs = priv->current_vs;
    DWORD fvf = 0;
    DWORD texcount = 0;
    int i;
    BYTE streamnum; 

    TRACE("converting vertex shader declaration to gl stride\n");
    /* position */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_POSITION].type) ) {
      streamnum = vs->input_reg[D3DVSDE_POSITION].stream;
      gl->position.dwStride = priv->d3dp->stream[streamnum].stride;
      gl->position.lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
			     + vs->input_reg[D3DVSDE_POSITION].offset;
      switch(GET_INPUT_REG_TYPE(vs->input_reg[D3DVSDE_POSITION].type)) {
	case D3DVSDT_FLOAT3: fvf |= D3DFVF_XYZ; break;
	default: TRACE("unhandled data format for D3DVSDE_POSITION\n");
      }
    }
    /* blend weights */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_BLENDWEIGHT].type) ) {
      streamnum = vs->input_reg[D3DVSDE_BLENDWEIGHT].stream;
      /* if the position has not been set, these are XYZRHW, otherwise blend weights 
       * (at least that's the best I can decipher from the MS docs) */
      if( !fvf ) {
	gl->position.dwStride = priv->d3dp->stream[streamnum].stride;
	gl->position.lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
			       + vs->input_reg[D3DVSDE_BLENDWEIGHT].offset;
	fvf = D3DFVF_XYZRHW;
      } else {
	for( i=0; i<4; i++ ) {    
	  gl->blend_weights[i].dwStride = priv->d3dp->stream[streamnum].stride;
	  gl->blend_weights[i].lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
			       + vs->input_reg[D3DVSDE_BLENDWEIGHT].offset + i;
	}
	switch(GET_INPUT_REG_TYPE(vs->input_reg[D3DVSDE_BLENDWEIGHT].type)) {
	  case D3DVSDT_FLOAT1: fvf = D3DFVF_XYZB1; break;
	  case D3DVSDT_FLOAT2: fvf = D3DFVF_XYZB2; break;
	  case D3DVSDT_FLOAT3: fvf = D3DFVF_XYZB3; break;
	  case D3DVSDT_FLOAT4: fvf = D3DFVF_XYZB4; break;
	  default: TRACE("unhandled data format for D3DVSDE_BLEND_WEIGHT\n");
	}
      }
    }
    /* blend indices */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_BLENDINDICES].type) ) {
      streamnum = vs->input_reg[D3DVSDE_BLENDINDICES].stream;
      if (GET_INPUT_REG_TYPE(vs->input_reg[D3DVSDE_BLENDINDICES].type) == D3DVSDT_UBYTE4) {
	gl->matrix_indices.dwStride = priv->d3dp->stream[streamnum].stride;
	gl->matrix_indices.lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
	                             + vs->input_reg[D3DVSDE_BLENDINDICES].offset;
	if( (fvf & D3DFVF_POSITION_MASK) != D3DFVF_XYZB4 ) {
	  fvf = D3DFVF_XYZB4;
	}
	fvf |= D3DFVF_LASTBETA_UBYTE4; 
      } else {
	TRACE("invalid data format for D3DVSDE_BLENDINDICES\n");
      }
    }
     /* normal */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_NORMAL].type) ) {
      streamnum = vs->input_reg[D3DVSDE_NORMAL].stream;
      gl->normal.dwStride = priv->d3dp->stream[streamnum].stride;
      gl->normal.lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
			     + vs->input_reg[D3DVSDE_NORMAL].offset;
      fvf |= D3DFVF_NORMAL;
    }
    /* FIXME: psize */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_PSIZE].type) ) {
      FIXME("fixed function vertex declaration uses D3DVSDE_PSIZE -- not implemented\n");
      /* fvf |= D3DFVF_PSIZ; */
    }
    /* diffuse */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_DIFFUSE].type) ) {
      streamnum = vs->input_reg[D3DVSDE_DIFFUSE].stream;
      gl->color[0].dwStride = priv->d3dp->stream[streamnum].stride;
      gl->color[0].lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
			     + vs->input_reg[D3DVSDE_DIFFUSE].offset;
      fvf |= D3DFVF_DIFFUSE; 
    }
    /* specular */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_SPECULAR].type) ) {
      streamnum = vs->input_reg[D3DVSDE_SPECULAR].stream;
      gl->color[1].dwStride = priv->d3dp->stream[streamnum].stride;
      gl->color[1].lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
			     + vs->input_reg[D3DVSDE_SPECULAR].offset;
      fvf |= D3DFVF_SPECULAR; 
    }
    /* textures 0-7 */
    for( i=0; i<D3DDP_MAXTEXCOORD; i++ ) {
      if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_TEXCOORD0+i].type) ) {
	streamnum = vs->input_reg[D3DVSDE_TEXCOORD0+i].stream;
	gl->textureCoords[i].dwStride = priv->d3dp->stream[streamnum].stride;
	gl->textureCoords[i].lpvData = (LPVOID)priv->d3dp->stream[streamnum].data_t->gbl.fpVidMem
			       + vs->input_reg[D3DVSDE_TEXCOORD0+i].offset;
	switch(GET_INPUT_REG_TYPE(vs->input_reg[D3DVSDE_TEXCOORD0+i].type)) {
	case D3DVSDT_FLOAT1: fvf |= D3DFVF_TEXCOORDSIZE1(i); break;
	case D3DVSDT_FLOAT2: fvf |= D3DFVF_TEXCOORDSIZE2(i); break;
	case D3DVSDT_FLOAT3: fvf |= D3DFVF_TEXCOORDSIZE3(i); break;
	case D3DVSDT_FLOAT4: fvf |= D3DFVF_TEXCOORDSIZE4(i); break;
	default: TRACE("unhandled data format for D3DVSDE_TEXCOORD%d\n", i);
	}
	texcount++;
      }
    }
    fvf |= (texcount << D3DFVF_TEXCOUNT_SHIFT);
      
    /* FIXME: position2 -- used for vertex tweening */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_POSITION2].type) ) {
      FIXME("fixed function vertex declaration uses D3DVSDE_POSITION2 -- not implemented\n");
    }
    /* FIXME: normal2 -- used for vertex tweening */
    if( IS_INPUT_REG_ENABLED(vs->input_reg[D3DVSDE_NORMAL2].type) ) {
      FIXME("fixed function vertex declaration uses D3DVSDE_NORMAL2 -- not implemented\n");
    } 

    return fvf;
}

static void
GL_ConvertInterleavedToGLStride(DWORD fvf, LPCVOID lpVertices,
				LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    static const DWORD position_sizes[] = { 0, 3, 4, 4, 5, 6, 7, 8 };

    LPBYTE p = (LPBYTE)lpVertices;
    DWORD stride;
    int i;

    gl->position.lpvData = p;
    stride = position_sizes[(fvf & D3DFVF_POSITION_MASK) / 2]
	* sizeof(D3DVALUE);

    if (fvf & D3DFVF_NORMAL)
    {
	gl->normal.lpvData = p + stride;
	stride += 3 * sizeof(D3DVALUE);
    }

    if (fvf & D3DFVF_PSIZE) { 
        /* FIXME: point size, was D3DFVF_RESERVED1 before DX8 */
        stride += sizeof(D3DVALUE);
    }
      
    if (fvf & D3DFVF_DIFFUSE)
    {
	gl->color[0].lpvData = p + stride;
	stride += sizeof(D3DCOLOR);
    }

    if (fvf & D3DFVF_SPECULAR)
    {
	gl->color[1].lpvData = p + stride;
	stride += sizeof(D3DCOLOR);
    }

    /* XXX: This assumes 2D texture coordinates. */
    if (fvf & D3DFVF_TEXCOUNT_MASK)
    {
	for (i = 0; i < D3DDP_MAXTEXCOORD; i++)
	    gl->textureCoords[i].lpvData
		= p + stride + 2 * sizeof(D3DVALUE) * i;

	stride += 2 * D3DFVF_TEXCOUNT(fvf) * sizeof(D3DVALUE);
    }

    gl->position.dwStride = stride;
    gl->normal.dwStride = stride;
    gl->color[0].dwStride = stride;
    gl->color[1].dwStride = stride;
    for (i=0; i < D3DDP_MAXTEXCOORD; i++)
	gl->textureCoords[i].dwStride = stride;

    GL_SetupStrideBlendWeights(fvf, gl);
    GL_SetupStrideFogCoordinate(fvf, gl);
}

#ifdef DEBUG_VERTEX_DATA
static void D3D_GL_dump_vertex(GL_D3D_priv* priv, DWORD fvf, DWORD dwIndex)
{
    LPGLDRAWPRIMITIVESTRIDEDDATA stride = &priv->stride;
    D3DVALUE *x;
    DWORD *c;
    int i;

#define STRIDE_DATA(x) \
  (LPVOID)((LPBYTE)stride->x.lpvData + dwIndex * stride->x.dwStride)

    x = STRIDE_DATA(position);
    if ((fvf & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW)
      TRACE("  XYZRHW: %f, %f, %f, %f\n",  x[0], x[1], x[2], x[3]);
    else {
      TRACE("  XYZ   : %f, %f, %f\n",  x[0], x[1], x[2]);
      /* FIXME: blend weights */
    }

    if (fvf & D3DFVF_NORMAL) {
      x = STRIDE_DATA(normal);
      TRACE("  normal: %f, %f, %f\n",  x[0], x[1], x[2]);
    }
    if (fvf & D3DFVF_DIFFUSE) {
      c = STRIDE_DATA(color[0]);
      TRACE("  diffuse : %08lx\n", *c);
    }
    if (fvf & D3DFVF_SPECULAR) {
      c = STRIDE_DATA(color[1]);
      TRACE("  specular: %08lx\n", *c);
    }

    if (fvf & D3DFVF_TEXCOUNT_MASK) {
      for (i = 0; i < D3DFVF_TEXCOUNT(fvf); i++) {
        x = STRIDE_DATA(textureCoords[i]);
	/* assume 2D texture coordinates */
        TRACE("  tex-stage %d: %f, %f\n", i, x[0], x[1]);
      }
    }
}
#endif

DWORD D3D_GL_setup_fixed(GL_D3D_priv* priv, LPVOID data1, 
                         const WORD* indices, GLenum pt, DWORD dwCount, BOOL indexed)
{
    DWORD fvf;
    DWORD dwIndexBase;
    DWORD dwVStart, dwVCount;

    fvf = priv->dp2_dwVertexType;
    dwIndexBase = indexed ? ((LPD3DHAL_DP2STARTVERTEX)data1)->wVStart : 0;
    dwVStart = priv->dp2_dwVertexOffset;
    dwVCount = priv->dp2_dwVertexLength;

    /* Compute the initial GL stride. */
    if (priv->glFlags & GLF_VERTEX_DECLARATION)
    {
	fvf = GL_VertexDeclarationToGLStride(priv, &priv->stride);
    }
    else if (priv->glFlags & GLF_STRIDED_DATA)
    {
	GL_ConvertD3DStrideToGLStride(fvf, &priv->sd, &priv->stride);
    }
    else if (priv->glFlags & GLF_VERTEX_BUFFER)
    {
	const struct GL_D3D_DriverCallbacks *cb = &priv->callbacks;

	if (cb->D3dVbGetConvertedFVF)
	{
	    DWORD converted_fvf;
	    BOOL do_conversion;

	    converted_fvf = cb->D3dVbGetConvertedFVF(priv->dp2_VertexBuffer);

	    do_conversion = GL_IsDirectlyRenderable(priv, converted_fvf);

	    if (do_conversion)
	    {
		if (GL_CHECK_NEW_TNL(fvf, converted_fvf))
		    priv->glFlags |= GLF_NEW_TNL_MODE;

		fvf = priv->dp2_dwVertexType = converted_fvf;
	    }

	    cb->D3dVbPreRender(priv->dp2_VertexBuffer,
			       &priv->stride,
			       do_conversion,
			       dwVStart,
			       dwVCount,
			       indices, dwCount,
			       dwIndexBase);
	}
	else
	{
	    GL_ConvertInterleavedToGLStride(fvf,
					    (LPVOID)priv->dp2_VertexBuffer->lpGbl->fpVidMem,
					    &priv->stride);
	}
    }
    else
    {
	GL_ConvertInterleavedToGLStride(fvf,
					priv->dp2_lpVertices,
					&priv->stride);
    }

    TRACE("fvf=%lx, ibase=%ld, vstart=%ld, vcount=%ld\n", fvf, dwIndexBase, dwVStart, dwVCount);
    if (pt == GL_POINTS)
    {
        /* FIXME: debug dumps for points */
    }
    else if (indexed)
    {
	LPD3DHAL_DP2STARTVERTEX dat = (LPD3DHAL_DP2STARTVERTEX)data1;

	TRACE("start=%d, count=%ld [indexed]\n", dat->wVStart, dwCount);
#ifdef DEBUG_VERTEX_DATA
	{
	    DWORD i;
	    for (i=0; i<dwCount; i++) {
		WORD idx = indices[i];
		TRACE(" index%03ld: %03d\n", i, idx);
		D3D_GL_dump_vertex(priv, fvf, dwIndexBase + idx);
	    }
	}
#endif
    }
    else
    {
	LPD3DHAL_DP2STARTVERTEX dat = (LPD3DHAL_DP2STARTVERTEX)data1;

	TRACE("start=%d, count=%ld\n", dat->wVStart, dwCount);
#ifdef DEBUG_VERTEX_DATA
	{
	    DWORD i;
	    for (i=0; i<dwCount; i++) {
		D3D_GL_dump_vertex(priv, fvf, dwVStart + dat->wVStart + i);
	    }
	}
#endif
    }

    /* We have the raw D3D arrays in the GL stride structure.
     * Now we have to replace the arrays that are not supported by OpenGL. */

    /* FIXME: recompute vertex materials only on (NEW_LIGHTING | NEW_TNL_MODE) ? */
    fvf = GL_fixup_vertex_materials(priv, fvf, &priv->stride);
    GL_fixup_position_array(priv, fvf, dwVStart, dwVCount,
			    &priv->stride, indices, dwCount, dwIndexBase);
    GL_fixup_color_arrays(priv, fvf, dwVStart, dwVCount,
			  &priv->stride, indices, dwCount, dwIndexBase);

    GL_load_arrays(priv, fvf, dwIndexBase, &priv->stride);

    GL_setup_transforms(priv, fvf);
    GL_setup_lighting(priv, fvf);

    priv->glFlags &= ~(GLF_NEW_TNL_MODE | GLF_NEW_LIGHTING);

    CHECK_GL_ERROR("setup_fixed");
    return fvf;
}

void D3D_GL_finish_fixed(GL_D3D_priv* priv)
{
    if ((priv->glFlags & GLF_VERTEX_BUFFER)
	&& priv->callbacks.D3dVbPostRender)
	priv->callbacks.D3dVbPostRender(priv->dp2_VertexBuffer);
}
