/*	Direct3D OpenGL driver
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <math.h> /* for M_PI */

#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define GL_LIGHT_DATA_CHANGE 0x10
#define GL_LIGHT_VIEW_CHANGE 0x20

static void GL_LoadLightPosition(D3D_Light *light)
{
    GLenum gl = GL_LIGHT0 + light->devnum;

    TRACE("positioning GL light %lu\n", light->devnum);
    glLightfv(gl, GL_POSITION, (const float *)&light->devpos);
    if ((light->l.dltType == D3DLIGHT_SPOT) || (light->l.dltType == D3DLIGHT_GLSPOT))
	glLightfv(gl, GL_SPOT_DIRECTION, (const float *)&(light->l.dvDirection));

    light->flags &= ~GL_LIGHT_VIEW_CHANGE;
}

void GL_LoadLightPositions(GL_D3D_priv* priv, BOOL force)
{
    DWORD fl;
    unsigned u;
    LPD3D_P d3dp = priv->d3dp;

    fl = D3DLIGHT_ACTIVE;
    if (!force) fl |= GL_LIGHT_VIEW_CHANGE;

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadMatrixf((const float *)&d3dp->xfrm[D3DTRANSFORMSTATE_VIEW]);
    for (u=0; u<d3dp->numlights; u++)
	if ((d3dp->light[u].flags & fl) == fl)
	    GL_LoadLightPosition(&d3dp->light[u]);
    glPopMatrix();
}

static void GL_LoadLightData(D3D_Light *light)
{
    GLenum gl = GL_LIGHT0 + light->devnum;

    TRACE("loading GL light %lu\n", light->devnum);
    glLightfv(gl, GL_DIFFUSE, (float *)&(light->l.dcvDiffuse));
    glLightfv(gl, GL_SPECULAR, (float *)&(light->l.dcvSpecular));
    glLightfv(gl, GL_AMBIENT, (float *)&(light->l.dcvAmbient));

    /* I can't find a GL parameter for dvRange, and no exact match for dvFalloff */
    glLightf(gl, GL_CONSTANT_ATTENUATION, light->l.dvAttenuation0);
    glLightf(gl, GL_LINEAR_ATTENUATION, light->l.dvAttenuation1);
    glLightf(gl, GL_QUADRATIC_ATTENUATION, light->l.dvAttenuation2);
    if (light->l.dltType == D3DLIGHT_SPOT) {
	float theta, ctheta, phi, cphi, pos, cpos, lit;
	theta = light->l.dvTheta / 2;
	phi = light->l.dvPhi / 2;
	if (theta < phi) {
	    /* Direct3D spotlights don't work exactly like OpenGL spotlights,
	     * but to approximate it as closely as possible, we'll map the
	     * dvTheta and dvFalloff to the OpenGL spot exponent by trying to
	     * make some point between dvTheta and dvPhi have an equal
	     * brightness in the two models */
	    /* choosing a point near phi makes the cutoff smoother */
	    pos = (phi - theta)*7/8 + theta;

	    ctheta = cos(theta);
	    cphi = cos(phi);
	    cpos = cos(pos);
	    /* Direct3D: light = ((cos(p)-cos(phi))/(cos(theta)-cos(phi)))^falloff */
	    lit = (cpos-cphi)/(ctheta-cphi);
	    /* ignore falloff, since we're just approximating anyway */
	    /* OpenGL: light = cos(p)^exponent */
	    /*   i.e.: exponent = ln light / ln cos(p) */
	    glLightf(gl, GL_SPOT_EXPONENT, log(lit) / log(cpos));
	} else glLightf(gl, GL_SPOT_EXPONENT, 0.0);
	glLightf(gl, GL_SPOT_CUTOFF, phi * 180 / M_PI);
    }
    else if (light->l.dltType == D3DLIGHT_GLSPOT) {
	/* not sure about this */
	glLightf(gl, GL_SPOT_EXPONENT, light->l.dvFalloff);
	glLightf(gl, GL_SPOT_CUTOFF, light->l.dvPhi * 90 / M_PI);
    } else {
	glLightf(gl, GL_SPOT_EXPONENT, 0.0);
	glLightf(gl, GL_SPOT_CUTOFF, 180);
    }
    light->flags |= GL_LIGHT_VIEW_CHANGE;
    light->flags &= ~GL_LIGHT_DATA_CHANGE;
}

HRESULT D3D_GL_create_light(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    return D3D_OK;
}

HRESULT D3D_GL_set_light(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2SETLIGHT dat = (LPD3DHAL_DP2SETLIGHT)data1;
    LPD3DLIGHT7 ldata = (LPD3DLIGHT7)data2;
    D3D_Light *light = &priv->d3dp->light[dat->dwIndex];

    switch (dat->dwDataType) {
    case D3DHAL_SETLIGHT_ENABLE:
	/* need to allocate GL light - is the previous allocation still free? */
	if (priv->lights & (1 << light->devnum)) {
	    /* it's taken, allocate another
	     * (hmm, we could use a better algorithm than this)
	     */
	    unsigned u;
	    for (u=0; u<32 && (priv->lights & (1 << u)); u++);
	    if (u >= priv->lpCaps->dwMaxActiveLights) {
		ERR("too many allocated lights, can't activate!\n");
		light->flags &= ~D3DLIGHT_ACTIVE;
		return D3D_OK;
	    }
	    light->devnum = u;
	    /* light now needs to be loaded */
	    light->flags |= GL_LIGHT_DATA_CHANGE;
	}
	else {
	    /* the position may have changed */
	    light->flags |= GL_LIGHT_VIEW_CHANGE;
	}
	priv->lights |= (1 << light->devnum);
	glEnable(GL_LIGHT0 + light->devnum);
	if (light->flags & GL_LIGHT_DATA_CHANGE)
	    GL_LoadLightData(light);
	break;
    case D3DHAL_SETLIGHT_DISABLE:
	glDisable(GL_LIGHT0 + light->devnum);
	priv->lights &= ~(1 << light->devnum);
	break;
    case D3DHAL_SETLIGHT_DATA:
	if (light->l.dltType == D3DLIGHT_DIRECTIONAL) {
	    light->devpos[0] = -ldata->dvDirection.x;
	    light->devpos[1] = -ldata->dvDirection.y;
	    light->devpos[2] = -ldata->dvDirection.z;
	    light->devpos[3] = 0.0;
	}
	else {
	    light->devpos[0] = ldata->dvPosition.x;
	    light->devpos[1] = ldata->dvPosition.y;
	    light->devpos[2] = ldata->dvPosition.z;
	    light->devpos[3] = 1.0;
	}
	light->flags |= GL_LIGHT_DATA_CHANGE;
	if (light->flags & D3DLIGHT_ACTIVE)
	    GL_LoadLightData(light);
	break;
    }
    return D3D_OK;
}
