/*	Direct3D OpenGL driver
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

static const GLenum GL_D3DCMP[] = {
    0, GL_NEVER, GL_LESS, GL_EQUAL, GL_LEQUAL, GL_GREATER,
    GL_NOTEQUAL, GL_GEQUAL, GL_ALWAYS
};

static const GLint GL_D3DFOGMODE[4] = {
    0, GL_EXP, GL_EXP2, GL_LINEAR
};

/* This is intended for D3DRENDERSTATE_SRCBLEND, D3DRENDERSTATE_DESTBLEND,
 * but has two problems. First, GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR are not
 * allowed source blend factors, and the corresponding destination-based
 * factors are not allowed destination blend factors, unless the
 * GL_NV_blend_square extension is present.
 * Second, we don't support the BOTH modes yet. */
static const GLenum GL_D3DBLEND[] = {
    0, GL_ZERO, GL_ONE, GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR, GL_SRC_ALPHA,
    GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA,
    GL_DST_COLOR, GL_ONE_MINUS_DST_COLOR, GL_SRC_ALPHA_SATURATE, 0, 0
};

static const GLenum GL_D3DBLENDOP[] = {
    0, GL_FUNC_ADD, GL_FUNC_SUBTRACT, GL_FUNC_REVERSE_SUBTRACT, GL_MIN, GL_MAX
};

static const GLenum GL_D3DSTENCILOP[] = {
    0, GL_KEEP, GL_ZERO, GL_REPLACE, GL_INCR, GL_DECR, GL_INVERT,
    GL_INCR_WRAP_EXT, GL_DECR_WRAP_EXT
};

static inline float GL_RSF(DWORD dwState)
{
    return *(float*)&dwState;
}

static inline void GL_RS2Col(GLfloat col[4], DWORD dwState)
{
    D3DCOLOR rcol = (D3DCOLOR)dwState;
    col[0] = RGBA_GETRED(rcol)   / 255.0;
    col[1] = RGBA_GETGREEN(rcol) / 255.0;
    col[2] = RGBA_GETBLUE(rcol)  / 255.0;
    col[3] = RGBA_GETALPHA(rcol) / 255.0;
}

static void GL_FogEnable(GL_D3D_priv* priv)
{
    LPD3D_P d3dp = priv->d3dp;
    if (d3dp->renderstate[D3DRENDERSTATE_FOGENABLE]) {
	if (d3dp->renderstate[D3DRENDERSTATE_FOGTABLEMODE]) {
	    TRACE("enabling table fog\n");
	    glEnable(GL_FOG);
	    glHint(GL_FOG_HINT, GL_NICEST);
	    glFogi(GL_FOG_MODE, GL_D3DFOGMODE[d3dp->renderstate[D3DRENDERSTATE_FOGTABLEMODE]]);
	}
	else if (d3dp->renderstate[D3DRENDERSTATE_FOGVERTEXMODE]) {
	    TRACE("enabling vertex fog\n");
	    glEnable(GL_FOG);
	    glHint(GL_FOG_HINT, GL_FASTEST);
	    glFogi(GL_FOG_MODE, GL_D3DFOGMODE[d3dp->renderstate[D3DRENDERSTATE_FOGVERTEXMODE]]);
	}
	else
	    glDisable(GL_FOG);
    }
    else glDisable(GL_FOG);
}

HRESULT D3D_GL_sync_render_state(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2RENDERSTATE dat = (LPD3DHAL_DP2RENDERSTATE)data1;
    LPD3D_P d3dp = priv->d3dp;

    TRACE("state %d, value %08lx\n", dat->RenderState, dat->u.dwState);

    switch (dat->RenderState) {
    /* ANTIALIAS */
    case D3DRENDERSTATE_TEXTUREPERSPECTIVE:
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, dat->u.dwState ? GL_NICEST : GL_FASTEST);
	break;
    case D3DRENDERSTATE_ZENABLE:
	if (dat->u.dwState) glEnable(GL_DEPTH_TEST);
	else glDisable(GL_DEPTH_TEST);
	break;
    case D3DRENDERSTATE_FILLMODE:
#ifndef WIREFRAME
	switch (dat->u.dwState) {
	case D3DFILL_POINT:	glPolygonMode(GL_FRONT_AND_BACK, GL_POINT); break;
	case D3DFILL_WIREFRAME:	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); break;
	case D3DFILL_SOLID:	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); break;
	default: FIXME("Unknown fill mode: %ld\n", dat->u.dwState);
	}
#else
        if (!gWireframe)
        {
	    switch (dat->u.dwState) {
	    case D3DFILL_POINT:	        glPolygonMode(GL_FRONT_AND_BACK, GL_POINT); break;
	    case D3DFILL_WIREFRAME:	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); break;
	    case D3DFILL_SOLID:   	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); break;
	    default: FIXME("Unknown fill mode: %ld\n", dat->u.dwState);
            }
	}
        else
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
#endif
	break;
    case D3DRENDERSTATE_SHADEMODE:
	switch (dat->u.dwState) {
	case D3DSHADE_FLAT:	glShadeModel(GL_FLAT); break;
	case D3DSHADE_GOURAUD:	glShadeModel(GL_SMOOTH); break;
	default: FIXME("Unknown shading model: %ld\n", dat->u.dwState);
	}
	break;
    /* LINEPATTERN, MONOENABLE, ROP2, PLANEMASK */
    case D3DRENDERSTATE_ZWRITEENABLE:
	glDepthMask(dat->u.dwState);
	break;
    case D3DRENDERSTATE_ALPHATESTENABLE:
	if (dat->u.dwState) glEnable(GL_ALPHA_TEST);
	else glDisable(GL_ALPHA_TEST);
	break;
    case D3DRENDERSTATE_LASTPIXEL:
	if (!dat->u.dwState) {
	    /* how to do this? */
	    WARN("disable last pixel\n");
	}
        break;
    case D3DRENDERSTATE_SRCBLEND:
    case D3DRENDERSTATE_DESTBLEND:
	glBlendFunc(GL_D3DBLEND[d3dp->renderstate[D3DRENDERSTATE_SRCBLEND]],
		    GL_D3DBLEND[d3dp->renderstate[D3DRENDERSTATE_DESTBLEND]]);
	break;
    /* TEXTUREMAPBLEND */
    case D3DRENDERSTATE_CULLMODE:
	switch (dat->u.dwState) {
	case D3DCULL_NONE:	glDisable(GL_CULL_FACE); break;
	case D3DCULL_CW:	glEnable(GL_CULL_FACE); glCullFace(GL_FRONT); break;
	case D3DCULL_CCW:	glEnable(GL_CULL_FACE); glCullFace(GL_BACK); break;
	default: FIXME("Unknown cull mode: %ld\n", dat->u.dwState);
	}
	break;
    case D3DRENDERSTATE_ZFUNC:
	glDepthFunc(GL_D3DCMP[dat->u.dwState]);
	break;
    case D3DRENDERSTATE_ALPHAREF:
    case D3DRENDERSTATE_ALPHAFUNC:
	glAlphaFunc(GL_D3DCMP[d3dp->renderstate[D3DRENDERSTATE_ALPHAFUNC]],
		    (d3dp->renderstate[D3DRENDERSTATE_ALPHAREF] & 0xFF) / 255.0);
	break;
    case D3DRENDERSTATE_DITHERENABLE:
	if (dat->u.dwState) glEnable(GL_DITHER);
	else glDisable(GL_DITHER);
	break;
    case D3DRENDERSTATE_ALPHABLENDENABLE:
	if (dat->u.dwState) glEnable(GL_BLEND);
	else glDisable(GL_BLEND);
	break;
    case D3DRENDERSTATE_FOGENABLE:
	GL_FogEnable(priv);
	break;
    case D3DRENDERSTATE_SPECULARENABLE:
	priv->glFlags |= GLMF_SPECULAR_DIRTY;
	break;
    /* ZVISIBLE, SUBPIXEL, SUBPIXELX, SUBPIXELY, STIPPLEDALPHA */
    case D3DRENDERSTATE_FOGCOLOR:
	{
	    GLfloat col[4];
	    GL_RS2Col(col, dat->u.dwState);
	    glFogfv(GL_FOG_COLOR, col);
	}
	break;
    case D3DRENDERSTATE_FOGTABLEMODE:
	GL_FogEnable(priv);
	break;
    case D3DRENDERSTATE_FOGSTART:
	glFogf(GL_FOG_START, dat->u.dvState);
	break;
    case D3DRENDERSTATE_FOGEND:
	glFogf(GL_FOG_END, dat->u.dvState);
	break;
    case D3DRENDERSTATE_FOGDENSITY:
	glFogf(GL_FOG_DENSITY, dat->u.dvState);
	break;
    /* STIPPLEENABLE, EDGEANTIALIAS, COLORKEYENABLE, BORDERCOLOR,
       TEXTUREADDRESSU, TEXTUREADDRESSV, MIPMAPLODBIAS */
    case D3DRENDERSTATE_ZBIAS:
	glPolygonOffset(0, -(GLfloat)dat->u.dwState);
	if (dat->u.dwState) glEnable(GL_POLYGON_OFFSET_FILL);
	else glDisable(GL_POLYGON_OFFSET_FILL);
	break;
    /* RANGEFOGENABLE, ANISOTROPY, FLUSHBATCH */
    case D3DRENDERSTATE_STENCILENABLE:
	if (dat->u.dwState) glEnable(GL_STENCIL_TEST);
	else glDisable(GL_STENCIL_TEST);
	break;
    case D3DRENDERSTATE_STENCILFAIL:
    case D3DRENDERSTATE_STENCILZFAIL:
    case D3DRENDERSTATE_STENCILPASS:
	glStencilOp(GL_D3DSTENCILOP[d3dp->renderstate[D3DRENDERSTATE_STENCILFAIL]],
		    GL_D3DSTENCILOP[d3dp->renderstate[D3DRENDERSTATE_STENCILZFAIL]],
		    GL_D3DSTENCILOP[d3dp->renderstate[D3DRENDERSTATE_STENCILPASS]]);
	break;
    case D3DRENDERSTATE_STENCILFUNC:
    case D3DRENDERSTATE_STENCILREF:
    case D3DRENDERSTATE_STENCILMASK:
	glStencilFunc(GL_D3DCMP[d3dp->renderstate[D3DRENDERSTATE_STENCILFUNC]],
		      d3dp->renderstate[D3DRENDERSTATE_STENCILREF],
		      d3dp->renderstate[D3DRENDERSTATE_STENCILMASK]);
	break;
    case D3DRENDERSTATE_STENCILWRITEMASK:
	glStencilMask(dat->u.dwState);
	break;
    case D3DRENDERSTATE_TEXTUREFACTOR:
	{
	    GLfloat col[4];
	    GL_RS2Col(col, dat->u.dwState);
	    GL_LoadTextureFactor(priv, col);
	}
	break;
    /* STIPPLEPATTERN00-31, WRAP0-7 */
    case D3DRENDERSTATE_CLIPPING:
	if (HAVE_GL_EXT(priv, EXT_clip_volume_hint))
	    glHint(GL_CLIP_VOLUME_CLIPPING_HINT_EXT,
		   dat->u.dwState ? GL_NICEST : GL_FASTEST);
	break;
    case D3DRENDERSTATE_LIGHTING:
	/* checked in glprim.c */
	priv->glFlags |= GLF_NEW_LIGHTING;
	break;
    case D3DRENDERSTATE_EXTENTS:
	/* Handled by HAL. */
	break;
    case D3DRENDERSTATE_AMBIENT:
	{
	    GLfloat col[4];
	    GL_RS2Col(col, dat->u.dwState);
	    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, col);
	}
	break;
    case D3DRENDERSTATE_FOGVERTEXMODE:
	GL_FogEnable(priv);
	break;
    case D3DRENDERSTATE_COLORVERTEX:
	/* checked in glprim.c */
	priv->glFlags |= GLF_NEW_LIGHTING;
	break;
    case D3DRENDERSTATE_LOCALVIEWER:
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, dat->u.dwState);
	break;
    case D3DRENDERSTATE_NORMALIZENORMALS:
	if (dat->u.dwState) glEnable(GL_NORMALIZE);
	else glDisable(GL_NORMALIZE);
	break;
    /* COLORKEYBLENDENABLE */
    case D3DRENDERSTATE_DIFFUSEMATERIALSOURCE:
    case D3DRENDERSTATE_SPECULARMATERIALSOURCE:
    case D3DRENDERSTATE_AMBIENTMATERIALSOURCE:
    case D3DRENDERSTATE_EMISSIVEMATERIALSOURCE:
	/* checked in glprim.c */
	priv->glFlags |= GLF_NEW_LIGHTING;
	break;
    case D3DRENDERSTATE_VERTEXBLEND:
	/* checked in glprim.c */
	priv->glFlags |= GLF_NEW_TNL_MODE;
	break;
    case D3DRENDERSTATE_CLIPPLANEENABLE:
	{
	    unsigned u;
	    for (u=0; u<priv->lpCaps->wMaxUserClipPlanes; u++)
		if (dat->u.dwState & (1 << u))
		    glEnable(GL_CLIP_PLANE0 + u);
		else
		    glDisable(GL_CLIP_PLANE0 + u);
	}
	break;
    /* Direct3D 8 states */
    /* SOFTWAREVERTEXPROCESSING, POINTSIZE, POINTSIZE_MIN, POINTSPRITEENABLE,
     * POINTSCALEENABLE, POINTSCALE_A, POINTSCALE_B, POINTSCALE_C,
     * MULTISAMPLEANTIALIAS, MULTISAMPLEMASK, PATCHEDGESTYLE, PATCHSEGMENTS,
     * DEBUGMONITORTOKEN, POINTSIZE_MAX, INDEXEDVERTEXBLENDENABLE */
    case D3DRS_COLORWRITEENABLE:
	glColorMask((dat->u.dwState & D3DCOLORWRITEENABLE_RED)   != 0,
		    (dat->u.dwState & D3DCOLORWRITEENABLE_GREEN) != 0,
		    (dat->u.dwState & D3DCOLORWRITEENABLE_BLUE)  != 0,
		    (dat->u.dwState & D3DCOLORWRITEENABLE_ALPHA) != 0);
	break;
    /* TWEENFACTOR */
    case D3DRS_BLENDOP:
	if (HAVE_GL_EXT(priv, ARB_imaging))
	    glBlendEquation(GL_D3DBLENDOP[dat->u.dwState]);
	break;
    /* POSITIONORDER, NORMALORDER */
    default:
	FIXME("unsupported render state %d\n", dat->RenderState);
	break;
    }
    CHECK_GL_ERROR("RenderState");
    return D3D_OK;
}

HRESULT D3D_GL_async_render_state(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2RENDERSTATE dat = (LPD3DHAL_DP2RENDERSTATE)data1;
    LPD3D_P d3dp = priv->d3dp;

    TRACE("state %d, value %08lx\n", dat->RenderState, dat->u.dwState);

    d3dp->renderstate[dat->RenderState] = dat->u.dwState;

    /* FIXME: here we can check render state for validity */

    return D3D_OK;
}

HRESULT D3D_GL_viewport_info(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2VIEWPORTINFO dat = (LPD3DHAL_DP2VIEWPORTINFO)data1;
    GLint gl_y;
    TRACE("X: %ld, Y: %ld, Width: %ld, Height: %ld\n",
	  dat->dwX, dat->dwY, dat->dwWidth, dat->dwHeight);

    /* D3D uses MM_TEXT coordinates. GL has 0,0 in the bottom left,
     * with positive Y going up, and it wants the bottom left corner
     * of the viewport. */
#ifndef HAVE_OSMESA
    gl_y = priv->d3dp->surf->lpGbl->wHeight - (dat->dwY + dat->dwHeight);
#else
    gl_y = dat->dwY;
#endif

    glScissor(dat->dwX, gl_y, dat->dwWidth, dat->dwHeight);
    glViewport(dat->dwX, gl_y, dat->dwWidth, dat->dwHeight);
    priv->glFlags |= GLF_NEW_TRANSFORM;
    return D3D_OK;
}

HRESULT D3D_GL_z_range(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2ZRANGE dat = (LPD3DHAL_DP2ZRANGE)data1;
    TRACE("Zmin: %f, Zmax: %f\n", dat->dvMinZ, dat->dvMaxZ);
    glDepthRange(dat->dvMinZ, dat->dvMaxZ);
    /* priv->glFlags |= GLF_NEW_TRANSFORM; */
    return D3D_OK;
}

HRESULT D3D_GL_set_material(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DMATERIAL7 dat = (LPD3DMATERIAL7)data1;
    LPD3D_P d3dp = priv->d3dp;

    if (memcmp(&d3dp->material.u.diffuse, &dat->u.diffuse, sizeof(D3DCOLORVALUE)))
    {
	memcpy(&d3dp->material.u.diffuse, &dat->u.diffuse, sizeof(D3DCOLORVALUE));
	priv->material_flags |= GLMF_DIFFUSE_DIRTY;
    }

    if (memcmp(&d3dp->material.u1.ambient, &dat->u1.ambient, sizeof(D3DCOLORVALUE)))
    {
	memcpy(&d3dp->material.u1.ambient, &dat->u1.ambient, sizeof(D3DCOLORVALUE));
	priv->material_flags |= GLMF_AMBIENT_DIRTY;
    }

    if (memcmp(&d3dp->material.u2.specular, &dat->u2.specular, sizeof(D3DCOLORVALUE)))
    {
	memcpy(&d3dp->material.u2.specular, &dat->u2.specular, sizeof(D3DCOLORVALUE));
	priv->material_flags |= GLMF_SPECULAR_DIRTY;
    }

    if (memcmp(&d3dp->material.u3.emissive, &dat->u3.emissive, sizeof(D3DCOLORVALUE)))
    {
	memcpy(&d3dp->material.u3.emissive, &dat->u3.emissive, sizeof(D3DCOLORVALUE));
	priv->material_flags |= GLMF_EMISSIVE_DIRTY;
    }

    if (dat->u4.power != d3dp->material.u4.power)
    {
	d3dp->material.u4.power = dat->u4.power;
	priv->material_flags |= GLMF_SHININESS_DIRTY;
    }

    return D3D_OK;
}

HRESULT D3D_GL_set_transform(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2SETTRANSFORM dat = (LPD3DHAL_DP2SETTRANSFORM)data1;

    priv->glFlags |= GLF_NEW_TRANSFORM;
    if (dat->xfrmType == D3DTRANSFORMSTATE_VIEW) priv->glFlags |= GLF_NEW_VIEW;

    if (dat->xfrmType >= D3DTRANSFORMSTATE_TEXTURE0
	&& dat->xfrmType <= D3DTRANSFORMSTATE_TEXTURE7)
    {
	priv->glFlags |= GLF_NEW_TEXCOORD_TRANSFORM;
	priv->texture_flags &= ~GLTF_DEFAULT_TEXCOORD_TRANSFORM;
    }

    /* the real work is done in glprim.c */
    return D3D_OK;
}

HRESULT D3D_GL_set_clip_plane(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2SETCLIPPLANE dat = (LPD3DHAL_DP2SETCLIPPLANE)data1;
    GLdouble plane[4];

    plane[0] = dat->plane[0];
    plane[1] = dat->plane[1];
    plane[2] = dat->plane[2];
    plane[3] = dat->plane[3];

    glClipPlane(GL_CLIP_PLANE0 + dat->dwIndex, plane);
    return D3D_OK;
}

