/*		DirectInput Joystick device
 *
 * Copyright 1998 Marcus Meissner
 * Copyright 1998,1999 Lionel Ulmer
 * Copyright 2000-2002 TransGaming Technologies Inc.
 *
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#ifdef HAVE_SYS_ERRNO_H
# include <sys/errno.h>
#endif

#include "debugtools.h"
#include "winbase.h"
#include "winerror.h"
#include "windef.h"
#include "mmsystem.h"
#include "dinput.h"

#include "dinput_private.h"
#include "device_private.h"

DEFAULT_DEBUG_CHANNEL(dinput);

/* Wine joystick driver object instances */
#define WINE_JOYSTICK_AXIS_BASE   0
#define WINE_JOYSTICK_POV_BASE    8
#define WINE_JOYSTICK_BUTTON_BASE 16

typedef struct {
	LONG lMin;
	LONG lMax;
	LONG lDeadZone;
	LONG lSaturation;
} ObjProps;

typedef struct JoystickAImpl JoystickAImpl;
static ICOM_VTABLE(IDirectInputDevice8A) JoystickAvt;
struct JoystickAImpl
{
        LPVOID                          lpVtbl;
        DWORD                           ref;

	/* The 'parent' DInput */
	IDirectInputImpl *dinput;

	/* joystick private */
	DWORD				dwID;
	LPDIDATAFORMAT			drvdf;
	int*				offset_array;
	BYTE*				buffer;
	int				acquired;
	LPDIDATAFORMAT			df;
	DataFormat*			wine_df;
	ObjProps*			props;
        HANDLE				hEvent;
        LPDIDEVICEOBJECTDATA 		data_queue;
        int				queue_head, queue_tail, queue_len;
};

static DWORD js_count;
static HMODULE hJS;
static HRESULT (WINAPI *DIJOY_GetGuid)(DWORD dwID, LPGUID lpGuid);
static HRESULT (WINAPI *DIJOY_GetDevInfo)(DWORD dwID, DWORD dwVersion, LPDIDEVCAPS lpCaps, LPDIDEVICEINSTANCEA lpInfo);
static HRESULT (WINAPI *DIJOY_OpenDIDev)(DWORD dwID, DWORD dwVersion);
static HRESULT (WINAPI *DIJOY_CloseDIDev)(DWORD dwID);
static HRESULT (WINAPI *DIJOY_GetDIDataFormat)(DWORD dwID, LPDIDATAFORMAT* pdf);
static HRESULT (WINAPI *DIJOY_GetDIData)(DWORD dwID, BOOL req, DWORD* dwObj, DWORD* dwData, DWORD* dwTimeStamp);

static BOOL joydev_load(void)
{
  if (!hJS) {
    /* this will load the winmm joystick driver */
    js_count = joyGetNumDevs();
    TRACE("js_count=%ld\n", js_count);
    /* grab handle to driver */
    hJS = GetModuleHandleA("joystick.drv");
    if (!hJS) {
      ERR("couldn't load joystick driver\n");
      return FALSE;
    }
    /* load dinput entry points */
    DIJOY_GetGuid = (LPVOID)GetProcAddress(hJS, "dijoy_GetGuid");
    DIJOY_GetDevInfo = (LPVOID)GetProcAddress(hJS, "dijoy_GetDevInfo");
    DIJOY_OpenDIDev = (LPVOID)GetProcAddress(hJS, "dijoy_OpenDIDev");
    DIJOY_CloseDIDev = (LPVOID)GetProcAddress(hJS, "dijoy_CloseDIDev");
    DIJOY_GetDIDataFormat = (LPVOID)GetProcAddress(hJS, "dijoy_GetDIDataFormat");
    DIJOY_GetDIData = (LPVOID)GetProcAddress(hJS, "dijoy_GetDIData");
    if (!DIJOY_GetGuid) return FALSE;
  }
  return TRUE;
}

static BOOL joydev_enum_device(IDirectInputImpl *dinput, DWORD dwDevType, DWORD dwFlags, LPDIDEVICEINSTANCEA lpddi, DWORD dwID)
{
  if (!joydev_load())
    return FALSE;

  if (dwID >= js_count)
    return FALSE;

  if (dwFlags & DIEDFL_FORCEFEEDBACK)
    return FALSE;

  if ((dwDevType==0) || (GET_DIDEVICE_TYPE(dwDevType)==DIDEVTYPE_JOYSTICK) ||
      (GET_DIDEVICE_TYPE(dwDevType)==DI8DEVTYPE_JOYSTICK)) {
    if (SUCCEEDED(DIJOY_GetDevInfo(dwID, dinput->ver, NULL, lpddi))) {
      TRACE("Enumerating the Joystick%ld device\n", dwID);
      return TRUE;
    }
  }

  return FALSE;
}

static JoystickAImpl *alloc_device(DWORD dwID, REFGUID rguid, LPVOID jvt, IDirectInputImpl *dinput)
{
  JoystickAImpl* newDevice;
  unsigned u;

  newDevice = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(JoystickAImpl));
  newDevice->lpVtbl = jvt;
  newDevice->ref = 1;
  newDevice->dinput = dinput;
  newDevice->dwID = dwID;

  DIJOY_GetDIDataFormat(newDevice->dwID, &newDevice->drvdf);

  TRACE("(drvdf.dwDataSize=%ld)\n",newDevice->drvdf->dwDataSize);
  TRACE("(drvdf.dwNumObjs=%ld)\n",newDevice->drvdf->dwNumObjs);
  for (u=0; u<newDevice->drvdf->dwNumObjs; u++) {
    TRACE("drvdf.rgodf[%d].guid %s (%p)\n",u,debugstr_guid(newDevice->drvdf->rgodf[u].pguid), newDevice->drvdf->rgodf[u].pguid);
    TRACE("drvdf.rgodf[%d].dwOfs %ld\n",u,newDevice->drvdf->rgodf[u].dwOfs);
    TRACE("dwType 0x%02x,dwInstance %d\n",DIDFT_GETTYPE(newDevice->drvdf->rgodf[u].dwType),DIDFT_GETINSTANCE(newDevice->drvdf->rgodf[u].dwType));
    TRACE("drvdf.rgodf[%d].dwFlags 0x%08lx\n",u,newDevice->drvdf->rgodf[u].dwFlags);
  }

  newDevice->offset_array = HeapAlloc(GetProcessHeap(), 0, sizeof(int)*newDevice->drvdf->dwNumObjs);
  for (u=0; u<newDevice->drvdf->dwNumObjs; u++)
    newDevice->offset_array[u] = newDevice->drvdf->rgodf[u].dwOfs;
  newDevice->buffer = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, newDevice->drvdf->dwDataSize);
  newDevice->df = newDevice->drvdf;

  newDevice->props = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(ObjProps)*newDevice->drvdf->dwNumObjs);
  for (u=0; u<newDevice->drvdf->dwNumObjs; u++) {
    if (newDevice->drvdf->rgodf[u].dwType & DIDFT_AXIS) {
      newDevice->props[u].lMin = 0x0000;
      newDevice->props[u].lMax = 0xFFFF;
      newDevice->props[u].lDeadZone = 1000; /* 10% */
      newDevice->props[u].lSaturation = 0; /* 0% */
    }
  }

  return newDevice;
}

static HRESULT joydev_create_device(IDirectInputImpl *dinput, REFGUID rguid, REFIID riid, LPDIRECTINPUTDEVICEA* pdev)
{
  int i;
  GUID joyGuid;

  if (!joydev_load())
    return DIERR_DEVICENOTREG;

  for (i = 0; i < js_count; i++) {
    if (DIJOY_GetGuid(i, &joyGuid) != DI_OK)
      continue;
    if ((IsEqualGUID(&GUID_Joystick,rguid)) ||
	(IsEqualGUID(&joyGuid,rguid))) {
      if ((riid == NULL) ||
	  IsEqualGUID(&IID_IDirectInputDeviceA,riid) ||
	  IsEqualGUID(&IID_IDirectInputDevice2A,riid) ||
	  IsEqualGUID(&IID_IDirectInputDevice7A,riid) ||
	  IsEqualGUID(&IID_IDirectInputDevice8A,riid)) {
	HRESULT hr = DIJOY_OpenDIDev(i, dinput->ver);
	if (FAILED(hr)) return hr;

	*pdev=(IDirectInputDeviceA*) alloc_device(i, rguid, &JoystickAvt, dinput);

	TRACE("Creating the Joystick%d device (%p)\n", i, *pdev);
	return DI_OK;
      } else
	return DIERR_NOINTERFACE;
    }
  }

  return DIERR_DEVICENOTREG;
}

static dinput_device joydev = {
  30,
  joydev_enum_device,
  joydev_create_device
};

DECL_GLOBAL_CONSTRUCTOR(joydev_register) { dinput_register_device(&joydev); }

/******************************************************************************
 *	Joystick
 */
static ULONG WINAPI JoystickAImpl_Release(LPDIRECTINPUTDEVICE8A iface)
{
  ICOM_THIS(JoystickAImpl,iface);

  This->ref--;
  if (This->ref)
    return This->ref;

  DIJOY_CloseDIDev(This->dwID);

  /* Free the data queue */
  if (This->data_queue != NULL)
    HeapFree(GetProcessHeap(),0,This->data_queue);

  /* Free the DataFormat */
  HeapFree(GetProcessHeap(), 0, This->offset_array);
  HeapFree(GetProcessHeap(), 0, This->buffer);
  if (This->df != This->drvdf)
    HeapFree(GetProcessHeap(), 0, This->df);

  HeapFree(GetProcessHeap(),0,This);
  return 0;
}

/******************************************************************************
  *   SetDataFormat : the application can choose the format of the data
  *   the device driver sends back with GetDeviceState.
  */
static HRESULT WINAPI JoystickAImpl_SetDataFormat(
	LPDIRECTINPUTDEVICE8A iface,LPCDIDATAFORMAT df
)
{
  ICOM_THIS(JoystickAImpl,iface);
  int i;

  TRACE("(this=%p,%p)\n",This,df);

  TRACE("(df.dwSize=%ld)\n",df->dwSize);
  TRACE("(df.dwObjsize=%ld)\n",df->dwObjSize);
  TRACE("(df.dwFlags=0x%08lx)\n",df->dwFlags);
  TRACE("(df.dwDataSize=%ld)\n",df->dwDataSize);
  TRACE("(df.dwNumObjs=%ld)\n",df->dwNumObjs);

  for (i=0;i<df->dwNumObjs;i++) {
    TRACE("df.rgodf[%d].guid %s (%p)\n",i,debugstr_guid(df->rgodf[i].pguid), df->rgodf[i].pguid);
    TRACE("df.rgodf[%d].dwOfs %ld\n",i,df->rgodf[i].dwOfs);
    TRACE("dwType 0x%02x,dwInstance %d\n",DIDFT_GETTYPE(df->rgodf[i].dwType),DIDFT_GETINSTANCE(df->rgodf[i].dwType));
    TRACE("df.rgodf[%d].dwFlags 0x%08lx\n",i,df->rgodf[i].dwFlags);
  }

  /* Store the new data format */
  This->df = HeapAlloc(GetProcessHeap(),0,df->dwSize);
  memcpy(This->df, df, df->dwSize);
  This->df->rgodf = HeapAlloc(GetProcessHeap(),0,df->dwNumObjs*df->dwObjSize);
  memcpy(This->df->rgodf,df->rgodf,df->dwNumObjs*df->dwObjSize);

  /* Prepare all the data-conversion filters */
  This->wine_df = create_DataFormat(This->drvdf, df, This->offset_array);

  return 0;
}

#define map_axis(val) (This->lMax == 65535 && This->lMin == 0) ? val+32768 : \
			((val+32768)*(This->lMax-This->lMin)/65536+This->lMin)

static void joy_polldev(JoystickAImpl *This, BOOL req, BOOL gen_event) {
    HRESULT hr;
    DWORD dwObj;
    DWORD dwData;
    DWORD dwTimeStamp;

    do {
	LPDIOBJECTDATAFORMAT odf;
	hr = DIJOY_GetDIData(This->dwID, req, &dwObj, &dwData, &dwTimeStamp);
	if (FAILED(hr)) return;
	if (hr == S_FALSE) return;
	odf = &This->drvdf->rgodf[dwObj];

	/* update state buffer */
	if (odf->dwType & DIDFT_BUTTON)
	    *(BYTE*)(This->buffer+odf->dwOfs) = dwData;
	else if (odf->dwType & DIDFT_AXIS) {
	    /* FIXME: cache these */
	    LONG pmin = This->props[dwObj].lMin;
	    LONG pmax = This->props[dwObj].lMax;
	    LONG h_prange = (pmax - pmin) / 2;
	    /* we assume a physical range of 0-65534 */
	    LONG h_range = 32767 - This->props[dwObj].lDeadZone*65534/20000;
	    DWORD dwOut;
	    /* FIXME: saturation */

	    if (dwData > 65534) dwData = 65534;

	    if (dwData < h_range)
		dwOut = pmin + dwData*h_prange/h_range;
	    else if (dwData > (65534-h_range))
		dwOut = pmax - (65534-dwData)*h_prange/h_range;
	    else
		dwOut = (pmin + pmax) / 2;

	    *(DWORD*)(This->buffer+odf->dwOfs) = dwOut;
	} else
	    *(DWORD*)(This->buffer+odf->dwOfs) = dwData;

	/* generate event */
	if (gen_event) {
	    GEN_EVENT(This->offset_array[dwObj],dwData,dwTimeStamp,(This->dinput->evsequence)++);
	}
    } while (TRUE);
}

/******************************************************************************
  *     Acquire : gets exclusive control of the joystick
  */
static HRESULT WINAPI JoystickAImpl_Acquire(LPDIRECTINPUTDEVICE8A iface)
{
    ICOM_THIS(JoystickAImpl,iface);

    TRACE("(this=%p)\n",This);
    if (This->acquired) return S_FALSE;
    /* grab initial state */
    joy_polldev(This, TRUE, FALSE);
    This->acquired = TRUE;
    return DI_OK;
}

/******************************************************************************
  *     Unacquire : frees the joystick
  */
static HRESULT WINAPI JoystickAImpl_Unacquire(LPDIRECTINPUTDEVICE8A iface)
{
    ICOM_THIS(JoystickAImpl,iface);

    TRACE("(this=%p)\n",This);
    if (!This->acquired) return S_FALSE;
    This->acquired = FALSE;
    return DI_OK;
}

/******************************************************************************
  *     GetDeviceState : returns the "state" of the joystick.
  *
  */
static HRESULT WINAPI JoystickAImpl_GetDeviceState(
	LPDIRECTINPUTDEVICE8A iface,DWORD len,LPVOID ptr
) {
    ICOM_THIS(JoystickAImpl,iface);

    TRACE("(this=%p,0x%08lx,%p)\n",This,len,ptr);
    if (!This->acquired)
        return DIERR_NOTACQUIRED;

    joy_polldev(This, FALSE, TRUE);
    fill_DataFormat(ptr, This->buffer, This->wine_df);
    return 0;
}

/******************************************************************************
  *     GetDeviceData : gets buffered input data.
  */
static HRESULT WINAPI JoystickAImpl_GetDeviceData(LPDIRECTINPUTDEVICE8A iface,
					      DWORD dodsize,
					      LPDIDEVICEOBJECTDATA dod,
					      LPDWORD entries,
					      DWORD flags
) {
  ICOM_THIS(JoystickAImpl,iface);

  FIXME("(%p)->(dods=%ld,entries=%ld,fl=0x%08lx),STUB!\n",This,dodsize,*entries,flags);

  if (!This->acquired)
    return DIERR_NOTACQUIRED;

  joy_polldev(This, FALSE, TRUE);
  if (flags & DIGDD_PEEK)
    FIXME("DIGDD_PEEK\n");

  if (dod == NULL) {
  } else {
  }
  return 0;
}

int find_object(JoystickAImpl *This, DWORD dwHow, DWORD dwObj)
{
    int obj;

    switch (dwHow) {
    case DIPH_DEVICE:
	TRACE(" - DEVICE\n");
	return -1;
    case DIPH_BYID:
        for (obj = 0; obj < This->drvdf->dwNumObjs; obj++)
            if (This->drvdf->rgodf[obj].dwType == dwObj)
                break;
        break;
    case DIPH_BYOFFSET:
        for (obj = 0; obj < This->drvdf->dwNumObjs; obj++)
            if (This->offset_array[obj] == dwObj)
                break;
        break;
    default:
	FIXME("unhandled dwHow=%ld\n", dwHow);
	obj = This->drvdf->dwNumObjs;
        break;
    }

    if (obj == This->drvdf->dwNumObjs) {
	TRACE(" - obj not found\n");
	return -1;
    }

    TRACE(" - obj %d\n", obj);
    return obj;
}

/******************************************************************************
  *     SetProperty : change input device properties
  */
static HRESULT WINAPI JoystickAImpl_SetProperty(LPDIRECTINPUTDEVICE8A iface,
					    REFGUID rguid,
					    LPCDIPROPHEADER ph)
{
  ICOM_THIS(JoystickAImpl,iface);
  int obj;

  TRACE("(this=%p,%s,%p)\n",This,debugstr_guid(rguid),ph);
  TRACE("ph.dwSize = %ld, ph.dwHeaderSize =%ld, ph.dwObj = %ld, ph.dwHow= %ld\n",ph->dwSize, ph->dwHeaderSize,ph->dwObj,ph->dwHow);

  obj = find_object(This, ph->dwHow, ph->dwObj);
  if (ph->dwHow != DIPH_DEVICE && obj < 0) return DIERR_OBJECTNOTFOUND;

  if (!HIWORD(rguid)) {
    switch ((DWORD)rguid) {
    case (DWORD) DIPROP_BUFFERSIZE: {
      LPCDIPROPDWORD	pd = (LPCDIPROPDWORD)ph;

      FIXME("buffersize = %ld\n",pd->dwData);
      break;
    }
    case (DWORD)DIPROP_RANGE: {
      LPCDIPROPRANGE	pr = (LPCDIPROPRANGE)ph;

      TRACE("proprange(%ld,%ld)\n",pr->lMin,pr->lMax);
      if (obj < 0) return DIERR_INVALIDPARAM;
      This->props[obj].lMin = pr->lMin;
      This->props[obj].lMax = pr->lMax;
      break;
    }
    case (DWORD)DIPROP_DEADZONE: {
      LPCDIPROPDWORD	pd = (LPCDIPROPDWORD)ph;

      TRACE("deadzone(%ld)\n",pd->dwData);
      if (obj < 0) return DIERR_INVALIDPARAM;
      This->props[obj].lDeadZone = pd->dwData;
      break;
    }
    default:
      FIXME("Unknown type %ld (%s)\n",(DWORD)rguid,debugstr_guid(rguid));
      break;
    }
  }
  return 0;
}

/******************************************************************************
  *     SetEventNotification : specifies event to be sent on state change
  */
static HRESULT WINAPI JoystickAImpl_SetEventNotification(
	LPDIRECTINPUTDEVICE8A iface, HANDLE hnd
) {
    ICOM_THIS(JoystickAImpl,iface);

    TRACE("(this=%p,0x%08lx)\n",This,(DWORD)hnd);
    This->hEvent = hnd;
    return DI_OK;
}

static HRESULT WINAPI JoystickAImpl_GetCapabilities(
	LPDIRECTINPUTDEVICE8A iface,
	LPDIDEVCAPS lpDIDevCaps)
{
    ICOM_THIS(JoystickAImpl,iface);

    return DIJOY_GetDevInfo(This->dwID, This->dinput->ver, lpDIDevCaps, NULL);
}

static HRESULT WINAPI JoystickAImpl_Poll(LPDIRECTINPUTDEVICE8A iface)
{
    ICOM_THIS(JoystickAImpl,iface);
    TRACE("(this=%p)\n",This);

    if (!This->acquired)
	return DIERR_NOTACQUIRED;
    joy_polldev(This, TRUE, TRUE);
    return DI_OK;
}

/******************************************************************************
  *     EnumObjects : enumerate the different buttons and axis...
  */
static HRESULT WINAPI JoystickAImpl_EnumObjects(
	LPDIRECTINPUTDEVICE8A iface,
	LPDIENUMDEVICEOBJECTSCALLBACKA lpCallback,
	LPVOID lpvRef,
	DWORD dwFlags)
{
  ICOM_THIS(JoystickAImpl,iface);
  DIDEVICEOBJECTINSTANCEA ddoi;
  unsigned u;

  TRACE("(this=%p,%p,%p,%08lx)\n", This, lpCallback, lpvRef, dwFlags);
  if (TRACE_ON(dinput)) {
    DPRINTF("  - flags = ");
    _dump_EnumObjects_flags(dwFlags);
    DPRINTF("\n");
  }

  /* Only the fields till dwFFMaxForce are relevant */
  ddoi.dwSize = FIELD_OFFSET(DIDEVICEOBJECTINSTANCEA, dwFFMaxForce);

  for (u=0; u<This->drvdf->dwNumObjs; u++) {
    LPDIOBJECTDATAFORMAT odf = &This->drvdf->rgodf[u];
    if ((odf->dwType & dwFlags) || (dwFlags == DIDFT_ALL)) {
      memcpy(&ddoi.guidType, odf->pguid, sizeof(GUID));
      ddoi.dwOfs = odf->dwOfs;
      ddoi.dwType = odf->dwType;
      ddoi.dwFlags = odf->dwFlags;
      if (ddoi.dwType & DIDFT_AXIS)
	sprintf(ddoi.tszName, "Axis %d", DIDFT_GETINSTANCE(ddoi.dwType));
      else if (ddoi.dwType & DIDFT_BUTTON)
	sprintf(ddoi.tszName, "Button %d", DIDFT_GETINSTANCE(ddoi.dwType));
      else if (ddoi.dwType & DIDFT_POV)
	sprintf(ddoi.tszName, "POV Hat %d", DIDFT_GETINSTANCE(ddoi.dwType));
      else
	sprintf(ddoi.tszName, "Unknown");
      _dump_OBJECTINSTANCEA(&ddoi);
      if (lpCallback(&ddoi, lpvRef) != DIENUM_CONTINUE) return DI_OK;
    }
  }

  return DI_OK;
}

/******************************************************************************
  *     GetProperty : get input device properties
  */
static HRESULT WINAPI JoystickAImpl_GetProperty(LPDIRECTINPUTDEVICE8A iface,
						REFGUID rguid,
						LPDIPROPHEADER pdiph)
{
  ICOM_THIS(JoystickAImpl,iface);
  int obj;

  TRACE("(this=%p,%s,%p): stub!\n",
	iface, debugstr_guid(rguid), pdiph);

  if (TRACE_ON(dinput))
    _dump_DIPROPHEADER(pdiph);

  obj = find_object(This, pdiph->dwHow, pdiph->dwObj);
  if (pdiph->dwHow != DIPH_DEVICE && obj < 0) return DIERR_OBJECTNOTFOUND;

  if (!HIWORD(rguid)) {
    switch ((DWORD)rguid) {
    case (DWORD) DIPROP_BUFFERSIZE: {
      LPDIPROPDWORD	pd = (LPDIPROPDWORD)pdiph;

      TRACE(" return buffersize = %d\n",This->queue_len);
      pd->dwData = This->queue_len;
      break;
    }

    case (DWORD) DIPROP_RANGE: {
      LPDIPROPRANGE pr = (LPDIPROPRANGE) pdiph;

      if (obj < 0) return DIERR_INVALIDPARAM;
      pr->lMin = This->props[obj].lMin;
      pr->lMax = This->props[obj].lMax;
      break;
    }

    case (DWORD) DIPROP_DEADZONE: {
      LPDIPROPDWORD pd = (LPDIPROPDWORD) pdiph;

      if (obj < 0) return DIERR_INVALIDPARAM;
      pd->dwData = This->props[obj].lDeadZone;
      break;
    }

    default:
      FIXME("Unknown type %ld (%s)\n",(DWORD)rguid,debugstr_guid(rguid));
      break;
    }
  }


  return DI_OK;
}

static HRESULT WINAPI JoystickAImpl_GetDeviceInfo(
	LPDIRECTINPUTDEVICE8A iface,
	LPDIDEVICEINSTANCEA lpddi)
{
	ICOM_THIS(JoystickAImpl,iface);

	TRACE("(this=%p,%p)\n",
	      iface, lpddi);

	return DIJOY_GetDevInfo(This->dwID, This->dinput->ver, NULL, lpddi);
}

static ICOM_VTABLE(IDirectInputDevice8A) JoystickAvt =
{
	ICOM_MSVTABLE_COMPAT_DummyRTTIVALUE
	IDirectInputDevice2AImpl_QueryInterface,
	IDirectInputDevice2AImpl_AddRef,
	JoystickAImpl_Release,
	JoystickAImpl_GetCapabilities,
	JoystickAImpl_EnumObjects,
	JoystickAImpl_GetProperty,
	JoystickAImpl_SetProperty,
	JoystickAImpl_Acquire,
	JoystickAImpl_Unacquire,
	JoystickAImpl_GetDeviceState,
	JoystickAImpl_GetDeviceData,
	JoystickAImpl_SetDataFormat,
	JoystickAImpl_SetEventNotification,
	IDirectInputDevice2AImpl_SetCooperativeLevel,
	IDirectInputDevice2AImpl_GetObjectInfo,
	JoystickAImpl_GetDeviceInfo,
	IDirectInputDevice2AImpl_RunControlPanel,
	IDirectInputDevice2AImpl_Initialize,
	IDirectInputDevice2AImpl_CreateEffect,
	IDirectInputDevice2AImpl_EnumEffects,
	IDirectInputDevice2AImpl_GetEffectInfo,
	IDirectInputDevice2AImpl_GetForceFeedbackState,
	IDirectInputDevice2AImpl_SendForceFeedbackCommand,
	IDirectInputDevice2AImpl_EnumCreatedEffectObjects,
	IDirectInputDevice2AImpl_Escape,
	JoystickAImpl_Poll,
	IDirectInputDevice2AImpl_SendDeviceData,
	IDirectInputDevice7AImpl_EnumEffectsInFile,
	IDirectInputDevice7AImpl_WriteEffectToFile,
	IDirectInputDevice8AImpl_BuildActionMap,
	IDirectInputDevice8AImpl_SetActionMap,
	IDirectInputDevice8AImpl_GetImageInfo
};
