/*
 * Linux 2.4 joystick interface
 *
 * Copyright 2002 TransGaming Technologies Inc.
 */

#include "config.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef HAVE_LINUX_INPUT_H
#include <linux/input.h>
#endif
#ifdef HAVE_SYS_ERRNO_H
#include <sys/errno.h>
#endif

#include "windef.h"
#include "winbase.h"
#include "wingdi.h"
#include "winuser.h"
#include "mmddk.h"
#include "dinput.h"
#include "debugtools.h"

#include "joy_drv.h"

DEFAULT_DEBUG_CHANNEL(joystick);

extern void LINUX22_DevName(WINE_JOYSTICK* jstck);
extern LONG LINUX22_GetDevCaps(WINE_JOYSTICK* jstck);
extern LONG LINUX22_GetPosEx(WINE_JOYSTICK* jstck, LPJOYINFOEX lpInfo);

#ifdef HAVE_CORRECT_LINUXINPUT_H

#define test_bit(bit,arr) (((BYTE*)(arr))[(bit)>>3]&(1<<((bit)&7)))

#define FIRST_BTN BTN_MISC

typedef struct {
    DWORD dwVersion;
    DIDEVCAPS dicaps;
    DIDEVICEINSTANCEA diinfo;
    int fd;
    int abs_axes;
    int rel_axes;
    int buttons;
    int hats;
    char abs_axis_map[ABS_MAX+1];
    char rel_axis_map[REL_MAX+1];
    char button_map[(KEY_MAX-FIRST_BTN)+1];
    char hat_map[4];
    int abs_axis_info[ABS_MAX+1][5];
    DIDATAFORMAT df;
    int init;
} JOY_DATA;

/* Not all event devices are joysticks.  Try to find the appropriate
 * event device for the requested joystick instance.  For our
 * purposes, a joystick must have absolute X and Y axes or an absolute
 * wheel axis, and a trigger, a or 1 button. */

#define MAX_EVENT_DEVS 64

static int find_event_joystick(int n)
{
    int		i, fd, ndet = 0;
    char	devname[32];
    BYTE	absbits[(ABS_MAX+7)/8];
    BYTE	keybits[(KEY_MAX+7)/8];

    for (i = 0; i < MAX_EVENT_DEVS; i++) {
	sprintf(devname, "/dev/input/event%d", i);
	if ((fd = open(devname, O_RDONLY | O_NONBLOCK)) < 0)
	    continue;
	if (ioctl(fd, EVIOCGBIT(EV_ABS, sizeof(absbits)), absbits) < 0 ||
	    ioctl(fd, EVIOCGBIT(EV_KEY, sizeof(keybits)), keybits) < 0) {
	    close(fd);
	    continue;
	}

	if ((!test_bit(ABS_X, absbits) || !test_bit(ABS_Y, absbits)) &&
	    !test_bit(ABS_WHEEL, absbits))
	    continue;
	if (!test_bit(BTN_TRIGGER, keybits) && !test_bit(BTN_A, keybits) &&
	    !test_bit(BTN_1, keybits))
	    continue;

	if (ndet == n) {
	    TRACE("joystick%d found on %s\n", n, devname);
	    return i;
	}
	ndet++;
    }

    TRACE("joystick%d not found\n", n);
    return -1;
}

/* Because there seems to be no way to get the current button state after
 * an open() like we rely on in GetPosEx, we'll delegate ordinary winmm
 * requests like that to the linux 2.2 compatibility device /dev/input/jsX,
 * and only use the new /dev/input/eventX device in DirectInput mode */

void LINUXINPUT_DevName(WINE_JOYSTICK* jstck)
{
    TRACE("(%p)\n", jstck);
    /* the 2.2 compatibility device on 2.4 systems is /dev/input/js,
     * not /dev/js, which does not always exist */
    sprintf(jstck->devname, "/dev/input/js%d", jstck->joyIntf);
    /* Fix devIntf for event devices */
    jstck->devIntf = find_event_joystick(jstck->joyIntf);
}

LONG LINUXINPUT_GetDevCaps(WINE_JOYSTICK* jstck)
{
    LPJOYCAPSA lpCaps = &jstck->caps;
    LPDIDEVCAPS lpDICaps = &jstck->dicaps;
    char	devname[32];
    int		dev;
    int		driverVersion;
    LONG	err;

    TRACE("(%p)\n", jstck);

    /* check that the new API is available */
    sprintf(devname, "/dev/input/event%d", jstck->devIntf);
    if ((dev = open(devname, O_RDONLY | O_NONBLOCK)) < 0) {
	if (errno == ENODEV) return JOYERR_UNPLUGGED;
	else if (errno == ENOENT) return MMSYSERR_NODRIVER;
	else if (errno == EBUSY) return MMSYSERR_ALLOCATED;
	else return MMSYSERR_ERROR;
    }
    if (ioctl(dev, EVIOCGVERSION, &driverVersion) < 0) {
	close(dev);
	return MMSYSERR_NODRIVER;
    }
    close(dev);

    /* use old API for standard GetDevCaps */
    err = LINUX22_GetDevCaps(jstck);
    if (err == MMSYSERR_ALLOCATED) return err;

    /* winmm joystick API can't be supported */
    memset(lpCaps, 0, sizeof(JOYCAPSA));
    /* dinput caps will be retrieved when needed */
    memset(lpDICaps, 0, sizeof(DIDEVCAPS));

    /* the guidInstance is supposed to identify the joystick,
     * this is enough for that purpose */ 
    jstck->guidInstance.Data1 = jstck->joyIntf;

    /* return success */
    return JOYERR_NOERROR;
}

static LONG LINUXINPUT_OpenDev(WINE_JOYSTICK* jstck, DWORD dwVersion, JOY_DATA* data)
{
    char	devname[32];
    LPDIDEVCAPS lpDICaps;
    LPDIDEVICEINSTANCEA lpDIInfo;
    int		dev;
    char	identString[MAXPNAMELEN];
    BYTE	absbits[(ABS_MAX+7)/8];
    BYTE	keybits[(KEY_MAX+7)/8];
    BYTE	relbits[(REL_MAX+7)/8];
    int		driverVersion;
    unsigned	u;

    TRACE("(%p,%lx,%p)\n", jstck, dwVersion, data);
    sprintf(devname, "/dev/input/event%d", jstck->devIntf);
    if ((dev = open(devname, O_RDONLY | O_NONBLOCK)) < 0) {
	if (errno == ENODEV) return JOYERR_UNPLUGGED;
	else if (errno == ENOENT) return MMSYSERR_NODRIVER;
	else if (errno == EBUSY) return MMSYSERR_ALLOCATED;
	else return MMSYSERR_ERROR;
    }
    if (ioctl(dev, EVIOCGVERSION, &driverVersion) < 0) {
	close(dev);
	return MMSYSERR_NODRIVER;
    }
    /* EVIOCGNAME doesn't nul terminate if the buffer is too small so
     * do it here just in case. */
    identString[sizeof(identString)-1] = 0;
    if (ioctl(dev, EVIOCGNAME(sizeof(identString)-1), &identString) < 0) strcpy(identString, "");
    TRACE("Driver: 0x%06x, Name: %s\n", driverVersion, identString);

    if (!data) {
	if (!jstck->internal)
	    jstck->internal = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(JOY_DATA));
	data = jstck->internal;
    }
    lpDICaps = &data->dicaps;
    lpDIInfo = &data->diinfo;

    data->dwVersion = dwVersion;

    /* count and map buttons */
    data->buttons = 0;
    if (ioctl(dev, EVIOCGBIT(EV_KEY, sizeof(keybits)), keybits) >= 0) {
	for (u=BTN_JOYSTICK; u<=KEY_MAX && data->buttons<32; u++) {
	    if (test_bit(u, keybits)) {
		TRACE("button[%d] = %d\n", data->buttons, u);
		data->button_map[u-FIRST_BTN] = data->buttons++;
	    } else
		data->button_map[u-FIRST_BTN] = -1;
	}
	for (u=FIRST_BTN; u<BTN_JOYSTICK && data->buttons<32; u++) {
	    if (test_bit(u, keybits)) {
		TRACE("button[%d] = %d\n", data->buttons, u);
		data->button_map[u-FIRST_BTN] = data->buttons++;
	    } else
		data->button_map[u-FIRST_BTN] = -1;
	}
    }
    /* count and map axes and hats */
    data->abs_axes = 0;
    data->hats = 0;
    if (ioctl(dev, EVIOCGBIT(EV_ABS, sizeof(absbits)), absbits) >= 0) {
	for (u=0; u<=ABS_MAX; u++) {
	    if (((u < ABS_HAT0X) || (u > ABS_HAT3Y)) &&
		test_bit(u, absbits)) {
		TRACE("abs_axis[%d] = %d\n", data->abs_axes, u);
		data->abs_axis_map[u] = data->abs_axes++;
		ioctl(dev, EVIOCGABS(u), data->abs_axis_info[u]);
	    } else
		data->abs_axis_map[u] = -1;
	}
	for (u=ABS_HAT0X; u<=ABS_HAT3Y; u+=2) {
	    if (test_bit(u, absbits) &&
		test_bit(u+1, absbits)) {
		TRACE("hat[%d] = %d,%d\n", data->hats, u, u+1);
		data->hat_map[(u-ABS_HAT0X)/2] = data->hats++;
	    } else
		data->hat_map[(u-ABS_HAT0X)/2] = -1;
	}
    }
    data->rel_axes = 0;
    if (ioctl(dev, EVIOCGBIT(EV_REL, sizeof(relbits)), relbits) >= 0) {
	for (u=0; u<=REL_MAX; u++) {
	    if (test_bit(u, relbits)) {
		TRACE("rel_axis[%d] = %d\n", data->rel_axes, u);
		data->rel_axis_map[u] = data->rel_axes++;
	    } else
		data->rel_axis_map[u] = -1;
	}
    }

    TRACE("detected: abs axes=%d, rel axes=%d, hats=%d, buttons=%d\n",
	data->abs_axes, data->rel_axes, data->hats, data->buttons);

    /* FIXME: handle relative axes somehow */

    memset(lpDICaps, 0, sizeof(DIDEVCAPS));
    lpDICaps->dwSize = sizeof(DIDEVCAPS);
    lpDICaps->dwFlags = DIDC_ATTACHED;
    /* FIXME: we might be able to figure out controller type
     * by checking what kind of axes and buttons it has */
    if (HIBYTE(dwVersion) >= 8)
	lpDICaps->dwDevType = DI8DEVTYPE_JOYSTICK |
			     (DI8DEVTYPEJOYSTICK_STANDARD<<8);
    else
	lpDICaps->dwDevType = DIDEVTYPE_JOYSTICK |
			     (DIDEVTYPEJOYSTICK_TRADITIONAL<<8);
    lpDICaps->dwAxes = data->abs_axes;
    lpDICaps->dwButtons = data->buttons;
    lpDICaps->dwPOVs = data->hats;

    memset(lpDIInfo, 0, sizeof(DIDEVICEINSTANCEA));
    lpDIInfo->dwSize = sizeof(DIDEVICEINSTANCEA);
    memcpy(&lpDIInfo->guidInstance, &jstck->guidInstance, sizeof(GUID));
    memcpy(&lpDIInfo->guidProduct, &GUID_Joystick, sizeof(GUID));
    lpDIInfo->dwDevType = jstck->dicaps.dwDevType;
    sprintf(lpDIInfo->tszInstanceName, "Joystick %d", jstck->joyIntf+1);
    strncpy(lpDIInfo->tszProductName, identString, MAX_PATH);
    lpDIInfo->tszProductName[MAX_PATH-1] = '\0';

    data->fd = dev;
    data->init = 0;

    return JOYERR_NOERROR;
}

HRESULT LINUXINPUT_GetDIInfo(WINE_JOYSTICK* jstck, DWORD dwVersion, LPDIDEVCAPS lpCaps, LPDIDEVICEINSTANCEA lpInfo)
{
    /* since, in the equation /dev/input/jsX = /dev/input/eventY,
     * it is not necessarily such that X = Y, we'll ignore the dicaps
     * from the js device and compute our own here */
    JOY_DATA data;
    JOY_DATA *pdata = &data;
    TRACE("(%p,%lx,%p,%p)\n", jstck, dwVersion, lpCaps, lpInfo);
    if (jstck->internal) pdata = jstck->internal;
    else {
	LONG err = LINUXINPUT_OpenDev(jstck, dwVersion, pdata);
	if (err == MMSYSERR_ALLOCATED) return DIERR_ACQUIRED;
	if (err != JOYERR_NOERROR) return DIERR_GENERIC;
	close(data.fd);
    }
    if (lpCaps) {
	DWORD dwSize = (lpCaps->dwSize > sizeof(pdata->dicaps)) ? sizeof(pdata->dicaps) : lpCaps->dwSize;
	memcpy(lpCaps, &pdata->dicaps, dwSize);
    }
    if (lpInfo) {
	DWORD dwSize = (lpInfo->dwSize > sizeof(pdata->diinfo)) ? sizeof(pdata->diinfo) : lpInfo->dwSize;
	memcpy(lpInfo, &pdata->diinfo, dwSize);
    }
    return DI_OK;
}

HRESULT LINUXINPUT_OpenDIDev(WINE_JOYSTICK* jstck, DWORD dwVersion)
{
    LONG err;
    JOY_DATA *pdata;
    LPDIOBJECTDATAFORMAT obj;
    unsigned u, o;

    TRACE("(%p,%lx)\n", jstck, dwVersion);
    err = LINUXINPUT_OpenDev(jstck, dwVersion, NULL);
    if (err == MMSYSERR_ALLOCATED) return DIERR_ACQUIRED;
    if (err != JOYERR_NOERROR) return DIERR_GENERIC;

    pdata = jstck->internal;

    /* build data format */
    pdata->df.dwNumObjs = pdata->dicaps.dwAxes + pdata->dicaps.dwPOVs + pdata->dicaps.dwButtons;
    pdata->df.rgodf = HeapAlloc(GetProcessHeap(), 0, sizeof(DIOBJECTDATAFORMAT) * pdata->df.dwNumObjs);
    obj = pdata->df.rgodf;
    o = 0;
    /* add axes */
    for (u=0; u<=ABS_MAX; u++) if (pdata->abs_axis_map[u] != -1) {
	switch (u) {
	case ABS_X:
	case ABS_WHEEL:
	    obj->pguid = &GUID_XAxis;
	    break;
	case ABS_Y:
	case ABS_GAS:
	    obj->pguid = &GUID_YAxis;
	    break;
	case ABS_Z:
	case ABS_BRAKE:
	    obj->pguid = &GUID_ZAxis;
	    break;
	case ABS_RX:
	    obj->pguid = &GUID_RxAxis;
	    break;
	case ABS_RY:
	    obj->pguid = &GUID_RyAxis;
	    break;
	case ABS_RZ:
	case ABS_RUDDER:
	    obj->pguid = &GUID_RzAxis;
	    break;
	default:
	    obj->pguid = &GUID_Slider;
	    break;
	}
	obj->dwOfs = o;
	obj->dwType = DIDFT_ABSAXIS | DIDFT_MAKEINSTANCE(pdata->abs_axis_map[u]);
	obj->dwFlags = 0;
	o += sizeof(LONG);
	obj++;
    }
    /* add hats */
    for (u=0; u<pdata->dicaps.dwPOVs; u++,obj++) {
	obj->pguid = &GUID_POV;
	obj->dwOfs = o;
	obj->dwType = DIDFT_POV | DIDFT_MAKEINSTANCE(u);
	obj->dwFlags = 0;
	o += sizeof(DWORD);
    }
    /* add buttons */
    for (u=0; u<pdata->dicaps.dwButtons; u++,obj++) {
	obj->pguid = &GUID_Button;
	obj->dwOfs = o;
	obj->dwType = DIDFT_PSHBUTTON | DIDFT_MAKEINSTANCE(u);
	obj->dwFlags = 0;
	o += sizeof(BYTE);
    }
    pdata->df.dwDataSize = (o+3)&~3; /* dword-aligned data size */

    return DI_OK;
}

HRESULT LINUXINPUT_CloseDIDev(WINE_JOYSTICK* jstck)
{
    JOY_DATA *pdata = jstck->internal;
    TRACE("(%p)\n", jstck);
    if (!pdata) return DI_OK;
    close(pdata->fd);
    HeapFree(GetProcessHeap(), 0, pdata->df.rgodf);
    HeapFree(GetProcessHeap(), 0, pdata);
    jstck->internal = 0;
    return DI_OK;
}

HRESULT LINUXINPUT_GetDIDataFormat(WINE_JOYSTICK* jstck, LPDIDATAFORMAT* pdf)
{
    JOY_DATA *pdata = jstck->internal;
    TRACE("(%p,%p)\n", jstck, pdf);
    *pdf = &pdata->df;
    return DI_OK;
}

#define AXE_ABS		0
#define AXE_ABSMIN	1
#define AXE_ABSMAX	2
#define AXE_ABSFUZZ	3
#define AXE_ABSFLAT	4

/* expand axis values to the full range expected by dinput,
 * and let dinput deal with the rest, like dead zones */
static DWORD map_abs_axis(JOY_DATA *pdata, unsigned axis, int val)
{
    int xmin = pdata->abs_axis_info[axis][AXE_ABSMIN];
    int xmax = pdata->abs_axis_info[axis][AXE_ABSMAX];

    if (xmin == xmax) return val; /* avoid division by zero */

    /* clamp just in case */
    if (val < xmin) val = xmin;
    if (val > xmax) val = xmax;

    return (val-xmin)*65534/(xmax-xmin);
}

HRESULT LINUXINPUT_GetDIData(WINE_JOYSTICK* jstck, BOOL req, DWORD* dwObj, DWORD* dwData, DWORD* dwTimeStamp)
{
    JOY_DATA *pdata = jstck->internal;
    struct input_event ev;
    DWORD Time;
    unsigned u;
    char map;

    if (pdata->init < pdata->dicaps.dwAxes) {
	unsigned u, i;
	*dwObj = pdata->init;
	*dwTimeStamp = GetTickCount();
	for (u=0,i=0; u<=ABS_MAX; u++) if (pdata->abs_axis_map[u] != -1) {
	    if (i++ != pdata->init) continue;
	    *dwData = map_abs_axis(pdata, u, pdata->abs_axis_info[u][AXE_ABS]);
	    pdata->init++;
	    return DI_OK;
	}
	/* shouldn't get here */
	pdata->init = pdata->dicaps.dwAxes;
    }

    if (pdata->init < pdata->dicaps.dwAxes + pdata->dicaps.dwPOVs) {
	unsigned u, i;
	*dwObj = pdata->init;
	*dwTimeStamp = GetTickCount();
	for (u=0,i=pdata->dicaps.dwAxes; u<4; u++) if (pdata->hat_map[u] != -1) {
	    if (i++ != pdata->init) continue;
	    *dwData = JSTCK_calc_povcts(pdata->abs_axis_info[ABS_HAT0X + u*2][AXE_ABS],
					pdata->abs_axis_info[ABS_HAT0Y + u*2][AXE_ABS]);
	    pdata->init++;
	    return DI_OK;
	}
	/* shouldn't get here */
	pdata->init = pdata->dicaps.dwAxes + pdata->dicaps.dwPOVs;
    }

    while (TRUE) {
	if (read(pdata->fd, &ev, sizeof(struct input_event)) < 0)
	    return S_FALSE;

	/* FIXME: derive timestamp from ev.time */
	Time = GetTickCount();

	switch (ev.type) {
	case EV_KEY:
	    TRACE("key %d -> %d\n", ev.code, ev.value);
	    if (ev.code < FIRST_BTN) break;
	    map = pdata->button_map[ev.code-FIRST_BTN];
	    if (map == -1) break;
	    *dwObj = pdata->dicaps.dwAxes + pdata->dicaps.dwPOVs + map;
	    *dwData = ev.value ? 0x80 : 0x00;
	    *dwTimeStamp = Time;
	    return DI_OK;
	case EV_ABS:
	    TRACE("abs axis %d -> %d\n", ev.code, ev.value);
	    pdata->abs_axis_info[ev.code][AXE_ABS] = ev.value;
	    if (ev.code < ABS_HAT0X || ev.code > ABS_HAT3Y) {
		map = pdata->abs_axis_map[ev.code];
		if (map == -1) break;
		*dwObj = map;
		*dwData = map_abs_axis(pdata, ev.code, ev.value);
		*dwTimeStamp = Time;
		return DI_OK;
	    }
	    /* handle hat */
	    u = (ev.code - ABS_HAT0X) / 2;
	    map = pdata->hat_map[u];
	    if (map == -1) break;
	    *dwObj = pdata->dicaps.dwAxes + map;
	    *dwData = JSTCK_calc_povcts(pdata->abs_axis_info[ABS_HAT0X + u*2][AXE_ABS],
					pdata->abs_axis_info[ABS_HAT0Y + u*2][AXE_ABS]);
	    *dwTimeStamp = Time;
	    return DI_OK;
	}
    }
}

WINE_JOYSTICK_DRV jdrv_linuxinput = {
    LINUXINPUT_DevName, LINUXINPUT_GetDevCaps, LINUX22_GetPosEx,
    LINUXINPUT_GetDIInfo, LINUXINPUT_OpenDIDev, LINUXINPUT_CloseDIDev,
    LINUXINPUT_GetDIDataFormat, LINUXINPUT_GetDIData
};

#endif
