/*
 * Direct3D GLX execute buffer implementation
 *
 * Copyright 2001 TransGaming Technologies, Inc.
 */


#include "config.h"
#include "glx.h"
#include "executebuffer.h"
#include "d3d.h"
#include "d3dhal.h"
#include "wine_gl.h"
#include "wine/d3dhalgl.h"

#include "winbase.h"
#include "debugtools.h"

DEFAULT_DEBUG_CHANNEL(x11drv);

/* GL_NV_vertex_array_range(2) & GL_NV_fence
 * The VAR arena is shared by all contexts, as are the function pointers,
 * but the extension must be tested in each context.
 */
#ifndef GL_ALL_COMPLETED_NV
#define GL_ALL_COMPLETED_NV 0x84F2
#endif

#ifndef GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV
#define GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV 0x8533
#endif

static BOOL X11DRV_GLX_VAR_Enable;
static PFNGLFLUSHVERTEXARRAYRANGENVPROC X11DRV_GLX_glFlushVertexArrayRangeNV;
static PFNGLVERTEXARRAYRANGENVPROC X11DRV_GLX_glVertexArrayRangeNV;
static void *(*X11DRV_GLX_glXAllocateMemoryNV)(GLsizei size, GLfloat readfrequency, GLfloat writefrequency, GLfloat priority);
static void (*X11DRV_GLX_glXFreeMemoryNV)(void *pointer);

#if defined(WINE_DDRAW_need_NV_fence_pfns) || !defined(GL_NV_fence)
typedef void (*PFNGLGENFENCESNVPROC)(GLsizei n, GLuint *fences);
typedef void (*PFNGLDELETEFENCESNVPROC)(GLsizei n, const GLuint *fences);
typedef void (*PFNGLSETFENCENVPROC)(GLuint fence, GLenum condition);
typedef GLboolean (*PFNGLTESTFENCENVPROC)(GLuint fence);
typedef void (*PFNGLFINISHFENCENVPROC)(GLuint fence);
typedef GLboolean (*PFNGLISFENCENVPROC)(GLuint fence);
typedef void (*PFNGLGETFENCEIVNVPROC)(GLuint fence, GLenum pname,
				      GLint *params);
#endif

static PFNGLGENFENCESNVPROC X11DRV_GLX_glGenFencesNV;
static PFNGLDELETEFENCESNVPROC X11DRV_GLX_glDeleteFencesNV;
static PFNGLSETFENCENVPROC X11DRV_GLX_glSetFenceNV;
static PFNGLTESTFENCENVPROC X11DRV_GLX_glTestFenceNV;
static PFNGLFINISHFENCENVPROC X11DRV_GLX_glFinishFenceNV;
static PFNGLISFENCENVPROC X11DRV_GLX_glIsFenceNV;

static LPVOID X11DRV_GLX_VertexArray;
static DWORD X11DRV_GLX_VertexArraySize;

void X11DRV_GLX_EB_Init(void)
{
#ifdef HAVE_GLX_GETPROCADDRESS
    X11DRV_GLX_glFlushVertexArrayRangeNV
	= glXGetProcAddressARB("glFlushVertexArrayRangeNV");
    X11DRV_GLX_glVertexArrayRangeNV = glXGetProcAddressARB("glVertexArrayRangeNV");
    X11DRV_GLX_glXAllocateMemoryNV = glXGetProcAddressARB("glXAllocateMemoryNV");
    X11DRV_GLX_glXFreeMemoryNV = glXGetProcAddressARB("glXFreeMemoryNV");
    X11DRV_GLX_glGenFencesNV = glXGetProcAddressARB("glGenFencesNV");
    X11DRV_GLX_glDeleteFencesNV = glXGetProcAddressARB("glDeleteFencesNV");
    X11DRV_GLX_glSetFenceNV = glXGetProcAddressARB("glSetFenceNV");
    X11DRV_GLX_glTestFenceNV = glXGetProcAddressARB("glTestFenceNV");
    X11DRV_GLX_glFinishFenceNV = glXGetProcAddressARB("glFinishFenceNV");
    X11DRV_GLX_glIsFenceNV = glXGetProcAddressARB("glIsFenceNV");
#endif

    /* This is the global enable, now it's up to individual contexts. */
    X11DRV_GLX_VAR_Enable = (X11DRV_GLX_glFlushVertexArrayRangeNV
			     && X11DRV_GLX_glVertexArrayRangeNV
			     && X11DRV_GLX_glXAllocateMemoryNV
			     && X11DRV_GLX_glXFreeMemoryNV
			     && X11DRV_GLX_glGenFencesNV
			     && X11DRV_GLX_glDeleteFencesNV
			     && X11DRV_GLX_glSetFenceNV
			     && X11DRV_GLX_glTestFenceNV
			     && X11DRV_GLX_glFinishFenceNV
			     && X11DRV_GLX_glIsFenceNV);
}

void X11DRV_GLX_EB_Cleanup(void)
{
    if (X11DRV_GLX_VAR_Enable && (X11DRV_GLX_VertexArray != NULL))
    {
	X11DRV_GLX_glXFreeMemoryNV(X11DRV_GLX_VertexArray);
	X11DRV_GLX_VertexArray = NULL;
    }
}

void X11DRV_GLX_EB_Context_Init(void)
{
    const char *exts = glGetString(GL_EXTENSIONS);

    if (X11DRV_GLX_VAR_Enable && exts
	&& strstr(exts, "GL_NV_vertex_array_range")
	&& strstr(exts, "GL_NV_vertex_array_range2")
	&& strstr(exts, "GL_NV_fence")) {

	/* One-time VertexArrayRange setup. */
	if (!X11DRV_GLX_VertexArray) {
	    static const GLfloat readfrequency = 0;
	    static const GLfloat writefrequency = 0;
	    static const GLfloat agp_priority = 0.75;
	    static const GLfloat vidmem_priority = 1.0;

	    X11DRV_GLX_VertexArraySize = 16 * 1024 * 1024; /*TODO: tweakable*/

	    X11DRV_GLX_VertexArraySize *= 2;

	    do
	    {
		X11DRV_GLX_VertexArraySize /= 2;

		X11DRV_GLX_VertexArray
		    = X11DRV_GLX_glXAllocateMemoryNV(X11DRV_GLX_VertexArraySize,
						     readfrequency,
						     writefrequency,
						     agp_priority);
	    } while(X11DRV_GLX_VertexArray == NULL
		    && X11DRV_GLX_VertexArraySize > 1);

	    TRACE("allocated %p %ld\n", X11DRV_GLX_VertexArray,
		  X11DRV_GLX_VertexArraySize);

	    if (X11DRV_GLX_VertexArray == NULL)
		X11DRV_GLX_VAR_Enable = FALSE;
	}

	/* Per-context setup */
	if (X11DRV_GLX_VAR_Enable)
	{
	    X11DRV_GLX_glVertexArrayRangeNV(X11DRV_GLX_VertexArraySize,
					    X11DRV_GLX_VertexArray);
	}
    }
}

/** Generic VAR allocation */

struct X11DRV_GLX_VARBlock
{
    BOOL free;
    DWORD size;
    struct X11DRV_GLX_VARBlock* prev;
    struct X11DRV_GLX_VARBlock* next;
};

static struct X11DRV_GLX_VARBlock *X11DRV_GLX_var_head = NULL;

static LPVOID X11DRV_GLX_AllocateVARMemory(DWORD size)
{
    struct X11DRV_GLX_VARBlock* avail;

    if (!X11DRV_GLX_VAR_Enable) return NULL;
    if (!X11DRV_GLX_VertexArray) return NULL;

    /* one-time init */
    if (X11DRV_GLX_var_head == NULL)
    {
	X11DRV_GLX_var_head = X11DRV_GLX_VertexArray;
	X11DRV_GLX_var_head->free = TRUE;
	X11DRV_GLX_var_head->size = X11DRV_GLX_VertexArraySize;
	X11DRV_GLX_var_head->prev = X11DRV_GLX_var_head;
	X11DRV_GLX_var_head->next = X11DRV_GLX_var_head;
    }

    size = (size + 31) &~ 31;

    /* find free block */

    avail = X11DRV_GLX_var_head;

    do
    {
	if (avail->free && avail->size >= size + 32) break;
	avail = avail->next;
    }
    while (avail != X11DRV_GLX_var_head);

    if (!avail->free)
    {
	WARN("oh no! out of memory\n");
	return NULL;
    }

    /* split block */
    if (avail->size >= size + 96)
    {
	struct X11DRV_GLX_VARBlock *newblock;

	newblock = (struct X11DRV_GLX_VARBlock *)((char *)avail + 32 + size);
	newblock->free = TRUE;

	newblock->next = avail->next;
	newblock->prev = avail;
	newblock->next->prev = newblock;
	newblock->prev->next = newblock;

	newblock->size = avail->size - size - 32;
	avail->size = size + 32;
    }

    avail->free = FALSE;

    return (LPBYTE)avail + 32;
}

static void
X11DRV_GLX_CombineFreeVARBlocks(struct X11DRV_GLX_VARBlock* first,
				struct X11DRV_GLX_VARBlock* second)
{
    /* When there's only one block, we might try to combine it with itself.
     * The caller is responsible for checking that it isn't trying to combine
     * the last block with the first block, since it needs to know about that
     * case anyway. */
    if (first != second)
    {
	first->next = second->next;
	second->next->prev = first;

	first->size += second->size;
    }
}

static void X11DRV_GLX_FreeVARMemory(LPVOID ptr)
{
    struct X11DRV_GLX_VARBlock *current_block
	= (struct X11DRV_GLX_VARBlock*)((LPBYTE)ptr - 32);

    if (ptr == NULL) return;

    current_block->free = TRUE;

    if (current_block != X11DRV_GLX_var_head && current_block->prev->free)
    {
	X11DRV_GLX_CombineFreeVARBlocks(current_block->prev, current_block);
	current_block = current_block->prev;
    }

    if (current_block->next != X11DRV_GLX_var_head
	&& current_block->next->free)
    {
	X11DRV_GLX_CombineFreeVARBlocks(current_block, current_block->next);
    }
}


/** VAR rename buffers */

#define MAX_RENAME_DEPTH 8

struct X11DRV_GLX_VAR_Buffer
{
    LPVOID data;
    GLuint fence;
};

struct X11DRV_GLX_VAR_BufferControl
{
    DWORD num_buffers;
    DWORD current_buffer;
    DWORD buffer_size;

    struct X11DRV_GLX_VAR_BufferControl *prev, *next;

    struct X11DRV_GLX_VAR_Buffer buffers[MAX_RENAME_DEPTH];
};

static struct X11DRV_GLX_VAR_BufferControl X11DRV_GLX_VAR_BufferControlList =
{
    0, 0, 0,
    &X11DRV_GLX_VAR_BufferControlList, &X11DRV_GLX_VAR_BufferControlList,
};


static BOOL X11DRV_GLX_VAR_TestBuffer(struct X11DRV_GLX_VAR_BufferControl* ctl,
				      DWORD buffer)
{
    return !X11DRV_GLX_glIsFenceNV(ctl->buffers[buffer].fence)
	|| X11DRV_GLX_glTestFenceNV(ctl->buffers[buffer].fence);
}

static void X11DRV_GLX_VAR_WaitBuffer(struct X11DRV_GLX_VAR_BufferControl* ctl,
				      DWORD buffer)
{
    if (X11DRV_GLX_glIsFenceNV(ctl->buffers[buffer].fence))
	X11DRV_GLX_glFinishFenceNV(ctl->buffers[buffer].fence);
}

/* Returns the reclaimed size. (May be changed to TRUE/FALSE in the future.)
 * Does not attempt to reclaim from skip. (Which can be NULL.)
 */
static DWORD
X11DRV_GLX_VAR_ReclaimBuffers(struct X11DRV_GLX_VAR_BufferControl* skip)
{
    struct X11DRV_GLX_VAR_BufferControl* victim;
    DWORD reclaimed_size = 0;

    for (victim = X11DRV_GLX_VAR_BufferControlList.next;
	 victim != &X11DRV_GLX_VAR_BufferControlList; victim = victim->next)
    {
	int i;

	if (victim == skip) continue;

	/* we won't be able to steal any anyway. */
	if (victim->num_buffers < 2) continue;

	for (i=0; i < victim->num_buffers; i++)
	{
	    if (i == victim->current_buffer) continue;

	    if (X11DRV_GLX_VAR_TestBuffer(victim, i))
	    {
		struct X11DRV_GLX_VAR_Buffer *buffer = &victim->buffers[i];

		/* free the buffer's resources */
		X11DRV_GLX_glDeleteFencesNV(1, &buffer->fence);
		X11DRV_GLX_FreeVARMemory(buffer->data);

		/* remove the buffer from the array */
		victim->num_buffers--;

		/* If i was the last buffer, then i = num_buffers + 1,
		 * so we compute -sizeof(*buffer) below which overflows. */
		if (victim->num_buffers > i)
		    memmove(buffer, buffer+1,
			    sizeof(*buffer) * victim->num_buffers - i);

		if (victim->current_buffer >= i)
		{
		    if (i == 0) victim->current_buffer = victim->num_buffers;
		    else victim->current_buffer--;
		}

		reclaimed_size += victim->buffer_size;
		/* cancel the i++ and rescan this entry. */
		i--;
	    }
	}
    }

    return reclaimed_size;
}

static BOOL
X11DRV_GLX_VAR_AllocateBuffer(struct X11DRV_GLX_VAR_BufferControl* ctl)
{
    struct X11DRV_GLX_VAR_Buffer* buffer;

    /* What return value makes sense here? */
    if (ctl->num_buffers == MAX_RENAME_DEPTH) return FALSE;

    buffer = &ctl->buffers[ctl->num_buffers];

    buffer->data = X11DRV_GLX_AllocateVARMemory(ctl->buffer_size);
    if (buffer->data == NULL)
    {
	X11DRV_GLX_VAR_ReclaimBuffers(ctl);

	buffer->data = X11DRV_GLX_AllocateVARMemory(ctl->buffer_size);
	if (buffer->data == NULL)
	    return FALSE;
    }

    X11DRV_GLX_glGenFencesNV(1, &buffer->fence);

    ctl->num_buffers++;

    return TRUE;
}

static BOOL
X11DRV_GLX_VAR_InitBufferControl(struct X11DRV_GLX_VAR_BufferControl* ctl,
				 DWORD buffer_size)
{
    ctl->num_buffers = 0;
    ctl->current_buffer = 0;
    ctl->buffer_size = buffer_size;

    /* This hack makes it possible to safely call FreeBufferControl even
     * if this fails. */
    ctl->prev = ctl;
    ctl->next = ctl;

    if (X11DRV_GLX_VAR_AllocateBuffer(ctl))
    {
	ctl->prev = &X11DRV_GLX_VAR_BufferControlList;
	ctl->next = X11DRV_GLX_VAR_BufferControlList.next;
	X11DRV_GLX_VAR_BufferControlList.next = ctl;
	ctl->next->prev = ctl;

	return TRUE;
    }
    else
    {
	return FALSE;
    }
}

static void
X11DRV_GLX_VAR_FreeBufferControl(struct X11DRV_GLX_VAR_BufferControl* ctl)
{
    DWORD i;

    for (i=0; i < ctl->num_buffers; i++)
    {
	if (X11DRV_GLX_glIsFenceNV(ctl->buffers[i].fence))
	{
	    X11DRV_GLX_glFinishFenceNV(ctl->buffers[i].fence);
	    X11DRV_GLX_glDeleteFencesNV(1, &ctl->buffers[i].fence);
	}

	X11DRV_GLX_FreeVARMemory(ctl->buffers[i].data);
    }

    ctl->prev->next = ctl->next;
    ctl->next->prev = ctl->prev;
}

/* The caller wishes to overwrite the current buffer.
 * Rename or wait as appropriate. */
static void X11DRV_GLX_VAR_Overwrite(struct X11DRV_GLX_VAR_BufferControl* ctl,
				     BOOL can_rename)
{
    DWORD oldest_buffer;

    if (!can_rename)
    {
	X11DRV_GLX_VAR_WaitBuffer(ctl, ctl->current_buffer);
	return;
    }

    /* Try the current buffer first.
     * If we can reuse the current buffer, we don't need to do anything.
     */
    if (X11DRV_GLX_VAR_TestBuffer(ctl, ctl->current_buffer))
	return;

    /* Try the oldest buffer. */
    oldest_buffer = (ctl->current_buffer + 1) % ctl->num_buffers;

    if (X11DRV_GLX_VAR_TestBuffer(ctl, oldest_buffer))
    {
	/* rename to the oldest buffer */
	ctl->current_buffer = oldest_buffer;
	return;
    }

    /* Try to allocate a new rename buffer. */
    if (X11DRV_GLX_VAR_AllocateBuffer(ctl))
    {
	/* num_buffers has changed, so this is not the same as oldest_buffer */
	ctl->current_buffer = (ctl->current_buffer + 1) % ctl->num_buffers;
	return;
    }

    /* Just wait on the oldest buffer, and rename it to that.
     * (AllocateBuffer failed, so it hasn't changed num_buffers, so
     * oldest_buffer is still valid.) */
    X11DRV_GLX_VAR_WaitBuffer(ctl, oldest_buffer);
    ctl->current_buffer = oldest_buffer;
    return;
}

static inline LPVOID
X11DRV_GLX_VAR_GetCurrentBuffer(struct X11DRV_GLX_VAR_BufferControl* ctl)
{
    return ctl->buffers[ctl->current_buffer].data;
}

static inline GLuint
X11DRV_GLX_VAR_GetCurrentFence(struct X11DRV_GLX_VAR_BufferControl* ctl)
{
    return ctl->buffers[ctl->current_buffer].fence;
}


/** Execute buffer implementation */

struct X11DRV_GLX_EB_Info
{
    LPVOID lock_buffer;
    LPVOID render_buffer;

    BOOL var;
    BOOL lock_buffer_is_var;
    BOOL render_buffer_is_var;

    struct X11DRV_GLX_VAR_BufferControl var_buffers;
    LPVOID system_buffer;

    DWORD num_vertices;
    DWORD size;

    DWORD original_fvf;
    DWORD converted_fvf;
    DWORD original_fvf_size;
    DWORD converted_fvf_size;

    GLDRAWPRIMITIVESTRIDEDDATA lock_strided_data, render_strided_data;

    /* These always apply to the current buffer. */
    DWORD overwritten;
    BYTE uptodate[1];
};

/* This is common to setting up from either a D3D stride, or from
 * a pointer and FVF. */
static void
X11DRV_GLX_EB_SetupStrideBlendWeights(DWORD fvf,
				      LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    int num_weights = min((fvf & D3DFVF_POSITION_MASK) / 2, 2) - 2;
    int i;

    /* Blend weights immediately follow the position data.
     * There are a maximum of 4: XYZB5 does not actually specify 5 blend
     * weights since it must be used with D3DFVF_LASTBETA_UBYTE4 flag,
     * meaning that the last float is actually 4 matrix palette indices. */
    for (i=0; i < 4; i++)
    {
	gl->blend_weights[i].lpvData = (LPFLOAT)gl->position.lpvData + 3 + i;
	gl->blend_weights[i].dwStride = gl->position.dwStride;
    }

    /* Setup the last blend weight as the matrix index pointer.
     * (In case that's what it actually is.) */
    gl->matrix_indices.lpvData
	= (LPFLOAT)gl->position.lpvData + 3 + num_weights - 1;
}

/* Apparently there is a super-secret fog value hidden in specular alpha. */
static void
X11DRV_GLX_EB_SetupStrideFogCoordinate(DWORD fvf,
				       LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    gl->fog_coordinate.lpvData = (LPBYTE)gl->color[1].lpvData + 3;
    gl->fog_coordinate.dwStride = gl->color[1].dwStride;
}

static void
X11DRV_GLX_EB_ConvertInterleavedToGLStride(DWORD fvf, LPCVOID lpVertices,
					   LPGLDRAWPRIMITIVESTRIDEDDATA gl)
{
    static const DWORD position_sizes[] = { 0, 3, 4, 4, 5, 6, 7, 8 };

    LPBYTE p = (LPBYTE)lpVertices;
    DWORD stride;
    int i;

    gl->position.lpvData = p;
    stride = position_sizes[(fvf & D3DFVF_POSITION_MASK) / 2]
	* sizeof(D3DVALUE);

    if (fvf & D3DFVF_RESERVED1)
	stride += sizeof(D3DVALUE);

    if (fvf & D3DFVF_NORMAL)
    {
	gl->normal.lpvData = p + stride;
	stride += 3 * sizeof(D3DVALUE);
    }

    if (fvf & D3DFVF_DIFFUSE)
    {
	gl->color[0].lpvData = p + stride;
	stride += sizeof(D3DCOLOR);
    }

    if (fvf & D3DFVF_SPECULAR)
    {
	gl->color[1].lpvData = p + stride;
	stride += sizeof(D3DCOLOR);
    }

    /* XXX: This assumes 2D texture coordinates. */
    if (fvf & D3DFVF_TEXCOUNT_MASK)
    {
	for (i = 0; i < D3DDP_MAXTEXCOORD; i++)
	    gl->textureCoords[i].lpvData
		= p + stride + 2 * sizeof(D3DVALUE) * i;

	stride += 2 * ((fvf & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT)
	    * sizeof(D3DVALUE);
    }

    gl->position.dwStride = stride;
    gl->normal.dwStride = stride;
    gl->color[0].dwStride = stride;
    gl->color[1].dwStride = stride;
    for (i=0; i < D3DDP_MAXTEXCOORD; i++)
	gl->textureCoords[i].dwStride = stride;

    X11DRV_GLX_EB_SetupStrideBlendWeights(fvf, gl);
    X11DRV_GLX_EB_SetupStrideFogCoordinate(fvf, gl);
}

static BOOL X11DRV_GLX_DoesFVFNeedConversion(DWORD fvf)
{
    return ((fvf & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW
	    || (fvf & (D3DFVF_DIFFUSE|D3DFVF_SPECULAR)));
}

static DWORD X11DRV_GLX_ConvertedFVF(DWORD fvf)
{
    if ((fvf & D3DFVF_POSITION_MASK) == D3DFVF_XYZRHW)
	fvf |= D3DFVF_GLW;

    if (fvf & (D3DFVF_DIFFUSE|D3DFVF_SPECULAR))
	fvf |= D3DFVF_GLCOLORORDER;

    return fvf;
}

static DWORD X11DRV_GLX_FVFSize(DWORD fvf)
{
    static const DWORD position_size[] = { 0, 3, 4, 4, 5, 6, 7, 8 };

    DWORD size;

    size = position_size[(fvf & D3DFVF_POSITION_MASK) / 2] * sizeof(DWORD);

    if (fvf & D3DFVF_RESERVED1)
	size += sizeof(DWORD);

    if (fvf & D3DFVF_NORMAL)
	size += 3 * sizeof(float);

    if (fvf & D3DFVF_DIFFUSE)
	size += sizeof(D3DCOLOR);

    if (fvf & D3DFVF_SPECULAR)
	size += sizeof(D3DCOLOR);

    size += 2 * ((fvf & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT)
	* sizeof(float);

    return size;
}

static inline DWORD rot16(DWORD x)
{
    return (x >> 16) | (x << 16);
}

static inline void
X11DRV_GLX_ConvertVertex(LPCVOID _in, LPVOID _out, DWORD fvf)
{
    const DWORD* in = _in;
    LPDWORD out = _out;

    switch (fvf & D3DFVF_POSITION_MASK)
    {
    case D3DFVF_XYZRHW:
    {
	float w;
	const FLOAT* inf = (const FLOAT*)in;
	LPFLOAT outf = (LPFLOAT)out;

	/* Reportedly Microsoft uses 2^20 in their ProcessVertices impl. */
	if (inf[3] == 0.0) w = 1048576.0;
	else w = 1.0 / inf[3];

	outf[0] = inf[0] * w;
	outf[1] = inf[1] * w;
	outf[2] = inf[2] * w;
	outf[3] = w;

	in += 4;
	out += 4;
	}
    break;

    case D3DFVF_XYZB5:
	*out++ = *in++;
    case D3DFVF_XYZB4:
	*out++ = *in++;
    case D3DFVF_XYZB3:
	*out++ = *in++;
    case D3DFVF_XYZB2:
	*out++ = *in++;
    case D3DFVF_XYZB1:
	*out++ = *in++;
    case D3DFVF_XYZ:
	*out++ = *in++;
	*out++ = *in++;
	*out++ = *in++;
	break;
    }

    if (fvf & D3DFVF_NORMAL)
    {
	*out++ = *in++;
	*out++ = *in++;
	*out++ = *in++;
    }

    if (fvf & D3DFVF_RESERVED1)
    {
	out++;
	in++;
    }

    if (fvf & D3DFVF_DIFFUSE)
    {
	DWORD tmp = *in++;
	*out++ = (tmp & 0xFF00FF00) | (rot16(tmp) & 0x00FF00FF);
    }

    if (fvf & D3DFVF_SPECULAR)
    {
	DWORD tmp = *in++;
	*out++ = (tmp & 0xFF00FF00) | (rot16(tmp) & 0x00FF00FF);
    }

    switch ((fvf & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT)
    {
    case 8:
	*out++ = *in++;
	*out++ = *in++;
    case 7:
	*out++ = *in++;
	*out++ = *in++;
    case 6:
	*out++ = *in++;
	*out++ = *in++;
    case 5:
	*out++ = *in++;
	*out++ = *in++;
    case 4:
	*out++ = *in++;
	*out++ = *in++;
    case 3:
	*out++ = *in++;
	*out++ = *in++;
    case 2:
	*out++ = *in++;
	*out++ = *in++;
    case 1:
	*out++ = *in++;
	*out++ = *in++;
    case 0:
	break;
    }
}

static inline BOOL X11DRV_GLX_EB_IsUptodate(const BYTE* uptodate, DWORD index)
{
    return uptodate[index / 8] & (1 << (index % 8));
}

static inline void X11DRV_GLX_EB_MarkUptodate(LPBYTE uptodate, DWORD index)
{
    uptodate[index / 8] |= (1 << (index % 8));
}

static void X11DRV_GLX_EB_ResetUptodate(struct X11DRV_GLX_EB_Info* info)
{
    memset(info->uptodate, 0, (info->num_vertices + 7) / 8);
}

static void X11DRV_GLX_Convert(struct X11DRV_GLX_EB_Info* info,
			       DWORD offset, DWORD num,
			       const WORD* indices, DWORD nindices,
			       DWORD indexbase)
{
    const BYTE* in;
    LPBYTE out;

    DWORD fvf = info->original_fvf;
    DWORD in_size = info->original_fvf_size;
    DWORD out_size = info->converted_fvf_size;

    LPBYTE uptodate = info->uptodate;

    if (indices != NULL)
    {
	in = (LPBYTE)info->lock_buffer;
	out = (LPBYTE)info->render_buffer;

	TRACE("indices=%ld\n", nindices);
	while (nindices--)
	{
	    DWORD index = indexbase + *indices++;

	    if (!X11DRV_GLX_EB_IsUptodate(uptodate, index))
	    {
		X11DRV_GLX_EB_MarkUptodate(uptodate, index);
		X11DRV_GLX_ConvertVertex(in + index * in_size,
					 out + index * out_size,
					 fvf);
	    }
	}
    }
    else
    {
	in = (LPBYTE)info->lock_buffer + offset * in_size;
	out = (LPBYTE)info->render_buffer + offset * out_size;

	while (num--)
	{
	    if (!X11DRV_GLX_EB_IsUptodate(uptodate, offset))
	    {
		X11DRV_GLX_EB_MarkUptodate(uptodate, offset);
		X11DRV_GLX_ConvertVertex(in, out, fvf);
	    }

	    in += in_size;
	    out += out_size;
	    offset++;
	}
    }
}

static DWORD WINAPI X11DRV_GLX_CanCreateEB(LPDDHAL_CANCREATESURFACEDATA data)
{
    data->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/* not necessarily called with locks held */
static DWORD WINAPI X11DRV_GLX_CreateEB(LPDDHAL_CREATESURFACEDATA data)
{
    LPDDRAWI_DDRAWSURFACE_LCL lcl = *data->lplpSList;
    LPDDRAWI_DDRAWSURFACE_GBL gbl = lcl->lpGbl;
    LPDDRAWI_DDRAWSURFACE_MORE more = lcl->lpSurfMore;

    DWORD size = gbl->u4.dwLinearSize;
    DWORD fvf = more->dwFVF;

    struct X11DRV_GLX_EB_Info* info;
    DWORD num_vertices;
    DWORD uptodate_bytes;

    TRACE("\n");

    num_vertices = size / X11DRV_GLX_FVFSize(fvf);
    uptodate_bytes = (num_vertices + 7) / 8;

    info = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
		     sizeof(struct X11DRV_GLX_EB_Info) + uptodate_bytes);
    if (info == NULL)
    {
	data->ddRVal = E_OUTOFMEMORY;
	return DDHAL_DRIVER_HANDLED;
    }

    /* Don't waste VAR memory on command buffers and implicit vertex buffers.*/
    if (more->ddsCapsEx.dwCaps2 & DDSCAPS2_VERTEXBUFFER)
	info->var = X11DRV_GLX_VAR_InitBufferControl(&info->var_buffers, size);

    if (!info->var
	|| X11DRV_GLX_DoesFVFNeedConversion(fvf)
	|| (lcl->ddsCaps.dwCaps & DDSCAPS_SYSTEMMEMORY)
	|| !(lcl->ddsCaps.dwCaps & DDSCAPS_WRITEONLY))
    {
	info->system_buffer = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
					size);

	if (info->system_buffer == NULL)
	{
	    HeapFree(GetProcessHeap(), 0, info);

	    data->ddRVal = D3DERR_VBUF_CREATE_FAILED;
	    return DDHAL_DRIVER_HANDLED;
	}
    }

    if (info->system_buffer)
	info->lock_buffer = info->system_buffer;
    else
    {
	info->lock_buffer = X11DRV_GLX_VAR_GetCurrentBuffer(&info->var_buffers);
	info->lock_buffer_is_var = TRUE;
    }

    if (info->var)
    {
	info->render_buffer = X11DRV_GLX_VAR_GetCurrentBuffer(&info->var_buffers);
	info->render_buffer_is_var = TRUE;
    }
    else
	info->render_buffer = info->system_buffer;

    info->original_fvf = fvf;
    if (info->lock_buffer != info->render_buffer)
	info->converted_fvf = X11DRV_GLX_ConvertedFVF(fvf);
    else
	info->converted_fvf = fvf;

    TRACE("%lx %lx\n", info->original_fvf, info->converted_fvf);

    info->original_fvf_size = X11DRV_GLX_FVFSize(info->original_fvf);
    info->converted_fvf_size = X11DRV_GLX_FVFSize(info->converted_fvf);

    info->overwritten = FALSE;

    info->num_vertices = num_vertices;
    info->size = size;

    X11DRV_GLX_EB_ConvertInterleavedToGLStride(info->original_fvf, NULL,
					       &info->lock_strided_data);

    X11DRV_GLX_EB_ConvertInterleavedToGLStride(info->converted_fvf, NULL,
					       &info->render_strided_data);

    gbl->dwReserved1 = (DWORD)info;

    gbl->fpVidMem = (FLATPTR)info->lock_buffer;
    data->ddRVal = DD_OK;

    return DDHAL_DRIVER_HANDLED;
}

/* not necessarily called with locks held */
static DWORD WINAPI X11DRV_GLX_DestroyEB(LPDDHAL_DESTROYSURFACEDATA data)
{
    LPDDRAWI_DDRAWSURFACE_GBL gbl = data->lpDDSurface->lpGbl;
    struct X11DRV_GLX_EB_Info* info
	= (struct X11DRV_GLX_EB_Info*)gbl->dwReserved1;

    TRACE("\n");

    ENTER_GL();
    glXWaitX();
    X11DRV_GLX_VAR_FreeBufferControl(&info->var_buffers);
    glXWaitGL();
    LEAVE_GL();

    if (info->system_buffer)
    {
	HeapFree(GetProcessHeap(), 0, info->system_buffer);
    }

    gbl->fpVidMem = (FLATPTR)NULL;
    data->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

static void X11DRV_GLX_EB_Rename(struct X11DRV_GLX_EB_Info* info,
				 BOOL can_rename)
{
    LPVOID new_buffer;

    X11DRV_GLX_VAR_Overwrite(&info->var_buffers, can_rename);

    new_buffer = X11DRV_GLX_VAR_GetCurrentBuffer(&info->var_buffers);

    if (info->lock_buffer_is_var)
	info->lock_buffer = new_buffer;

    if (info->render_buffer_is_var)
	info->render_buffer = new_buffer;
}

static void X11DRV_GLX_EB_OverwriteLock(struct X11DRV_GLX_EB_Info* info,
					BOOL discard_contents)
{
    info->overwritten = TRUE;
    X11DRV_GLX_EB_ResetUptodate(info);

    if (info->lock_buffer_is_var)
	X11DRV_GLX_EB_Rename(info, discard_contents);
}

static void X11DRV_GLX_EB_OverwriteRender(struct X11DRV_GLX_EB_Info* info)
{
    if (info->render_buffer_is_var && !info->lock_buffer_is_var)
	X11DRV_GLX_EB_Rename(info, TRUE);
}

/* not necessarily called with locks held */
static DWORD WINAPI X11DRV_GLX_LockEB(LPDDHAL_LOCKDATA data)
{
    struct X11DRV_GLX_EB_Info* info
	= (struct X11DRV_GLX_EB_Info*)data->lpDDSurface->lpGbl->dwReserved1;

    TRACE("\n");

    if (!(data->dwFlags & (DDLOCK_READONLY|DDLOCK_NOOVERWRITE))) {
	ENTER_GL();
	glXWaitX();
	X11DRV_GLX_EB_OverwriteLock(info,
				    data->dwFlags & DDLOCK_DISCARDCONTENTS);
	glXWaitGL();
	LEAVE_GL();
    }

    data->lpSurfData = info->lock_buffer;

    data->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/* not necessarily called with locks held */
static DWORD WINAPI X11DRV_GLX_UnlockEB(LPDDHAL_UNLOCKDATA data)
{
    TRACE("\n");
    data->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

static DWORD X11DRV_GLX_D3dVbGetConvertedFVF(LPDDRAWI_DDRAWSURFACE_LCL vb)
{
    struct X11DRV_GLX_EB_Info* info
	= (struct X11DRV_GLX_EB_Info*)vb->lpGbl->dwReserved1;

    return info->converted_fvf;
}

static inline void
X11DRV_GLX_EB_SSBAOOne(D3DDP_PTRSTRIDE* in, D3DDP_PTRSTRIDE* out, LPVOID base,
		       DWORD ofs)
{
    out->lpvData = (LPBYTE)base + (DWORD)in->lpvData + ofs * in->dwStride;
    out->dwStride = in->dwStride;
}

/* If a GLDRAWPRIMITIVESTRIDEDDATA struct was generated with a NULL pointer,
 * then each stride pointer is actually just an offset relative to a vertex
 * pointer.
 *
 * This adds in the vertex pointer to create a real stride structure, then
 * shifts it by offset vertices.
 */
static void
X11DRV_GLX_EB_SetStridedBaseAndOffset(LPGLDRAWPRIMITIVESTRIDEDDATA in,
				      LPGLDRAWPRIMITIVESTRIDEDDATA out,
				      LPVOID base, DWORD offset)
{
    int i;

    X11DRV_GLX_EB_SSBAOOne(&in->position, &out->position, base, offset);
    X11DRV_GLX_EB_SSBAOOne(&in->normal, &out->normal, base, offset);
    X11DRV_GLX_EB_SSBAOOne(&in->color[0], &out->color[0], base, offset);
    X11DRV_GLX_EB_SSBAOOne(&in->color[1], &out->color[1], base, offset);

    for (i = 0; i < D3DDP_MAXTEXCOORD; i++)
	X11DRV_GLX_EB_SSBAOOne(&in->textureCoords[i],
			       &out->textureCoords[i], base, offset);

    for (i = 0; i < 4; i++)
	X11DRV_GLX_EB_SSBAOOne(&in->blend_weights[i],
			       &out->blend_weights[i], base, offset);

    X11DRV_GLX_EB_SSBAOOne(&in->matrix_indices, &out->matrix_indices, base,
			   offset);
    X11DRV_GLX_EB_SSBAOOne(&in->fog_coordinate, &out->fog_coordinate, base,
			   offset);
}

static void X11DRV_GLX_D3dVbPreRender(LPDDRAWI_DDRAWSURFACE_LCL vb,
				      LPGLDRAWPRIMITIVESTRIDEDDATA data,
				      BOOL convert, DWORD offset, DWORD num,
				      const WORD* indices, DWORD nindices,
				      DWORD indexbase)
{
    struct X11DRV_GLX_EB_Info* info
	= (struct X11DRV_GLX_EB_Info*)vb->lpGbl->dwReserved1;
    BOOL var;

    TRACE("(%p,%p,%d,%ld,%ld,%p,%ld,%ld)\n",vb,data,convert,offset,num,indices,nindices,indexbase);
    TRACE("lockbuf=%p, renderbuf=%p, overwritten=%ld\n",info->lock_buffer,info->render_buffer,info->overwritten);
    if (convert && info->lock_buffer != info->render_buffer)
    {
	/* As long as they only lock with DDLOCK_NOOVERWRITE, we don't
	 * need to wait since any in-use vertex data we overwrite will be
	 * overwritten with the previous data. */
	if (info->overwritten)
	    X11DRV_GLX_EB_OverwriteRender(info);

	X11DRV_GLX_Convert(info, offset, num, indices, nindices, indexbase);

	info->overwritten = FALSE;

	X11DRV_GLX_EB_SetStridedBaseAndOffset(&info->render_strided_data,
					      data, info->render_buffer,
					      0);

	var = info->render_buffer_is_var;
    }
    else
    {
	X11DRV_GLX_EB_SetStridedBaseAndOffset(&info->lock_strided_data,
					      data, info->lock_buffer,
					      0);
	var = info->lock_buffer_is_var;
    }

    TRACE("is_var=%d\n",var);
    if (var)
	glEnableClientState(GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV);
}

static void X11DRV_GLX_D3dVbPostRender(LPDDRAWI_DDRAWSURFACE_LCL vb)
{
    struct X11DRV_GLX_EB_Info* info
	= (struct X11DRV_GLX_EB_Info*)vb->lpGbl->dwReserved1;

    TRACE("(%p): render_is_var=%d\n",vb,info->render_buffer_is_var);
    if (info->render_buffer_is_var)
    {
#if 0
	GLboolean valid;
	glGetBooleanv(GL_VERTEX_ARRAY_RANGE_VALID_NV, &valid);
	if (!valid)
	{
	    void* p;
	    GLint x;

	    ERR("VAR not consistent\n");

	    if (!glXIsDirect(glXGetCurrentDisplay(), glXGetCurrentContext()))
		ERR("context not direct\n");

	    if (!strstr(glGetString(GL_EXTENSIONS),
			"GL_NV_vertex_array_range2"))
		ERR("VAR2 extension missing\n");

	    if (!strstr(glGetString(GL_EXTENSIONS), "GL_NV_fence"))
		ERR("fence extension missing\n");

	    if (!glIsEnabled(GL_VERTEX_ARRAY_RANGE_NV))
		ERR("VAR not enabled\n");

	    glGetPointerv(GL_VERTEX_ARRAY_RANGE_POINTER_NV, &p);
	    if ((DWORD)p % 32 != 0) ERR("pointer not aligned\n");

	    glGetIntegerv(GL_VERTEX_ARRAY_RANGE_LENGTH_NV, &x);

	    ERR("%p %d\n", p, x);

	    if (!glIsEnabled(GL_VERTEX_ARRAY))
		ERR("GL_VERTEX_ARRAY disabled\n");
	    else
	    {
		glGetIntegerv(GL_VERTEX_ARRAY_STRIDE, &x);
		if (x >= 256) ERR("vertex stride too large\n");
		if (x % 4 != 0) ERR("vertex stride unaligned\n");

		glGetPointerv(GL_VERTEX_ARRAY_POINTER, &p);
		if ((DWORD)p % 4 != 0) ERR("vertex pointer unaligned\n");

		glGetIntegerv(GL_VERTEX_ARRAY_SIZE, &x);
		if (x < 2 || x > 4) ERR("vertex array size wrong\n");
	    }

	    if (glIsEnabled(GL_NORMAL_ARRAY))
	    {
		glGetIntegerv(GL_NORMAL_ARRAY_STRIDE, &x);
		if (x >= 256) ERR("normal stride too large\n");
		if (x % 4 != 0) ERR("normal stride unaligned\n");

		glGetPointerv(GL_NORMAL_ARRAY_POINTER, &p);
		if ((DWORD)p % 4 != 0) ERR("normal pointer unaligned\n");
	    }

	    if (glIsEnabled(GL_COLOR_ARRAY))
	    {
		glGetIntegerv(GL_COLOR_ARRAY_TYPE, &x);
		if (x != GL_UNSIGNED_BYTE) ERR("color array type wrong\n");

		glGetIntegerv(GL_COLOR_ARRAY_STRIDE, &x);
		if (x >= 256) ERR("color stride too large\n");
		if (x % 4 != 0) ERR("color stride unaligned\n");

		glGetPointerv(GL_COLOR_ARRAY_POINTER, &p);
		if ((DWORD)p % 4 != 0) ERR("color pointer unaligned\n");

		glGetIntegerv(GL_COLOR_ARRAY_SIZE, &x);
		if (x != 4) ERR("color array size wrong\n");
	    }

	    if (glIsEnabled(GL_SECONDARY_COLOR_ARRAY_EXT))
	    {
		glGetIntegerv(GL_SECONDARY_COLOR_ARRAY_TYPE_EXT, &x);
		if (x != GL_UNSIGNED_BYTE) ERR("2color array type wrong\n");

		glGetIntegerv(GL_SECONDARY_COLOR_ARRAY_STRIDE_EXT, &x);
		if (x >= 256) ERR("2color stride too large\n");
		if (x % 4 != 0) ERR("2color stride unaligned\n");

		glGetPointerv(GL_SECONDARY_COLOR_ARRAY_POINTER_EXT, &p);
		if ((DWORD)p % 4 != 0) ERR("2color pointer unaligned\n");

		glGetIntegerv(GL_SECONDARY_COLOR_ARRAY_SIZE_EXT, &x);
		if (x != 4) ERR("2color array size wrong\n");
	    }

	    if (glIsEnabled(GL_TEXTURE_COORD_ARRAY))
	    {
		glGetIntegerv(GL_TEXTURE_COORD_ARRAY_TYPE, &x);
		if (x != GL_FLOAT) ERR("tex array type wrong\n");

		glGetIntegerv(GL_TEXTURE_COORD_ARRAY_STRIDE, &x);
		if (x >= 256) ERR("tex stride too large\n");
		if (x % 4 != 0) ERR("tex stride unaligned\n");

		glGetPointerv(GL_TEXTURE_COORD_ARRAY_POINTER, &p);
		if ((DWORD)p % 4 != 0) ERR("tex pointer unaligned\n");

		glGetIntegerv(GL_TEXTURE_COORD_ARRAY_SIZE, &x);
		if (x < 1 || x > 4) ERR("tex array size wrong\n");
	    }

	    if (glIsEnabled(GL_EDGE_FLAG_ARRAY))
		ERR("edge flag array enabled\n");

#if 0
	    if (glIsEnabled(GL_VERTEX_WEIGHT_ARRAY_NV))
		ERR("vertex weight array enabled\n");
#endif

	    if (glIsEnabled(GL_FOG_COORDINATE_ARRAY_EXT))
		ERR("fog coordinary array enabled\n");

	    glGetIntegerv(GL_LIGHT_MODEL_TWO_SIDE, &x);
	    if (x) ERR("two-sided lighting is on\n");
	}
#endif

	/* It would be nice to only SetFence if we actually rendered
	 * from VAR memory. */
	X11DRV_GLX_glSetFenceNV(X11DRV_GLX_VAR_GetCurrentFence(&info->var_buffers),
				GL_ALL_COMPLETED_NV);

	glDisableClientState(GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV);
    }
}

DDHAL_DDEXEBUFCALLBACKS X11DRV_GLX_exebufcallbacks = {
  sizeof(DDHAL_DDEXEBUFCALLBACKS),
  0x1f, /* all callbacks are 32-bit */
  X11DRV_GLX_CanCreateEB, /* CanCreateExecuteBuffer */
  X11DRV_GLX_CreateEB, /* CreateExecuteBuffer */
  X11DRV_GLX_DestroyEB, /* DestroyExecuteBuffer */
  X11DRV_GLX_LockEB, /* LockExecuteBuffer */
  X11DRV_GLX_UnlockEB  /* UnlockExecuteBuffer */
};

struct GL_D3D_DriverCallbacks X11DRV_GLX_GL_D3D_driver_callbacks =
{
    sizeof(struct GL_D3D_DriverCallbacks),
    X11DRV_GLX_D3dVbGetConvertedFVF,
    X11DRV_GLX_D3dVbPreRender,
    X11DRV_GLX_D3dVbPostRender
};
