/*
 * Direct3D GLX interface
 *
 * Copyright 2001 TransGaming Technologies, Inc.
 */

#include "config.h"

#include "ts_xlib.h"
#include "x11drv.h"
#include "x11font.h"
#include "x11ddraw.h"
#include "glx.h"
#include "executebuffer.h"
#include "d3d.h"
#include "d3dhal.h"
#include "wine_gl.h"
#include "wine/d3dhalgl.h"

#include "windef.h"
#include "wingdi.h"
#include "ddrawi.h"
#include "bitmap.h"
#include "win.h"
#include "debugtools.h"
#include "options.h"

DEFAULT_DEBUG_CHANNEL(x11drv);

#define RENDER_TO_WINDOW
/* #define TRANSGAMING_LOGO */
/* #define NO_INDIRECT */
#define ENABLE_BORKEN_MIGRATE

DWORD X11DRV_GLX_TLS;

static BOOL glx_available = FALSE;

void X11DRV_GLX_Init(void)
{
  int glx_event, glx_error, glx_major, glx_minor;

  if (glx_available) return; /* already initialized? */

  ENTER_GL();
  if (glXQueryExtension(gdi_display, &glx_event, &glx_error) &&
      glXQueryVersion(gdi_display, &glx_major, &glx_minor)) {

    TRACE("GLX extension detected and enabled\n");
    glx_available = TRUE;

    X11DRV_GLX_TLS = TlsAlloc();

    X11DRV_GLX_EB_Init();
  }
  LEAVE_GL();
}

void X11DRV_GLX_Cleanup(void)
{
  X11DRV_GLX_EB_Cleanup();
  if (TlsFree(X11DRV_GLX_TLS))
    X11DRV_GLX_TLS = 0;
}

/* actual Direct3D HAL stuff */

#define CACHESIZE 2

typedef struct _D3DTCONTEXT {
  struct _D3DCONTEXT* parent;
  HANDLE	thread;
  GLXContext	ctx;
  GLXContext	save_ctx;
  GLXDrawable	save_draw;
  DWORD		stamp;
  struct _D3DTCONTEXT*	prev;
  struct _D3DTCONTEXT*	next;
} D3DTCONTEXT,*LPD3DTCONTEXT;

typedef struct {
  GLXDrawable draw;
  DWORD flags;
  HBITMAP bmp;
  X11DRV_DIBSECTION *dib;
  BITMAPOBJ *lockedBmp;
  DWORD locks, lockf;
  BOOL rect_lock;
  RECTL rect;
} D3DTARGET,*LPD3DTARGET;

typedef struct _D3DCONTEXT {
  D3DTCONTEXT	p;

  XVisualInfo*	vis;
  DWORD		owner_tid;
  GLXContext	owner_ctx;
  BOOL		borked;
  DWORD		refs;
  LPDDRAWI_DDRAWSURFACE_LCL lsurf, zsurf, nsurf;
  LPDDRAWI_DDRAWSURFACE_GBL surf;
  Font		lfont;

  D3DTARGET	zbuf;
  D3DTARGET	cache[CACHESIZE];
  DWORD		curp;

  GLint		texunits;
  void (*glActiveTextureARB)(GLenum texture);

  int		glx_major, glx_minor;

  /* SGIX_fbconfig */
  int (*glXGetFBConfigAttribSGIX)(Display*,GLXFBConfig,int,int*);
  GLXFBConfig* (*glXChooseFBConfigSGIX)(Display*,int,const int*,int*);
  GLXPixmap (*glXCreateGLXPixmapWithConfigSGIX)(Display*,GLXFBConfig,Pixmap);
  GLXContext (*glXCreateContextWithConfigSGIX)(Display*,GLXFBConfig,int,GLXContext,Bool);
  XVisualInfo* (*glXGetVisualFromFBConfigSGIX)(Display*,GLXFBConfig);
  GLXFBConfig (*glXGetFBConfigFromVisualSGIX)(Display*,XVisualInfo*);

  /* SGIX_pbuffer */
  GLXPbuffer (*glXCreateGLXPbufferSGIX)(Display*,GLXFBConfig,unsigned int,unsigned int,int*);
  void (*glXDestroyGLXPbufferSGIX)(Display*,GLXPbuffer);
  void (*glXQueryGLXPbufferSGIX)(Display*,GLXPbuffer,int,unsigned int);
  void (*glXSelectEventSGIX)(Display*,GLXDrawable,unsigned long);
  void (*glXGetSelectedEventSGIX)(Display*,GLXDrawable,unsigned long*);

#ifdef TRANSGAMING_LOGO
  GLuint logo;
  GLuint logo_list;
  DWORD logo_size;
#endif
} D3DCONTEXT,*LPD3DCONTEXT;

#define FLAG_PIXMAP	0x0001
#define FLAG_CINDIRECT	0x0002
#define FLAG_PINDIRECT	0x0004
#define FLAG_WINDIRECT  0x0008
#define FLAG_PDIRECT	0x0010
#define FLAG_WDIRECT	0x0020
#define FLAG_INDIRECT	(FLAG_PINDIRECT|FLAG_WINDIRECT)
#define FLAG_DIRECT	(FLAG_PDIRECT|FLAG_WDIRECT)
#define FLAG_BACKBUF	0x0040
#define FLAG_FLIPPED	0x0080
#define FLAG_PBUFFER	0x0100
#define FLAG_OFFSCREEN	(FLAG_PIXMAP|FLAG_PBUFFER)
#define FLAG_ZBUFFER	0x0200
#define FLAG_DESTROYED	0x8000

static DWORD X11DRV_GLX_MakeCurrent(LPD3DCONTEXT ctx);
static BOOL X11DRV_GLX_KillCurrent(LPD3DCONTEXT ctx);

static void X11DRV_GLX_CleanTarget(LPD3DCONTEXT ctx, LPD3DTARGET tgt);
static void X11DRV_GLX_DestroyTarget(LPD3DCONTEXT ctx, LPD3DTARGET tgt);

static void X11DRV_GLX_CleanTargets(LPD3DCONTEXT ctx);
static void X11DRV_GLX_DestroyTargets(LPD3DCONTEXT ctx);

static LPD3DTARGET X11DRV_GLX_FindTargetByDIB(LPD3DCONTEXT ctx, X11DRV_DIBSECTION* dib);

static void X11DRV_GLX_UpdateTexture(LPD3DCONTEXT ctx);

static void X11DRV_GLX_SetupLogo(LPD3DCONTEXT ctx);
static void X11DRV_GLX_RenderLogo(LPD3DCONTEXT ctx);

static inline LPD3DTCONTEXT X11DRV_GLX_GetCurrentT(void)
{
  return TlsGetValue(X11DRV_GLX_TLS);
}

static inline LPD3DCONTEXT X11DRV_GLX_GetCurrent(void)
{
  LPD3DTCONTEXT ctx = X11DRV_GLX_GetCurrentT();
  return ctx ? ctx->parent : NULL;
}

static inline void X11DRV_GLX_SetCurrentT(LPD3DTCONTEXT ctx)
{
  TlsSetValue(X11DRV_GLX_TLS, ctx);
}

static int GLXError;
static int GLXErrorHandler(Display *dpy, XErrorEvent *event)
{
  TRACE("(type=%d,code=%d,major=%d,minor=%d,serial=%ld,id=0x%08lx)\n",
    event->type, event->error_code,
    event->request_code, event->minor_code,
    event->serial, event->resourceid);
  GLXError = 1;
  return 0;
}

static unsigned X11DRV_GLX_Format(LPDIBSECTION dib, GLenum *fmt, GLenum *type)
{
  switch (dib->dsBmih.biBitCount) {
  case 15:
    *fmt=GL_BGRA; *type=GL_UNSIGNED_SHORT_1_5_5_5_REV;
    return 2;
  case 16:
    *fmt=GL_RGB; *type=GL_UNSIGNED_SHORT_5_6_5;
    return 2;
  case 24:
    *fmt=GL_BGR; *type=GL_UNSIGNED_BYTE;
    return 3;
  case 32:
    *fmt=GL_BGRA; *type=GL_UNSIGNED_BYTE;
    return 4;
  default:
    ERR("unhandled DIB depth %d\n", dib->dsBmih.biBitCount);
    *fmt=0; *type=0;
    return 0;
  }
}

static unsigned X11DRV_GLX_ZFormat(LPDIBSECTION dib, GLenum *fmt, GLenum *type)
{
  *fmt = GL_DEPTH_COMPONENT;
  switch (dib->dsBmih.biBitCount) {
  case 8:
    *type = GL_UNSIGNED_BYTE;
    return 1;
  case 15:
  case 16:
    *type = GL_UNSIGNED_SHORT;
    return 2;
  case 24:
  case 32:
    *type = GL_UNSIGNED_INT;
    return 4;
  default:
    ERR("unhandled DIB depth %d\n", dib->dsBmih.biBitCount);
    *fmt=0; *type=0;
    return 0;
  }
}

static BOOL X11DRV_GLX_Destroy(LPD3DCONTEXT ctx)
{
  BOOL ret;

  X11DRV_GLX_CleanTargets(ctx);
  X11DRV_GLX_CleanTarget(ctx, &ctx->zbuf);
  ret = X11DRV_GLX_KillCurrent(ctx);
  X11DRV_GLX_DestroyTargets(ctx);
  XFree(ctx->vis);
  return ret;
}

static inline RECTL* X11DRV_GLX_GetRect(LPD3DTARGET tgt)
{
  return tgt->rect_lock ? &tgt->rect : NULL;
}

static inline D3DTARGET* X11DRV_GLX_GetCurTgt(LPD3DCONTEXT ctx)
{
  return &ctx->cache[ctx->curp];
}

static inline DWORD X11DRV_GLX_GetFlags(LPD3DCONTEXT ctx)
{
  return X11DRV_GLX_GetCurTgt(ctx)->flags;
}

/* This function is called with GLX resources locked */
static void X11DRV_GLX_CopyDIB(LPD3DCONTEXT ctx, LPD3DTARGET tgt)
{
  BITMAPOBJ *bmp;
  X11DRV_DIBSECTION *dib;
  GLenum fmt, type;
  unsigned w, h, p, bpp;
  RECTL* rect;
  RECTL def;
  LPBYTE bits;
  int i;

  TRACE("copying from %s DIBsection to GL buffer\n", (tgt->flags&FLAG_ZBUFFER)?"Zbuffer":"RGB");

  bmp = tgt->lockedBmp;
  if (!bmp)
    return;

  dib = (X11DRV_DIBSECTION *)bmp->dib;

  /* make the DIB readable for us */
  if (dib->status == DIB_Status_GdiMod) {
    TRACE("state is GdiMod, converting to InSync\n");
    glXWaitGL();
    X11DRV_DIB_Coerce(bmp, DIB_Status_InSync, FALSE);
    dib->p_status = DIB_Status_InSync;
    glXWaitX();
  }
  w = dib->dibSection.dsBm.bmWidth;
  h = dib->dibSection.dsBm.bmHeight;
  p = dib->dibSection.dsBm.bmWidthBytes;
  bits = dib->dibSection.dsBm.bmBits;
  if (tgt->flags & FLAG_ZBUFFER) {
    bpp = X11DRV_GLX_ZFormat(&dib->dibSection, &fmt, &type);
  } else {
    bpp = X11DRV_GLX_Format(&dib->dibSection, &fmt, &type);
  }
  TRACE("fmt=%x, type=%x, w=%d, h=%d, p=%d, bits=%p, state=%d\n", fmt, type, w, h, p, bits, dib->status);
  rect = X11DRV_GLX_GetRect(tgt);
  if (!rect) {
    def.left = 0;
    def.top = 0;
    def.right = w;
    def.bottom = h;
    rect = &def;
  } else {
    bits += rect->top * p + rect->left * bpp;
  }
  TRACE("rect: (%ld,%ld)-(%ld,%ld)\n", rect->left, rect->top, rect->right, rect->bottom);

  /* Save GL state info.
   * glPixelZoom changes stuff in GL_PIXEL_MODE_BIT, but nobody else calls
   * glDrawPixels or glCopyPixels yet. */
  glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_COLOR_BUFFER_BIT |
               GL_DEPTH_BUFFER_BIT | GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);

  /* Set up the transformation so that we can set the rasterpos appropriately */
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0.0, w, h, 0.0, -1, 1);

  /* Disable OpenGL features that might slow down glDrawPixels. */
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_DITHER);
  glDisable(GL_FOG);
  glDisable(GL_LIGHTING);
  glDisable(GL_SCISSOR_TEST);
  glDisable(GL_STENCIL_TEST);
  if (ctx->texunits > 1) {
    for (i=0; i<ctx->texunits; i++) {
      ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
      glDisable(GL_TEXTURE_2D);
    }
  }
  else glDisable(GL_TEXTURE_2D);

  if (fmt == GL_DEPTH_COMPONENT) {
    glDepthFunc(GL_ALWAYS);
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
  } else {
    glDisable(GL_DEPTH_TEST);
  }

  /* Transfer pixels. */
  if (fmt == GL_DEPTH_COMPONENT)
  {
      /* This is much faster on nVidia and Mesa-3.5.
       * Don't know about others. */
      DWORD y, dw;

      glViewport(0, 0, ctx->lsurf->lpGbl->wWidth, ctx->lsurf->lpGbl->wHeight);
      glPixelZoom(1, 1);

      dw = rect->right - rect->left;
      for (y = rect->top; y < rect->bottom; y++)
      {
	  glRasterPos2i(rect->left, y);
	  glDrawPixels(dw, 1, fmt, type, bits);
	  bits += p;
      }
  }
  else
  {
      /* We need Y=1 as a side-effect of using glPixelZoom. */
      glViewport(0, 1, ctx->lsurf->lpGbl->wWidth, ctx->lsurf->lpGbl->wHeight);

      glPixelZoom(1, -1);
      glRasterPos2i(rect->left, rect->top);
      glPixelStorei(GL_UNPACK_ROW_LENGTH, p/bpp); /* the unpack alignment takes care of p%bpp */
      glDrawPixels(rect->right-rect->left, rect->bottom-rect->top, fmt, type, bits);
      glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
  }

  /* Restore GL state info and finish */
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPopAttrib();

  dib->status = DIB_Status_AuxMod;
  TRACE("complete\n");
}

static void X11DRV_GLX_copy_aux(BITMAPOBJ*bmp, void*c, int req)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)c;
  X11DRV_DIBSECTION *dib = (X11DRV_DIBSECTION *)bmp->dib;
  LPD3DTARGET tgt = X11DRV_GLX_FindTargetByDIB(ctx, dib);
  GLenum fmt, type;
  unsigned w, h, p, bpp;
  RECTL* rect;
  RECTL def;
  LPBYTE bits;
  DWORD back;

  X11DRV_DIB_Coerce(bmp, DIB_Status_AppMod, TRUE);

  if (!tgt) {
    ERR("render target not found!\n");
    return;
  }

  w = dib->dibSection.dsBm.bmWidth;
  h = dib->dibSection.dsBm.bmHeight;
  p = dib->dibSection.dsBm.bmWidthBytes;
  bits = dib->dibSection.dsBm.bmBits;
  back = tgt->flags & FLAG_BACKBUF;
  if (tgt->flags & FLAG_ZBUFFER) {
    bpp = X11DRV_GLX_ZFormat(&dib->dibSection, &fmt, &type);
  } else {
    bpp = X11DRV_GLX_Format(&dib->dibSection, &fmt, &type);
    /* check whether we have done a GLX flip since the last rendering */
    if (X11DRV_GLX_GetFlags(ctx) & FLAG_FLIPPED)
      back ^= FLAG_BACKBUF;
  }
  TRACE("copying from %s buffer to %s DIBsection\n", back?"GL_BACK":"GL_FRONT", (tgt->flags&FLAG_ZBUFFER)?"Zbuffer":"RGB");
  TRACE("fmt=%x, type=%x, w=%d, h=%d, p=%d, bits=%p, state=%d\n", fmt, type, w, h, p, bits, dib->status);
  rect = X11DRV_GLX_GetRect(tgt);
  if (!rect) {
    def.left = 0;
    def.top = 0;
    def.right = w;
    def.bottom = h;
    rect = &def;
  } else {
    bits += rect->top * p + rect->left * bpp;
  }

  ENTER_GL();

  /* should we set the context? */
  if (SUCCEEDED(X11DRV_GLX_MakeCurrent(ctx))) {
    DWORD y, dw;

    TRACE("rect: (%ld,%ld)-(%ld,%ld)\n", rect->left, rect->top, rect->right, rect->bottom);
    glFinish();
    glReadBuffer(back ? GL_BACK : GL_FRONT);
    /* because OpenGL's coordinates are upside-down relative to the DIBs DirectDraw uses,
     * we must read out pixel data line by line into app memory */
    dw = rect->right - rect->left;
    for (y = rect->top; y < rect->bottom; y++) {
      glReadPixels(rect->left, h-y, dw, 1, fmt, type, bits);
      bits += p;
    }
  }
  else {
    ERR("failed to update DIB section from OpenGL frame buffer\n");
  }
  glXWaitGL();
  LEAVE_GL();
  TRACE("complete\n");
}

#define FONT_BASE 4200

BOOL X11DRV_GLX_write_aux(BITMAPOBJ *bmp, void*c, DC *dc, INT x, INT y, UINT flags,
                          const RECT *lprect, LPCWSTR wstr, UINT count,
                          const INT *lpDx)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)c;
  X11DRV_PDEVICE *physDev = (X11DRV_PDEVICE *)dc->physDev;
  fontObject *pfo = XFONT_GetFontObject( physDev->font );
  XFontStruct *font = pfo->fs;
  int i;
  GLint rp[2];

  TRACE("hdc=%04x df=%04x %d,%d %s, %d  flags=%d lpDx=%p\n",
        dc->hSelf, (UINT16)(physDev->font), x, y,
        debugstr_wn (wstr, count), count, flags, lpDx);
  /* if an app draws onto a surface used for GLX rendering, it's probably
   * not very advanced text, we'll only handle common cases */
  /* hmm, we ought to be able to handle ETO_OPAQUE and ETO_CLIPPED? */
  if (flags) goto write_fallback;
  if (lpDx) goto write_fallback;
  /* we're not Unicode-aware yet */
  for (i=0; i<count; i++) if (wstr[i] >= 256) goto write_fallback;

  rp[0] = INTERNAL_XWPTODP( dc, x, y );
  rp[1] = INTERNAL_YWPTODP( dc, x, y ) + font->ascent+1;

  TRACE("pos: %d, %d, color: %08lx\n", rp[0], rp[1], dc->textColor);

  ENTER_GL();
  /* should we set the context? */
  if (SUCCEEDED(X11DRV_GLX_MakeCurrent(ctx))) {
    if (font->fid != ctx->lfont) {
      TRACE("generating gdi_display lists for font %ld\n", font->fid);
      glXUseXFont(font->fid, 0, 256, FONT_BASE);
      ctx->lfont = font->fid;
    }

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT |
                 GL_LIST_BIT | GL_TRANSFORM_BIT);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0.0, ctx->lsurf->lpGbl->wWidth, ctx->lsurf->lpGbl->wHeight, 0.0, -1, 1);

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_STENCIL_TEST);
    if (ctx->texunits > 1) {
      for (i=0; i<ctx->texunits; i++) {
        ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
        glDisable(GL_TEXTURE_2D);
      }
    }
    else glDisable(GL_TEXTURE_2D);

    glColor3ub(GetRValue(dc->textColor),
               GetGValue(dc->textColor),
               GetBValue(dc->textColor));
    glRasterPos2iv(rp);

    glListBase(FONT_BASE);
    glCallLists(count, GL_UNSIGNED_SHORT, wstr);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glPopAttrib();

    X11DRV_GLX_UpdateTexture(ctx);

    glXWaitGL();
    LEAVE_GL();
  }
  else {
    LEAVE_GL();
    goto write_fallback;
  }

  return TRUE;

write_fallback:
  return X11DRV_ExtTextOut(dc, x, y, flags, lprect, wstr, count, lpDx);
}

static X11DRV_DIBSECTION *X11DRV_GLX_AttachAux(LPD3DCONTEXT ctx, HBITMAP hBmp)
{
   BITMAPOBJ *bmp;
   X11DRV_DIBSECTION *dib;

   /* attach aux handlers to DIB sections
    * (hoping no more than one context uses it) */
   bmp = (BITMAPOBJ *)GDI_GetObjPtr(hBmp, BITMAP_MAGIC);
   dib = (X11DRV_DIBSECTION *)bmp->dib;
   if (dib) {
     dib->aux_ctx = ctx;
     dib->copy_aux = X11DRV_GLX_copy_aux;
     dib->write_aux = X11DRV_GLX_write_aux;
   }
   GDI_ReleaseObj(hBmp);
   TRACE("aux handler attached to %p (%p)\n", bmp, dib);
   return dib;
}

static void X11DRV_GLX_MakeZbuffer(LPD3DCONTEXT ctx, LPDDRAWI_DDRAWSURFACE_LCL lsurf)
{
  LPDDRAWI_DDRAWSURFACE_GBL surf = lsurf ? lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
  HBITMAP hBmp = surf ? GET_LPDDRAWSURFACE_GBL_MORE(surf)->hKernelSurface : (HBITMAP)0;
  if (hBmp) {
    X11DRV_DIBSECTION *dib;
    TRACE("setting surf %p to ctx %p\n", lsurf, ctx);
    dib = X11DRV_GLX_AttachAux(ctx, hBmp);
    X11DRV_GLX_CleanTarget(ctx, &ctx->zbuf);
    ctx->zbuf.dib = dib;
    ctx->zbuf.bmp = hBmp;
  }
  ctx->zsurf = lsurf;
}

static void X11DRV_GLX_LockTarget(LPDDRAWI_DDRAWSURFACE_LCL lsurf, BOOL lock)
{
  LPDDRAWI_DDRAWSURFACE_GBL surf = lsurf ? lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
  HBITMAP hBmp = surf ? GET_LPDDRAWSURFACE_GBL_MORE(surf)->hKernelSurface : (HBITMAP)0;
  if (!hBmp) return;
  if (lock) GDI_GetObjPtr( hBmp, BITMAP_MAGIC );
  else GDI_ReleaseObj( hBmp );
}

static void X11DRV_GLX_CleanTarget(LPD3DCONTEXT ctx, LPD3DTARGET tgt)
{
  TRACE("(%p)\n", tgt);
  if (tgt->bmp && tgt->draw) {
    if (!(tgt->flags & FLAG_PIXMAP)) {
      X11DRV_CoerceDIBSection2(tgt->bmp, DIB_Status_InSync, FALSE);
    }
  }
}

static void X11DRV_GLX_DestroyTarget(LPD3DCONTEXT ctx, LPD3DTARGET tgt)
{
  TRACE("(%p)\n", tgt);
  if (tgt->bmp) {
    if (tgt->draw) {
      if (tgt->flags & FLAG_PIXMAP) {
        TRACE("destroying GLXpixmap %08lx\n", tgt->draw);
        glXDestroyGLXPixmap(gdi_display, tgt->draw);
      }
      else if (tgt->flags & FLAG_PBUFFER) {
        TRACE("destroying Pbuffer %08lx\n", tgt->draw);
        ctx->glXDestroyGLXPbufferSGIX(gdi_display, tgt->draw);
      }
    }
    tgt->draw = None;
    tgt->bmp = 0;
    tgt->dib = 0;
    if (tgt->lockedBmp) ERR("destroying captured target\n");
    tgt->lockedBmp = NULL;
    if (tgt->locks) ERR("destroying locked target\n");
    tgt->locks = 0;
    tgt->lockf = 0;
    tgt->rect_lock = FALSE;
  }
  tgt->flags &= ~FLAG_OFFSCREEN;
}

static void X11DRV_GLX_CleanTargets(LPD3DCONTEXT ctx)
{
  DWORD c;
  for (c=0; c<CACHESIZE; c++)
    X11DRV_GLX_CleanTarget(ctx, &ctx->cache[c]);
}

static void X11DRV_GLX_DestroyTargets(LPD3DCONTEXT ctx)
{
  DWORD c;
  for (c=0; c<CACHESIZE; c++)
    X11DRV_GLX_DestroyTarget(ctx, &ctx->cache[c]);
}

static void X11DRV_GLX_SwitchToTarget(LPD3DCONTEXT ctx, LPD3DTARGET tgt)
{
  DWORD c = tgt - ctx->cache;
  LPD3DTARGET last;

  TRACE("(%p)\n", tgt);
  if (c>=CACHESIZE) {
    ERR("bad target %p (%p+%ld)\n", tgt, ctx->cache, c);
    return;
  }
  if (c == ctx->curp) return;
  last = &ctx->cache[ctx->curp];
  /* migrate flags */
  tgt->flags = last->flags;
  if (last->flags & FLAG_FLIPPED)
    last->flags = (last->flags^FLAG_BACKBUF)&~FLAG_FLIPPED;
  ctx->curp = c;
}

static LPD3DTARGET X11DRV_GLX_NewTarget(LPD3DCONTEXT ctx)
{
  DWORD c, n, next;
  LPD3DTARGET tgt;
  next = (ctx->curp+1)%CACHESIZE;
  for (n=ctx->curp,c=0; c<CACHESIZE; c++) {
    if (!ctx->cache[n].draw) {
      next = n;
      break;
    }
    n = (n+1)%CACHESIZE;
  }
  tgt = &ctx->cache[next];
  X11DRV_GLX_SwitchToTarget(ctx, tgt);
  /* for now, the caller must call CleanTarget and DestroyTarget */
  return tgt;
}

static LPD3DTARGET X11DRV_GLX_FindTargetByHBMP(LPD3DCONTEXT ctx, HBITMAP hBmp)
{
  DWORD c;

  if (ctx->zbuf.bmp == hBmp) return &ctx->zbuf;
  for (c=0; c<CACHESIZE; c++)
    if (ctx->cache[c].bmp == hBmp) return &ctx->cache[c];
  return NULL;
}

static LPD3DTARGET X11DRV_GLX_FindTargetByDIB(LPD3DCONTEXT ctx, X11DRV_DIBSECTION* dib)
{
  DWORD c;

  if (ctx->zbuf.dib == dib) return &ctx->zbuf;
  for (c=0; c<CACHESIZE; c++)
    if (ctx->cache[c].dib == dib) return &ctx->cache[c];
  return NULL;
}

static LPD3DTARGET X11DRV_GLX_FindTarget(LPD3DCONTEXT ctx, LPDDRAWI_DDRAWSURFACE_LCL lsurf)
{
  LPDDRAWI_DDRAWSURFACE_GBL surf = lsurf ? lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
  HBITMAP hBmp = surf ? GET_LPDDRAWSURFACE_GBL_MORE(surf)->hKernelSurface : (HBITMAP)0;
  return X11DRV_GLX_FindTargetByHBMP(ctx, hBmp);
}

static void X11DRV_GLX_CreateOffscreen(LPD3DCONTEXT ctx, LPD3DTARGET tgt, HBITMAP hbmp)
{
  BITMAPOBJ *bmp;
  GLXFBConfig config = None;
  Drawable draw;

  tgt->draw = 0;
  tgt->flags &= ~FLAG_OFFSCREEN;

  bmp = (BITMAPOBJ *)GDI_GetObjPtr( hbmp, BITMAP_MAGIC );
  tgt->bmp = hbmp;
  tgt->dib = (X11DRV_DIBSECTION *)bmp->dib;

  /* FIXME: use ChooseFBConfig instead? */
  if (ctx->glXGetFBConfigFromVisualSGIX)
    config = ctx->glXGetFBConfigFromVisualSGIX(gdi_display, ctx->vis);

  if (ctx->glXCreateGLXPbufferSGIX) {
    int attribs[] = {None};
    TRACE("creating Pbuffer with size %dx%d, config %p\n",
          bmp->bitmap.bmWidth, bmp->bitmap.bmHeight, config);
    tgt->draw = ctx->glXCreateGLXPbufferSGIX(gdi_display, config,
                                                 bmp->bitmap.bmWidth,
                                                 bmp->bitmap.bmHeight,
                                                 attribs);
    TRACE("created Pbuffer %08lx\n", tgt->draw);
    tgt->flags |= FLAG_PBUFFER;
    X11DRV_GLX_AttachAux(ctx, hbmp);
    goto finish;
  }

  if (!bmp->physBitmap)
    X11DRV_CreateBitmap(hbmp);
  draw = (Pixmap)bmp->physBitmap;

  /* FIXME: try ctx->glXCreateGLXPixmapWithConfigSGIX, perhaps? */

  tgt->draw = glXCreateGLXPixmap(gdi_display, ctx->vis, draw);
  TRACE("created GLXpixmap %08lx\n", tgt->draw);
  tgt->flags |= FLAG_PIXMAP;

finish:
  GDI_ReleaseObj( hbmp );
}

/* OK, this procedure is getting unmaintainable... */

static void X11DRV_GLX_MakeTarget(LPD3DCONTEXT ctx, LPDDRAWI_DDRAWSURFACE_LCL lsurf)
{
  LPDDRAWI_DDRAWSURFACE_GBL surf = lsurf ? lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
  HBITMAP hBmp = surf ? GET_LPDDRAWSURFACE_GBL_MORE(surf)->hKernelSurface : (HBITMAP)0;
  LPD3DTARGET tgt;
  BOOL is_new = FALSE;

  ENTER_GL();
  tgt = hBmp ? X11DRV_GLX_FindTargetByHBMP(ctx, hBmp) : NULL;
  if (tgt) X11DRV_GLX_SwitchToTarget(ctx, tgt);
  else {
    tgt = X11DRV_GLX_NewTarget(ctx);
    is_new = TRUE;
    X11DRV_GLX_CleanTarget(ctx, tgt);
  }
  tgt->flags &= ~FLAG_FLIPPED;
  if (hBmp) {
    /* save this context in surface (hoping only one context uses it) */
    lsurf->dwReserved1 = (DWORD)ctx;
    TRACE("setting surf %p to ctx %p\n", lsurf, ctx);
    if (lsurf->ddsCaps.dwCaps & DDSCAPS_BACKBUFFER) {
      /* find frontbuffer */
      LPATTACHLIST from = lsurf->lpAttachListFrom;
      LPDDRAWI_DDRAWSURFACE_LCL front = from ? from->lpAttached : NULL;
      if (front) {
        front->dwReserved1 = (DWORD)ctx;
        TRACE("setting front surf %p to ctx %p\n", front, ctx);
      }
    }
#ifdef RENDER_TO_WINDOW
    if (lsurf->ddsCaps.dwCaps & (DDSCAPS_PRIMARYSURFACE | DDSCAPS_BACKBUFFER)) {
      if (tgt->flags & FLAG_OFFSCREEN) {
        X11DRV_GLX_CleanTargets(ctx);
        /* destroy context if it's indirect for pixmaps only */
        if (ctx->p.ctx && ((tgt->flags & FLAG_INDIRECT) == FLAG_PINDIRECT)) {
          X11DRV_GLX_KillCurrent(ctx);
          glXDestroyContext(gdi_display, ctx->p.ctx);
          ctx->p.ctx = 0;
          ctx->p.stamp++;
          tgt->flags &= ~FLAG_CINDIRECT;
        }
        /* evict pixmaps */
        X11DRV_GLX_DestroyTargets(ctx);
      }
      if (is_new) {
        X11DRV_DIBSECTION *dib = X11DRV_GLX_AttachAux(ctx, hBmp);
        Drawable draw;
#if 0
        HWND hwnd;

        /* find the drawable for the window that ddraw uses */
        hwnd = (HWND)lsurf->lpSurfMore->lpDDRAWReserved;
        draw = X11DRV_get_client_window(hwnd);
        TRACE("WND: %x Draw: %ld\n", hwnd, draw);
        if (!draw) {
          TRACE("no drawable, host off desktop\n");
          draw = X11DRV_get_client_window(GetDesktopWindow());
          TRACE("Owner WND: %x Drawable: 0x%08lx\n", hwnd, draw);
        }
#else
        draw = X11DRV_DD_PrimaryClient;
        TRACE("Drawable: 0x%08lx\n", draw);
#endif
        tgt->draw = draw;
        tgt->bmp = hBmp;
        tgt->dib = dib;
      }
      if (lsurf->ddsCaps.dwCaps & DDSCAPS_PRIMARYSURFACE)
        tgt->flags &= ~FLAG_BACKBUF;
      else
        tgt->flags |= FLAG_BACKBUF;
      if (tgt->flags & FLAG_WINDIRECT) tgt->flags |= FLAG_CINDIRECT;
    }
    else
#endif
    if (is_new) {
      /* evict old cached pixmap */
      X11DRV_GLX_DestroyTarget(ctx, tgt);
      if (!(tgt->flags & FLAG_OFFSCREEN)) {
        /* destroy context if it's direct for windows only */
        if (ctx->p.ctx && ((tgt->flags & FLAG_DIRECT) == FLAG_WDIRECT)) {
          X11DRV_GLX_CleanTargets(ctx);
          X11DRV_GLX_KillCurrent(ctx);
          glXDestroyContext(gdi_display, ctx->p.ctx);
          ctx->p.ctx = 0;
          ctx->p.stamp++;
        }
        /* need to render to "front" of pixmap */
        tgt->flags &= ~FLAG_BACKBUF;
        X11DRV_GLX_DestroyTargets(ctx);
      }
      /* prepare new pixmap */
      X11DRV_GLX_CreateOffscreen(ctx, tgt, hBmp);
      if (tgt->flags & FLAG_PINDIRECT) tgt->flags |= FLAG_CINDIRECT;
    }
  } else {
    /* this is probably GL_DirectDraw_Init trying to determine device caps,
     * it's probably OK to use the root window for this purpose, but the context
     * needs to use the root_windows's visual, not the standard one.  Some GL drivers
     * (ie: NVidia) choke if you have mismatched context/window visuals.  */
    XWindowAttributes root_win_attr;
    Visual *rootVisual;
    XVisualInfo template;
    int num;

    tgt->draw = root_window;
    tgt->flags = 0;
    tgt->bmp = 0;
    tgt->dib = 0;

    if (XGetWindowAttributes( gdi_display, root_window, &root_win_attr ))
    {
      rootVisual = root_win_attr.visual;
    }
    else
    {
      /* Get the default visual, since we can't seem to get the attributes from the
         Root Window.  Let's hope that the Root Window Visual matches the DefaultVisual */
      rootVisual = DefaultVisual( gdi_display, DefaultScreen(gdi_display) );
    }

    template.visualid = XVisualIDFromVisual(rootVisual);
    XFree(ctx->vis);
    ctx->vis = XGetVisualInfo(gdi_display, VisualIDMask, &template, &num);

  }
  ctx->nsurf = lsurf;

  LEAVE_GL();
}

/* This function is called with GLX resources locked */
static DWORD X11DRV_GLX_MakeCurrent(LPD3DCONTEXT ctx)
{
  LPD3DTARGET tgt = X11DRV_GLX_GetCurTgt(ctx);
  BOOL direct = !(tgt->flags & FLAG_CINDIRECT);
  int (*WineXHandler)(Display *, XErrorEvent *);
  HRESULT res = D3D_OK;
  LPD3DTCONTEXT old_ctx = X11DRV_GLX_GetCurrentT();
  LPD3DTCONTEXT new_ctx = old_ctx;
  GLXContext save_ctx, used_ctx;
  GLXDrawable save_draw, used_draw;
  BOOL borked;

  if (!tgt->lockedBmp) glXWaitX();
  save_ctx = used_ctx = glXGetCurrentContext();
  save_draw = used_draw = glXGetCurrentDrawable();
  borked = FALSE;

  if (old_ctx && ctx == old_ctx->parent && ctx->p.stamp == old_ctx->stamp &&
      old_ctx->parent->owner_ctx == old_ctx->ctx && ctx->nsurf == ctx->lsurf)
    return D3D_OK; /* already current */

  if (old_ctx && ctx != old_ctx->parent) {
    LPD3DCONTEXT parent = old_ctx->parent;
    DWORD tid = parent->owner_tid;

    if (save_ctx && save_ctx == ctx->owner_ctx) borked = TRUE;
    else if (save_ctx && save_ctx == parent->owner_ctx && tid != GetCurrentThreadId()) borked = TRUE;
    else if (parent->borked) borked = TRUE;
    if (borked) TRACE("GL thread borkage detected (1)\n");

    save_ctx = old_ctx->save_ctx;
    save_draw = old_ctx->save_draw;

    if (used_ctx != ctx->owner_ctx) {
      if (!glXMakeCurrent(gdi_display, old_ctx->save_draw, old_ctx->save_ctx))
        ERR("couldn't clear context (1)\n");
      used_ctx = old_ctx->save_ctx;
      used_draw = old_ctx->save_draw;
      TRACE("cleared (1) to: ctx %p, draw %08lx\n", old_ctx->save_ctx, old_ctx->save_draw);
    } else TRACE("skipped clear (1)\n");

    if (tid == GetCurrentThreadId()) {
      parent->owner_tid = 0;
    }
    if (&parent->p != old_ctx) {
      TRACE("killing slave context %p (ctx %p, parent=%p)\n", old_ctx, old_ctx->ctx, parent);
      if (old_ctx->prev) old_ctx->prev->next = old_ctx->next;
      if (old_ctx->next) old_ctx->next->prev = old_ctx->prev;
      if (tid == GetCurrentThreadId() && parent->owner_ctx != parent->p.ctx) {
        if (parent->p.ctx) glXDestroyContext(gdi_display, parent->p.ctx);
        parent->p.ctx = parent->owner_ctx;
      }
      parent->refs--;
      if (old_ctx->ctx && old_ctx->ctx != parent->p.ctx) glXDestroyContext(gdi_display, old_ctx->ctx);
      CloseHandle(old_ctx->thread);
      if (ctx->owner_tid) {
        /* now reuse old_ctx as new_ctx */
        memset(old_ctx, 0, sizeof(*old_ctx));
      } else {
        /* will use "main" context below */
        HeapFree(GetProcessHeap(), 0, old_ctx);
        new_ctx = NULL;
      }
    }
    else new_ctx = NULL;
    X11DRV_GLX_SetCurrentT(NULL);
    if (!parent->refs) {
      TRACE("killing master context %p (ctx %p)\n", parent, parent->p.ctx);
      if (used_ctx == parent->p.ctx) used_ctx = 0;
      if (parent->p.ctx) glXDestroyContext(gdi_display, parent->p.ctx);
      CloseHandle(parent->p.thread);
      HeapFree(GetProcessHeap(), 0, parent);
    }
    old_ctx = NULL;
  }
  else if (!old_ctx && save_ctx && save_ctx == ctx->owner_ctx) {
    TRACE("GL thread borkage detected (2), using: ctx %p, draw %08lx\n", ctx->p.save_ctx, ctx->p.save_draw);
    save_ctx = ctx->p.save_ctx;
    save_draw = ctx->p.save_draw;
    borked = TRUE;
  }
  if (ctx->borked) borked = TRUE;

  if (old_ctx && ctx->p.stamp != old_ctx->stamp) {
    if (!glXMakeCurrent(gdi_display, old_ctx->save_draw, old_ctx->save_ctx))
      ERR("couldn't clear context (2)\n");
    TRACE("cleared (2) to: ctx %p, draw %08lx\n", old_ctx->save_ctx, old_ctx->save_draw);
    used_ctx = old_ctx->save_ctx;
    used_draw = old_ctx->save_draw;
    if (old_ctx->ctx && old_ctx->ctx != old_ctx->parent->p.ctx) glXDestroyContext(gdi_display, old_ctx->ctx);
    new_ctx->ctx = 0;
    new_ctx->stamp = ctx->p.stamp;
    old_ctx = NULL;
  }

  if (!new_ctx) {
    /* use "main" context if unused */
    if (!ctx->owner_tid) {
      new_ctx = &ctx->p;
      borked = FALSE;
    } else new_ctx = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(D3DTCONTEXT));
  }
  if (!new_ctx->parent) {
    TRACE("initializing slave context %p (parent=%p)\n", new_ctx, ctx);
    new_ctx->parent = ctx;
    DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), GetCurrentProcess(),
                    &new_ctx->thread, 0, FALSE, DUPLICATE_SAME_ACCESS);
    new_ctx->stamp = ctx->p.stamp;
    ctx->refs++;
    /* this is threadsafe because the X11/GL lock is held */
    new_ctx->prev = &ctx->p;
    new_ctx->next = ctx->p.next;
    if (new_ctx->next) new_ctx->next->prev = new_ctx;
    ctx->p.next = new_ctx;
  }

  TRACE("current on entry: ctx %p, draw %08lx\n", save_ctx, save_draw);
  if (!old_ctx) {
    if (borked) {
#ifdef ENABLE_BORKEN_MIGRATE
      /* this is the only way my nvidia drivers work (swapping contexts
       * with glXMakeCurrent a few times fails for some reason) */
      TRACE("trying to work around borkage with a single context\n");
      new_ctx->ctx = ctx->p.ctx;
#endif
      ctx->borked = TRUE;
    }
    new_ctx->save_ctx = save_ctx;
    new_ctx->save_draw = save_draw;
  }
  if (new_ctx->ctx) {
    /* hmm, just re-establish this context? */
    if (ctx->owner_ctx && ctx->owner_ctx != new_ctx->ctx) {
      if (old_ctx) {
        save_ctx = old_ctx->save_ctx;
        save_draw = old_ctx->save_draw;
        if (!glXMakeCurrent(gdi_display, old_ctx->save_draw, old_ctx->save_ctx))
          ERR("couldn't clear context (4)\n");
        used_ctx = old_ctx->save_ctx;
        used_draw = old_ctx->save_draw;
        TRACE("cleared (4) to: ctx %p, draw %08lx\n", old_ctx->save_ctx, old_ctx->save_draw);
      }
      TRACE("copying context from %p to %p\n", ctx->owner_ctx, new_ctx->ctx);
      WineXHandler = XSetErrorHandler(GLXErrorHandler);
      GLXError = 0;
      XSync(gdi_display, False);
      glXCopyContext(gdi_display, ctx->owner_ctx, new_ctx->ctx, GL_ALL_ATTRIB_BITS);
      XSync(gdi_display, False);
      XSetErrorHandler(WineXHandler);
      if (GLXError) {
        ERR("context copy failed! (1)\n");
      }
    }
    if (used_ctx != new_ctx->ctx ||
        used_draw != tgt->draw) {
      BOOL ok = glXMakeCurrent(gdi_display, tgt->draw, new_ctx->ctx);
      TRACE("resetting context %p, ctx %p, bmp %08x, draw %08lx\n", ctx, new_ctx->ctx, tgt->bmp, tgt->draw);
      if (!ok) {
        ERR("failed to reset OpenGL context\n");
        TRACE("context remains: ctx %p, draw %08lx\n", glXGetCurrentContext(), glXGetCurrentDrawable());
        /* res = D3DERR_INVALID_DEVICE; */
      }
    }
  }
  while (!new_ctx->ctx) {
    BOOL ok;
    new_ctx->ctx = glXCreateContext(gdi_display, ctx->vis, ctx->owner_ctx, direct);
    if (!new_ctx->ctx) {
      WARN("unable to create OpenGL rendering context\n");
      res = D3DERR_INVALID_DEVICE;
      break;
    }
    if (ctx->owner_ctx && ctx->owner_ctx != new_ctx->ctx) {
      TRACE("copying context from %p to %p\n", ctx->owner_ctx, new_ctx->ctx);
      WineXHandler = XSetErrorHandler(GLXErrorHandler);
      GLXError = 0;
      XSync(gdi_display, False);
      glXCopyContext(gdi_display, ctx->owner_ctx, new_ctx->ctx, GL_ALL_ATTRIB_BITS);
      XSync(gdi_display, False);
      XSetErrorHandler(WineXHandler);
      if (GLXError) {
        /* if the copy is only for copy_aux, this is harmless, but if the app really
         * uses a different rendering thread, this could be bad */
        ERR("context copy failed! (2)\n");
      }
    }
    TRACE("context %p, ctx %p, bmp %08x, draw %08lx, %s\n", ctx, new_ctx->ctx, tgt->bmp, tgt->draw, direct?"direct":"indirect");
    WineXHandler = XSetErrorHandler(GLXErrorHandler);
    GLXError = 0;
    ok = glXMakeCurrent(gdi_display, tgt->draw, new_ctx->ctx);
    XSetErrorHandler(WineXHandler);
    if (GLXError) {
      glXDestroyContext(gdi_display, new_ctx->ctx);
      new_ctx->ctx = 0;
      if (direct) {
#ifndef NO_INDIRECT
        if (ctx->owner_ctx) {
#endif
          ERR("direct rendering to GLXPixmap rejected by OpenGL implementation\n");
          res = D3DERR_INVALID_DEVICE;
          break;
#ifndef NO_INDIRECT
        }
        ERR("direct rendering to GLXPixmap rejected by OpenGL implementation, switching to indirect rendering\n");
        direct = FALSE;
#endif
      } else {
        ERR("rendering to GLXPixmap rejected by OpenGL implementation\n");
        res = D3DERR_INVALID_DEVICE;
        break;
      }
    }
    else if (!ok) {
      glXDestroyContext(gdi_display, new_ctx->ctx);
      new_ctx->ctx = 0;
      ERR("failed to set OpenGL context\n");
      res = D3DERR_INVALID_DEVICE;
      break;
    }
  }
  if (res == D3D_OK) {
    TRACE("initializing context\n");
    /* success - update state to reflect new render target */
    ctx->lsurf = ctx->nsurf;
    ctx->surf = ctx->lsurf ? ctx->lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
    /* cache results of direct/indirect check */
    if (direct) {
      if (tgt->flags & FLAG_OFFSCREEN) tgt->flags |= FLAG_PDIRECT;
      else tgt->flags |= FLAG_WDIRECT;
    } else {
      if (tgt->flags & FLAG_OFFSCREEN) tgt->flags |= FLAG_PINDIRECT;
      else tgt->flags |= FLAG_WINDIRECT;
    }
    /* choose front or back buffer for rendering */
    if (tgt->flags & FLAG_BACKBUF) {
      glDrawBuffer(GL_BACK);
      TRACE("rendering to GL_BACK\n");
    } else {
      glDrawBuffer(GL_FRONT);
      TRACE("rendering to GL_FRONT\n");
    }
    ctx->owner_tid = GetCurrentThreadId();
    ctx->owner_ctx = new_ctx->ctx;
    X11DRV_GLX_SetCurrentT(new_ctx);
  } else {
    /* failure - revert state to old render target */
    TRACE("resetting context\n");
    if (ctx->lsurf && ctx->lsurf != ctx->nsurf)
      X11DRV_GLX_MakeTarget(ctx, ctx->lsurf);
  }

  return res;
}

static BOOL X11DRV_GLX_DoKillCurrent(void)
{
  LPD3DTCONTEXT old_ctx = X11DRV_GLX_GetCurrentT();
  LPD3DCONTEXT parent;
  DWORD tid;
  BOOL ret = FALSE;

  if (!old_ctx) return FALSE;
  parent = old_ctx->parent;
  tid = parent->owner_tid;
  glXMakeCurrent(gdi_display, old_ctx->save_draw, old_ctx->save_ctx);
  if (tid == GetCurrentThreadId()) {
    parent->owner_tid = 0;
  }
  if (&parent->p != old_ctx) {
    TRACE("killing slave context %p (ctx %p, parent=%p)\n", old_ctx, old_ctx->ctx, parent);
    if (old_ctx->prev) old_ctx->prev->next = old_ctx->next;
    if (old_ctx->next) old_ctx->next->prev = old_ctx->prev;
    if (tid == GetCurrentThreadId() && parent->owner_ctx != parent->p.ctx) {
      if (parent->p.ctx) glXDestroyContext(gdi_display, parent->p.ctx);
      parent->p.ctx = parent->owner_ctx;
    }
    parent->refs--;
    if (old_ctx->ctx && old_ctx->ctx != parent->p.ctx) glXDestroyContext(gdi_display, old_ctx->ctx);
    CloseHandle(old_ctx->thread);
    HeapFree(GetProcessHeap(), 0, old_ctx);
  }
  else ret = TRUE;
  X11DRV_GLX_SetCurrentT(NULL);
  if (!parent->refs) {
    TRACE("killing master context %p (ctx %p)\n", parent, parent->p.ctx);
    if (parent->p.ctx) glXDestroyContext(gdi_display, parent->p.ctx);
    CloseHandle(parent->p.thread);
    HeapFree(GetProcessHeap(), 0, parent);
  }
  return ret;
}

static BOOL X11DRV_GLX_KillCurrent(LPD3DCONTEXT ctx)
{
  BOOL ret;

  if (ctx != X11DRV_GLX_GetCurrent())
    return FALSE;

  ENTER_GL();
  ret = X11DRV_GLX_DoKillCurrent();
  LEAVE_GL();
  return ret;
}

void X11DRV_GLX_UserSignal(CONTEXT* pctx)
{
  LPD3DCONTEXT ctx = X11DRV_GLX_GetCurrent();
  if (ctx) {
    WARN("deselecting GLX context from signal handler - can be risky\n");
    /* the X11/GL lock is held by the calling thread */
    X11DRV_GLX_DoKillCurrent();
  }
}

void X11DRV_GLX_Detach(void)
{
  LPD3DCONTEXT ctx = X11DRV_GLX_GetCurrent();
  if (ctx) X11DRV_GLX_KillCurrent(ctx);
}

static DWORD PASCAL X11DRV_GLX_ContextCreate(LPD3DHAL_CONTEXTCREATEDATA data)
{
  int num;
  XVisualInfo template;
  LPD3DCONTEXT ctx = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(D3DCONTEXT));
  const char *gl_exts;
  const char *glx_exts;
  HRESULT res;

  TRACE("creating 3D context %p\n", ctx);
  ctx->p.parent = ctx;
  DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), GetCurrentProcess(),
                  &ctx->p.thread, 0, FALSE, DUPLICATE_SAME_ACCESS);
  ctx->refs = 1;
  /* FIXME: do this properly */
  USER_Lock();
  TRACE("locking render target\n");
  X11DRV_GLX_LockTarget(data->u2.lpDDSLcl, TRUE);
  ENTER_GL();
  template.visualid = XVisualIDFromVisual(visual);
  ctx->vis = XGetVisualInfo(gdi_display, VisualIDMask, &template, &num);

  /* check GLX version */
  glXQueryVersion(gdi_display, &ctx->glx_major, &ctx->glx_minor);
  TRACE("detected GLX version %d.%d\n", ctx->glx_major, ctx->glx_minor);

#ifdef HAVE_GLX_GETPROCADDRESS
  glx_exts = glXQueryExtensionsString(gdi_display, DefaultScreen(gdi_display));
  if (!glx_exts) ERR("failed to retrieve GLX extension list\n");
  else TRACE("GLX extensions: %s\n", glx_exts);
  if (glx_exts && strstr(glx_exts, "GLX_SGIX_fbconfig")) {
    TRACE("found GLX_SGIX_fbconfig\n");
    ctx->glXGetFBConfigAttribSGIX         = (void*)glXGetProcAddressARB("glXGetFBConfigAttribSGIX");
    ctx->glXChooseFBConfigSGIX            = (void*)glXGetProcAddressARB("glXChooseFBConfigSGIX");
    ctx->glXCreateGLXPixmapWithConfigSGIX = (void*)glXGetProcAddressARB("glXCreateGLXPixmapWithConfigSGIX");
    ctx->glXCreateContextWithConfigSGIX   = (void*)glXGetProcAddressARB("glXCreateContextWithConfigSGIX");
    ctx->glXGetVisualFromFBConfigSGIX     = (void*)glXGetProcAddressARB("glXGetVisualFromFBConfigSGIX");
    ctx->glXGetFBConfigFromVisualSGIX     = (void*)glXGetProcAddressARB("glXGetFBConfigFromVisualSGIX");
  }
  if (glx_exts && strstr(glx_exts, "GLX_SGIX_pbuffer")) {
    TRACE("found GLX_SGIX_pbuffer\n");
    ctx->glXCreateGLXPbufferSGIX  = (void*)glXGetProcAddressARB("glXCreateGLXPbufferSGIX");
    ctx->glXDestroyGLXPbufferSGIX = (void*)glXGetProcAddressARB("glXDestroyGLXPbufferSGIX");
    ctx->glXQueryGLXPbufferSGIX   = (void*)glXGetProcAddressARB("glXQueryGLXPbufferSGIX");
    ctx->glXSelectEventSGIX       = (void*)glXGetProcAddressARB("glXSelectEventSGIX");
    ctx->glXGetSelectedEventSGIX  = (void*)glXGetProcAddressARB("glXGetSelectedEventSGIX");
  }
#endif

  TRACE("preparing render target\n");
  X11DRV_GLX_MakeTarget(ctx, data->u2.lpDDSLcl);
  res = X11DRV_GLX_MakeCurrent(ctx);

  if (res == D3D_OK) {
    TRACE("preparing Z buffer\n");
    ctx->zbuf.flags = FLAG_ZBUFFER;
    X11DRV_GLX_MakeZbuffer(ctx, data->u3.lpDDSZLcl);

    /* check GL and GLX extensions */
#ifdef HAVE_GLX_GETPROCADDRESS
    gl_exts = glGetString(GL_EXTENSIONS);
    if (!gl_exts) ERR("failed to retrieve OpenGL extension list\n");
    if (gl_exts && strstr(gl_exts, "GL_ARB_multitexture")) {
      ctx->glActiveTextureARB = (void*)glXGetProcAddressARB("glActiveTextureARB");
      glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &ctx->texunits);
    }

    X11DRV_GLX_EB_Context_Init();
#endif

    X11DRV_GLX_SetupLogo(ctx);
  }

  glXWaitGL();
  LEAVE_GL();

  if (res == D3D_OK) {
    TRACE("created 3D context\n");
  } else {
    ERR("error creating 3D context\n");
    ENTER_GL();
    X11DRV_GLX_Destroy(ctx);
    LEAVE_GL();
    HeapFree(GetProcessHeap(),0,ctx);
    ctx = NULL;
  }
  X11DRV_GLX_LockTarget(data->u2.lpDDSLcl, FALSE);
  USER_Unlock();
  data->dwhContext = (DWORD)ctx;
  data->ddrval = res;
  return DDHAL_DRIVER_HANDLED;
}

static void CALLBACK X11DRV_GLX_ContextDestroyer(ULONG_PTR the_arg)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)the_arg;
  LPD3DTCONTEXT this = X11DRV_GLX_GetCurrentT();
  BOOL mt;

  /* weed out redundant calls */
  if (!this) return;
  if (this->parent != ctx) return;
  if (!(X11DRV_GLX_GetFlags(ctx) & FLAG_DESTROYED)) return;
  /* ok, this looks like a valid request */
  mt = X11DRV_GLX_KillCurrent(ctx);
  TRACE("main: %d\n", mt);
}

static DWORD PASCAL X11DRV_GLX_ContextDestroy(LPD3DHAL_CONTEXTDESTROYDATA data)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->dwhContext;
  LPD3DTCONTEXT this = X11DRV_GLX_GetCurrentT();
  DWORD refs;
  BOOL mt;

  TRACE("destroyed 3D context %p\n", ctx);

  if (ctx) {
    X11DRV_GLX_LockTarget(ctx->lsurf, TRUE);
    ENTER_GL();
    mt = X11DRV_GLX_Destroy(ctx);
    refs = --ctx->refs;
    if (!refs) {
      TRACE("killing master context %p (ctx %p)\n", ctx, ctx->p.ctx);
      if (ctx->p.ctx) {
        glXDestroyContext(gdi_display, ctx->p.ctx);
        ctx->p.ctx = 0;
      }
    }
    X11DRV_GLX_GetCurTgt(ctx)->flags |= FLAG_DESTROYED;
    LEAVE_GL();
    X11DRV_GLX_LockTarget(ctx->lsurf, FALSE);
    TRACE("refs: %ld, main: %d\n", refs, mt);
    if (!refs) {
      CloseHandle(ctx->p.thread);
      HeapFree(GetProcessHeap(), 0, ctx);
    }
    else {
      LPD3DTCONTEXT cur = &ctx->p;
      while (cur) {
        if (cur != this)
          QueueUserAPC(X11DRV_GLX_ContextDestroyer, cur->thread, (ULONG_PTR)ctx);
        cur = cur->next;
      }
    }
  }

  data->ddrval = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_GLX_ContextDestroyAll(LPD3DHAL_CONTEXTDESTROYALLDATA data)
{
  /* since this isn't a real display driver, we probably don't need to do anything here,
   * the kernel and the X server will clean up after Wine exits */
  data->ddrval = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static void X11DRV_GLX_SetupLogo(LPD3DCONTEXT ctx)
{
#ifdef TRANSGAMING_LOGO
/* makedep is overzealous and doesn't obey the #ifdef. */
/*#include "logo.c"*/
#error "Uncomment the above line, and comment this one."

    ctx->logo_size = logo_size;

    glGenTextures(1, &ctx->logo);
    glBindTexture(GL_TEXTURE_2D, ctx->logo);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ctx->logo_size, ctx->logo_size,
		 0, GL_RGBA, GL_UNSIGNED_BYTE, logo);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);

    ctx->logo_list = glGenLists(1);

    glNewList(ctx->logo_list, GL_COMPILE);

    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_COLOR_BUFFER_BIT
		 | GL_DEPTH_BUFFER_BIT | GL_TEXTURE_BIT | GL_TRANSFORM_BIT
		 | GL_VIEWPORT_BIT);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0.0, ctx->logo_size, ctx->logo_size, 0.0, -1, 1);

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_SCISSOR_TEST);
    glMatrixMode(GL_TEXTURE);
    if (ctx->texunits > 1) {
	int i;
	for (i=0; i<ctx->texunits; i++) {
	    ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
	    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	    glDisable(GL_TEXTURE_2D);

	    glPushMatrix();
	    glLoadIdentity();
	    glDisable(GL_TEXTURE_GEN_S);
	    glDisable(GL_TEXTURE_GEN_T);
	    glDisable(GL_TEXTURE_GEN_R);
	    glDisable(GL_TEXTURE_GEN_Q);
	}
	ctx->glActiveTextureARB(GL_TEXTURE0_ARB);
    }
    else
    {
	glDisable(GL_TEXTURE_2D);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glPushMatrix();
	glLoadIdentity();
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);
    }
    /* disable cubemapping, 3d */

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glDisable(GL_DEPTH_TEST);

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glColor4f(1, 1, 1, 0.333);
    if (gluCheckExtension("GL_EXT_secondary_color",
			  glGetString(GL_EXTENSIONS)))
    {
	glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
	glSecondaryColor3fEXT(0, 0, 0);
    }

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, ctx->logo);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, 0);
    /* LOD bias shouldn't make a difference here. */

    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);	glVertex2f(0, 0);
    glTexCoord2f(1, 0);	glVertex2f(ctx->logo_size, 0);
    glTexCoord2f(1, 1);	glVertex2f(ctx->logo_size, ctx->logo_size);
    glTexCoord2f(0, 1);	glVertex2f(0, ctx->logo_size);
    glEnd();

    glMatrixMode(GL_TEXTURE);
    if (ctx->texunits > 1) {
	int i;
	for (i=0; i<ctx->texunits; i++) {
	    ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
	    glPopMatrix();
	}
	ctx->glActiveTextureARB(GL_TEXTURE0_ARB);
    }
    else
    {
	glPopMatrix();
    }
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glPopAttrib();
    glPopClientAttrib();

    glEndList();
#endif
}

static void X11DRV_GLX_RenderLogo(LPD3DCONTEXT ctx)
{
#ifdef TRANSGAMING_LOGO
    glPushAttrib(GL_VIEWPORT_BIT);
    glViewport(ctx->lsurf->lpGbl->wWidth - ctx->logo_size,
	       ctx->lsurf->lpGbl->wHeight - ctx->logo_size,
	       ctx->logo_size, ctx->logo_size);

    glCallList(ctx->logo_list);

    glPopAttrib();
#endif
}

static const GLenum GL_cube_targets[6] = {
  GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
  GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
  GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB
};

static inline GLenum GL_tgt_load(LPDDRAWI_DDRAWSURFACE_LCL lcl)
{
  LPDDRAWI_DDRAWSURFACE_MORE more = lcl->lpSurfMore;
  if (more->ddsCapsEx.dwCaps2 & DDSCAPS2_CUBEMAP) {
      unsigned u;
      for (u=0; u<6; u++)
        if (more->ddsCapsEx.dwCaps2 & (0x400 << u))
           return GL_cube_targets[u];
  }
  return GL_TEXTURE_2D;
}

static inline GLenum GL_tgt_par(LPDDRAWI_DDRAWSURFACE_LCL lcl)
{
  LPDDRAWI_DDRAWSURFACE_MORE more = lcl->lpSurfMore;
 if (more->ddsCapsEx.dwCaps2 & DDSCAPS2_CUBEMAP)
    return GL_TEXTURE_CUBE_MAP_ARB;
  return GL_TEXTURE_2D;
}

static inline GLenum GL_tgt_bind(LPDDRAWI_DDRAWSURFACE_LCL lcl)
{
  LPDDRAWI_DDRAWSURFACE_MORE more = lcl->lpSurfMore;
  if (more->ddsCapsEx.dwCaps2 & DDSCAPS2_CUBEMAP)
    return GL_TEXTURE_BINDING_CUBE_MAP_ARB;
  return GL_TEXTURE_BINDING_2D;
}

static void X11DRV_GLX_UpdateTexture(LPD3DCONTEXT ctx)
{
  if (ctx->lsurf->ddsCaps.dwCaps & DDSCAPS_TEXTURE &&
      (LPD3DCONTEXT)ctx->lsurf->dwReserved1 == ctx) {
    LPDDRAWI_DDRAWSURFACE_GBL_MORE gblm = GET_LPDDRAWSURFACE_GBL_MORE(ctx->surf);
    if (gblm->dwDriverReserved) {
      GLint oldtex;
      GLenum target = GL_tgt_par(ctx->lsurf);
      int level = 0;
      /* FIXME: detect level */
      TRACE("reloading texture %d level %d from render target\n", gblm->dwDriverReserved, level);
      glGetIntegerv(GL_tgt_bind(ctx->lsurf), &oldtex);
      glBindTexture(target, (GLuint)gblm->dwDriverReserved);
      glCopyTexSubImage2D(GL_tgt_load(ctx->lsurf), level, 0, 0, 0, 0,
                          ctx->surf->wWidth, ctx->surf->wHeight);
      glBindTexture(target, oldtex);
    }
  }
}

/* Acquire all the locks needed during a scene capture */
static void SceneCaptureAcquireLocks(LPD3DCONTEXT ctx) 
{
  int s = DIB_Status_None, sz = DIB_Status_None;
  LPD3DTARGET cur;

  USER_Lock();
  X11DRV_GLX_MakeTarget(ctx, ctx->nsurf); /* used to be lsurf */
  cur = X11DRV_GLX_GetCurTgt(ctx);
  TRACE("Locking bitmap %08x\n", cur->bmp);
  cur->lockedBmp = (BITMAPOBJ *)GDI_GetObjPtr(cur->bmp, BITMAP_MAGIC);
  s = X11DRV_DIB_Lock(cur->lockedBmp, (cur->flags & FLAG_PIXMAP) ? DIB_Status_GdiMod : DIB_Status_AuxMod, FALSE);
  if (ctx->zbuf.bmp) {
    ctx->zbuf.lockedBmp = (BITMAPOBJ *)GDI_GetObjPtr(ctx->zbuf.bmp, BITMAP_MAGIC);
    sz = X11DRV_DIB_Lock(ctx->zbuf.lockedBmp, DIB_Status_AuxMod, FALSE);
    GDI_ReleaseObj(ctx->zbuf.bmp);
  }
  GDI_ReleaseObj(cur->bmp);
  USER_Unlock();
  ENTER_GL();
  X11DRV_GLX_MakeCurrent(ctx);
#ifdef RENDER_TO_WINDOW
  if ((s != DIB_Status_AuxMod) && !(cur->flags & FLAG_PIXMAP)) X11DRV_GLX_CopyDIB(ctx, cur);
#endif
  if (ctx->zbuf.bmp && sz != DIB_Status_AuxMod) X11DRV_GLX_CopyDIB(ctx, &ctx->zbuf);
}  

/* Release all the locks held during a scene capture */
static void SceneCaptureReleaseLocks(LPD3DCONTEXT ctx, LPD3DTARGET cur) 
{
  LEAVE_GL();
  TRACE("Unlocking bitmap %08x\n", cur->bmp);
  if (ctx->zbuf.bmp) {
    X11DRV_DIB_Unlock(ctx->zbuf.lockedBmp, TRUE);
    ctx->zbuf.lockedBmp = NULL;
  }
  X11DRV_DIB_Unlock(cur->lockedBmp, TRUE);
  cur->lockedBmp = NULL;
}

static DWORD PASCAL X11DRV_GLX_SceneCapture(LPD3DHAL_SCENECAPTUREDATA data)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->dwhContext;
  LPD3DTARGET cur;

  switch (data->dwFlag) {
  case D3DHAL_SCENE_CAPTURE_START:
    SceneCaptureAcquireLocks(ctx);
    break;
  case D3DHAL_SCENE_CAPTURE_END:
    cur = X11DRV_GLX_GetCurTgt(ctx);
    X11DRV_GLX_RenderLogo(ctx);
    /* glFinish(); */
    X11DRV_GLX_UpdateTexture(ctx);
    glXWaitGL();
    SceneCaptureReleaseLocks(ctx, cur);
    break;
  }
  data->ddrval = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_GLX_SetRenderTarget(LPD3DHAL_SETRENDERTARGETDATA data)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->dwhContext;
  HRESULT res = D3DERR_INVALID_DEVICE;

  TRACE("target: %p, %p\n", data->u1.lpDDSLcl, data->u2.lpDDSZLcl);
  if (ctx) {
    ENTER_GL();
    X11DRV_GLX_MakeTarget(ctx, data->u1.lpDDSLcl);
    res = X11DRV_GLX_MakeCurrent(ctx);
    glXWaitGL();
    LEAVE_GL();
    if (res == D3D_OK) {
      X11DRV_GLX_MakeZbuffer(ctx, data->u2.lpDDSZLcl);
    }
  }

  data->ddrval = res;
  return DDHAL_DRIVER_HANDLED;
}

static LPDDHAL_CREATESURFACE X11DRV_GLX_old_create_surface;
static LPDDHALSURFCB_DESTROYSURFACE X11DRV_GLX_old_destroy_surface;
static LPDDHALSURFCB_FLIP X11DRV_GLX_old_flip;
static LPDDHALSURFCB_LOCK X11DRV_GLX_old_lock;
static LPDDHALSURFCB_UNLOCK X11DRV_GLX_old_unlock;

static DWORD PASCAL X11DRV_GLX_CreateSurface(LPDDHAL_CREATESURFACEDATA data)
{
  LPDDRAWI_DDRAWSURFACE_LCL lcl = *data->lplpSList;
  LPDDRAWI_DDRAWSURFACE_GBL gbl = lcl->lpGbl;
  LPDDRAWI_DDRAWSURFACE_GBL_MORE gblm = GET_LPDDRAWSURFACE_GBL_MORE(gbl);
  LPDDSURFACEDESC2 desc = (LPDDSURFACEDESC2)data->lpDDSurfaceDesc;
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)lcl->dwReserved1;

  if (desc->ddsCaps.dwCaps & DDSCAPS_TEXTURE) {
    /* only process late allocations (with a context) */
    if (ctx && !(desc->ddsCaps.dwCaps2 & DDSCAPS2_MIPMAPSUBLEVEL)) {
      ENTER_GL();
      X11DRV_GLX_MakeCurrent(ctx);
      glGenTextures(1, (GLuint*)&gblm->dwDriverReserved);
      LEAVE_GL();
      TRACE("generated texture %d\n", gblm->dwDriverReserved);
    }
  }

  if (X11DRV_GLX_old_create_surface)
    return X11DRV_GLX_old_create_surface(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_GLX_DestroySurface(LPDDHAL_DESTROYSURFACEDATA data)
{
  LPDDRAWI_DDRAWSURFACE_LCL lcl = data->lpDDSurface;
  /* LPDDRAWI_DDRAWSURFACE_MORE lclm = lcl->lpSurfMore; */
  LPDDRAWI_DDRAWSURFACE_GBL gbl = lcl->lpGbl;
  LPDDRAWI_DDRAWSURFACE_GBL_MORE gblm = GET_LPDDRAWSURFACE_GBL_MORE(gbl);
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)lcl->dwReserved1;

  if (lcl->ddsCaps.dwCaps & DDSCAPS_TEXTURE) {
    if (ctx) {
      LPD3DTARGET tgt = X11DRV_GLX_FindTarget(ctx, lcl);
      if (gblm->dwDriverReserved) {
        TRACE("deleted texture %d\n", gblm->dwDriverReserved);
        ENTER_GL();
        X11DRV_GLX_MakeCurrent(ctx);
        glDeleteTextures(1, (GLuint*)&gblm->dwDriverReserved);
        gblm->dwDriverReserved = 0;
        if (!X11DRV_GLX_GetCurTgt(ctx)->lockedBmp) glXWaitGL();
        LEAVE_GL();
      }
      if (tgt) {
        TRACE("destroying a known render target\n");
        X11DRV_GLX_CleanTarget(ctx, tgt);
        X11DRV_GLX_DestroyTarget(ctx, tgt);
        /* do we need to do more? */
      }
    }
  }

  if (X11DRV_GLX_old_destroy_surface)
    return X11DRV_GLX_old_destroy_surface(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_GLX_Flip(LPDDHAL_FLIPDATA data)
{
  if (data->lpSurfCurr == X11DRV_DD_Primary) {
    LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->lpSurfTarg->dwReserved1;
    TRACE("front=%p, back=%p, ctx=%p\n", data->lpSurfCurr, data->lpSurfTarg, ctx);
    if (ctx && ctx->lsurf == data->lpSurfTarg &&
        !(X11DRV_GLX_GetFlags(ctx) & FLAG_OFFSCREEN)) {
      /* we have a windowed render target */
      int state;
      LPD3DTARGET cur = X11DRV_GLX_GetCurTgt(ctx);
      BITMAPOBJ *bmp = NULL;

      if (cur->lockedBmp) {
        /* don't know if this could cause problems or not */
        TRACE("calling Flip on surface locked for rendering\n");
        state = cur->dib ? cur->dib->status : DIB_Status_None;
        bmp = cur->lockedBmp;
      }
      else
      {
        bmp = (BITMAPOBJ *)GDI_GetObjPtr(cur->bmp, BITMAP_MAGIC);
        state = X11DRV_DIB_Lock(bmp, DIB_Status_None, FALSE);
      }

      TRACE("performing GLX flip: state = %d\n", state );

      ENTER_GL();
      X11DRV_GLX_MakeCurrent(ctx);
      /* If we're not GL synced then do so now before flipping the GL buffer */
      if (state != DIB_Status_AuxMod) {
        X11DRV_DIB_Coerce(bmp, DIB_Status_AuxMod, FALSE);
        X11DRV_GLX_CopyDIB(ctx, cur);
      }

      glXSwapBuffers(gdi_display, cur->draw);
      if (!cur->lockedBmp) glXWaitGL();
      LEAVE_GL();
      TRACE("complete\n");

      cur->flags ^= FLAG_FLIPPED;

      if (!cur->lockedBmp)
      {
        X11DRV_DIB_Unlock(bmp, TRUE);
        GDI_ReleaseObj(cur->bmp);
      } else {
	/* this is to guard against applications holding the scene capture locks
	 * too long and starving another thread that may be waiting for them.
	 * e.g. app starts SceneCapture, but doesn't end SceneCapture for
	 * a long period of time, instead updating the screen using GLX_Flip. */
	SceneCaptureReleaseLocks(ctx, cur);
	SceneCaptureAcquireLocks(ctx);
      }

      return DDHAL_DRIVER_HANDLED;
    }
  }
  if (X11DRV_GLX_old_flip)
    return X11DRV_GLX_old_flip(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_GLX_Lock(LPDDHAL_LOCKDATA data)
{
  LPDDRAWI_DDRAWSURFACE_LCL lcl = data->lpDDSurface;
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)lcl->dwReserved1;
  if (ctx) {
    LPD3DTARGET tgt = X11DRV_GLX_FindTarget(ctx, lcl);
    if (tgt) {
      BITMAPOBJ *bmp = (BITMAPOBJ *)GDI_GetObjPtr(tgt->bmp, BITMAP_MAGIC);
      int state = X11DRV_DIB_Lock(bmp, DIB_Status_None, FALSE);
      TRACE("tgt=%p\n", tgt);
      tgt->locks++;
      if (tgt->locks == 1) {
        /* first lock */
        if (data->bHasRect && (state == DIB_Status_AuxMod)) {
          TRACE("using rectlock\n");
          /* only the specified rectangle will be accessed by the app,
           * so the subsequent copy_aux only needs to copy this rect,
           * not the whole surface */
          tgt->rect_lock = TRUE;
          tgt->rect = data->rArea;
        }
        tgt->lockf = data->dwFlags;
      } else {
        if (tgt->rect_lock) {
          /* another lock, cancel the above optimization, must read everything */
          if (state == DIB_Status_AuxMod) tgt->rect_lock = FALSE; 
          else {
            FIXME("read entire framebuffer! (except already-locked region)\n");
          }
        }
        tgt->lockf = 0;
      }
      X11DRV_DIB_Unlock(bmp, TRUE);
      GDI_ReleaseObj(tgt->bmp);
      /* return here, don't move on to x11ddraw */
      return DDHAL_DRIVER_NOTHANDLED;
    }
  }
  if (X11DRV_GLX_old_lock)
    return X11DRV_GLX_old_lock(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_GLX_Unlock(LPDDHAL_UNLOCKDATA data)
{
  LPDDRAWI_DDRAWSURFACE_LCL lcl = data->lpDDSurface;
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)lcl->dwReserved1;
  if (ctx) {
    LPD3DTARGET tgt = X11DRV_GLX_FindTarget(ctx, lcl);
    if (tgt) {
      BITMAPOBJ *bmp = (BITMAPOBJ *)GDI_GetObjPtr(tgt->bmp, BITMAP_MAGIC);
      int state = X11DRV_DIB_Lock(bmp, DIB_Status_None, FALSE);
      BOOL ret = FALSE;
      TRACE("tgt=%p\n", tgt);
      if (tgt->locks) {
        tgt->locks--;
        if (tgt->locks == 0) {
          /* last lock - copy back to render target immediately if we
           * a) just locked a small rectangle
           *    (so the rest of the DIB section is invalid), or
           * b) are currently rendering a scene
           *    (so subsequent rendering commands use the new image) */
          if (tgt->rect_lock || tgt->lockedBmp) {
            if (state != DIB_Status_AuxMod) {
              /* write modified rectangle back */
              X11DRV_DIB_Coerce(bmp, DIB_Status_AuxMod, FALSE);
              if (tgt->lockf & DDLOCK_READONLY) {
                TRACE("read-only lock released\n");
                tgt->dib->status = DIB_Status_AuxMod;
              } else {
                LPDDRAWI_DDRAWSURFACE_LCL old = ctx->lsurf;
                LPD3DTARGET cur = X11DRV_GLX_GetCurTgt(ctx);
                BITMAPOBJ *oldbmp;
                ENTER_GL();
                if (old != lcl) X11DRV_GLX_MakeTarget(ctx, lcl);
                X11DRV_GLX_MakeCurrent(ctx);
                oldbmp = tgt->lockedBmp;
                tgt->lockedBmp = bmp;
                X11DRV_GLX_CopyDIB(ctx, tgt);
                tgt->lockedBmp = oldbmp;
                if (old != lcl) {
                  X11DRV_GLX_MakeTarget(ctx, old);
                  if (cur->lockedBmp) X11DRV_GLX_MakeCurrent(ctx);
                }
                if (!cur->lockedBmp) glXWaitGL();
                LEAVE_GL();
              }
            }
          }
          tgt->rect_lock = FALSE;
          tgt->lockf = 0;
        }
        /* return here, don't move on to x11ddraw */
        ret = TRUE;
      }
      X11DRV_DIB_Unlock(bmp, TRUE);
      GDI_ReleaseObj(tgt->bmp);
      if (ret) return DDHAL_DRIVER_NOTHANDLED;
    }
  }
  if (X11DRV_GLX_old_unlock)
    return X11DRV_GLX_old_unlock(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}

static D3DHAL_CALLBACKS d3dhal_callbacks = {
  sizeof(D3DHAL_CALLBACKS),
  X11DRV_GLX_ContextCreate,
  X11DRV_GLX_ContextDestroy,
  X11DRV_GLX_ContextDestroyAll,
  X11DRV_GLX_SceneCapture,
  /* use lpReserved10 for something useful, like GetProcAddress */
#ifdef HAVE_GLX_GETPROCADDRESS
  glXGetProcAddressARB
#else
  NULL
#endif
};

static D3DHAL_CALLBACKS2 d3dhal_callbacks2 = {
  sizeof(D3DHAL_CALLBACKS2),
  0x1f, /* all callbacks are 32-bit */
  X11DRV_GLX_SetRenderTarget
};

static D3DHAL_GLOBALDRIVERDATA d3dhal_driverdata = {
  sizeof(D3DHAL_GLOBALDRIVERDATA),
  {	/* hwCaps */
   sizeof(D3DDEVICEDESC_V1),
   D3DDD_WINE_OPENGL_DEVICE
  },
  0,	/* dwNumVertices */
  0,	/* dwNumClipVertices */
  0,	/* dwNumTextureFormats */
  NULL	/* lpTextureFormats */
};

static X11DRIVERINFO driver_info[] = {
  {&GUID_D3DCallbacks2, sizeof(D3DHAL_CALLBACKS2), &d3dhal_callbacks2, NULL},
  {&GUID_WineGLD3DCallbacks, sizeof(struct GL_D3D_DriverCallbacks),
   &X11DRV_GLX_GL_D3D_driver_callbacks, NULL}
};

int X11DRV_GLX_CreateDriver(LPDDHALINFO info)
{
  LPX11DEVICE device = (LPX11DEVICE)info->lpPDevice;

  if (!glx_available) return 0;

  /* attach Direct3D callbacks */
  info->lpD3DGlobalDriverData = (ULONG_PTR)&d3dhal_driverdata;
  info->lpD3DHALCallbacks = (ULONG_PTR)&d3dhal_callbacks;
  info->lpDDExeBufCallbacks = &X11DRV_GLX_exebufcallbacks;

  /* add 3D caps bits */
  info->ddCaps.dwCaps |= DDCAPS_3D;
  info->ddCaps.dwCaps2 |= DDCAPS2_NO2DDURING3DSCENE;
  info->ddCaps.ddsCaps.dwCaps |= DDSCAPS_TEXTURE | DDSCAPS_3DDEVICE |
                                 DDSCAPS_ZBUFFER | DDSCAPS_MIPMAP |
                                 DDSCAPS_EXECUTEBUFFER | DDSCAPS_OPTIMIZED;

  /* attach our D3D callback structures to driver info linked list */
  driver_info[0].lpNext = driver_info+1;
  driver_info[1].lpNext = device->lpInfo;
  device->lpInfo = driver_info;

  /* chain some surface callbacks */
  X11DRV_GLX_old_create_surface = info->lpDDCallbacks->CreateSurface;
  X11DRV_GLX_old_destroy_surface = info->lpDDSurfaceCallbacks->DestroySurface;
  X11DRV_GLX_old_flip = info->lpDDSurfaceCallbacks->Flip;
  X11DRV_GLX_old_lock = info->lpDDSurfaceCallbacks->Lock;
  X11DRV_GLX_old_unlock = info->lpDDSurfaceCallbacks->Unlock;
  info->lpDDCallbacks->CreateSurface = X11DRV_GLX_CreateSurface;
  info->lpDDSurfaceCallbacks->DestroySurface = X11DRV_GLX_DestroySurface;
  info->lpDDSurfaceCallbacks->Flip = X11DRV_GLX_Flip;
  info->lpDDSurfaceCallbacks->Lock = X11DRV_GLX_Lock;
  info->lpDDSurfaceCallbacks->Unlock = X11DRV_GLX_Unlock;

  return TRUE;
}
